@@ -1,6 +1,3 @@
-#include "array.h"
-#include "buffer.h"
-
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -9,11 +6,14 @@
 #include <errno.h>
 #include <assert.h>
 
+#include "array.h"
+#include "buffer.h"
+
 array *array_init(void) {
 	array *a;
 
 	a = calloc(1, sizeof(*a));
-	force_assert(a);
+	assert(a);
 
 	a->next_power_of_2 = 1;
 
@@ -24,8 +24,6 @@ array *array_init_array(array *src) {
 	size_t i;
 	array *a = array_init();
 
-	if (0 == src->size) return a;
-
 	a->used = src->used;
 	a->size = src->size;
 	a->next_power_of_2 = src->next_power_of_2;
@@ -74,7 +72,7 @@ void array_reset(array *a) {
 data_unset *array_pop(array *a) {
 	data_unset *du;
 
-	force_assert(a->used != 0);
+	assert(a->used != 0);
 
 	a->used --;
 	du = a->data[a->used];
@@ -98,7 +96,7 @@ static int array_get_index(array *a, const char *key, size_t keylen, int *rndx)
 		} else if (pos >= (int)a->used) {
 			pos -= i;
 		} else {
-			cmp = buffer_caseless_compare(key, keylen, CONST_BUF_LEN(a->data[a->sorted[pos]]->key));
+			cmp = buffer_caseless_compare(key, keylen, a->data[a->sorted[pos]]->key->ptr, a->data[a->sorted[pos]]->key->used);
 
 			if (cmp == 0) {
 				/* found */
@@ -118,10 +116,10 @@ static int array_get_index(array *a, const char *key, size_t keylen, int *rndx)
 	return ndx;
 }
 
-data_unset *array_get_element(array *a, const char *key) {
+data_unset *array_get_element(array *a, const char *key, size_t keylen) {
 	int ndx;
 
-	if (-1 != (ndx = array_get_index(a, key, strlen(key), NULL))) {
+	if (-1 != (ndx = array_get_index(a, key, keylen + 1, NULL))) {
 		/* found, leave here */
 
 		return a->data[ndx];
@@ -132,27 +130,26 @@ data_unset *array_get_element(array *a, const char *key) {
 
 data_unset *array_get_unused_element(array *a, data_type_t t) {
 	data_unset *ds = NULL;
-	unsigned int i;
 
-	for (i = a->used; i < a->size; i++) {
-		if (a->data[i] && a->data[i]->type == t) {
-			ds = a->data[i];
+	UNUSED(t);
 
-			/* make empty slot at a->used for next insert */
-			a->data[i] = a->data[a->used];
-			a->data[a->used] = NULL;
+	if (a->size == 0) return NULL;
 
-			return ds;
-		}
+	if (a->used == a->size) return NULL;
+
+	if (a->data[a->used]) {
+		ds = a->data[a->used];
+
+		a->data[a->used] = NULL;
 	}
 
-	return NULL;
+	return ds;
 }
 
 void array_set_key_value(array *hdrs, const char *key, size_t key_len, const char *value, size_t val_len) {
 	data_string *ds_dst;
 
-	if (NULL != (ds_dst = (data_string *)array_get_element(hdrs, key))) {
+	if (NULL != (ds_dst = (data_string *)array_get_element(hdrs, key, key_len))) {
 		buffer_copy_string_len(ds_dst->value, value, val_len);
 		return;
 	}
@@ -166,12 +163,23 @@ void array_set_key_value(array *hdrs, const char *key, size_t key_len, const cha
 	array_insert_unique(hdrs, (data_unset *)ds_dst);
 }
 
+void array_append_key_value(array *hdrs, const char *key, size_t key_len, const char *value, size_t val_len) {
+	data_string *ds_dst;
+
+	if (NULL == (ds_dst = (data_string *)array_get_unused_element(hdrs, TYPE_STRING))) {
+		ds_dst = data_string_init();
+	}
+
+	buffer_copy_string_len(ds_dst->key, key, key_len);
+	buffer_copy_string_len(ds_dst->value, value, val_len);
+	array_insert_unique(hdrs, (data_unset *)ds_dst);
+}
+
 /* replace or insert data, return the old one with the same key */
 data_unset *array_replace(array *a, data_unset *du) {
 	int ndx;
 
-	force_assert(NULL != du);
-	if (-1 == (ndx = array_get_index(a, CONST_BUF_LEN(du->key), NULL))) {
+	if (-1 == (ndx = array_get_index(a, du->key->ptr, du->key->used, NULL))) {
 		array_insert_unique(a, du);
 		return NULL;
 	} else {
@@ -186,19 +194,19 @@ int array_insert_unique(array *a, data_unset *str) {
 	int pos = 0;
 	size_t j;
 
-	/* generate unique index if neccesary */
-	if (buffer_is_empty(str->key) || str->is_index_key) {
-		buffer_copy_int(str->key, a->unique_ndx++);
+	/* generate unique index if necessary */
+	if (str->key->used == 0 || str->is_index_key) {
+		buffer_copy_long(str->key, a->unique_ndx++);
 		str->is_index_key = 1;
 	}
 
 	/* try to find the string */
-	if (-1 != (ndx = array_get_index(a, CONST_BUF_LEN(str->key), &pos))) {
+	if (-1 != (ndx = array_get_index(a, str->key->ptr, str->key->used, &pos))) {
 		/* found, leave here */
 		if (a->data[ndx]->type == str->type) {
 			str->insert_dup(a->data[ndx], str);
 		} else {
-			SEGFAULT();
+			fprintf(stderr, "a\n");
 		}
 		return 0;
 	}
@@ -214,32 +222,29 @@ int array_insert_unique(array *a, data_unset *str) {
 		a->size   = 16;
 		a->data   = malloc(sizeof(*a->data)     * a->size);
 		a->sorted = malloc(sizeof(*a->sorted)   * a->size);
-		force_assert(a->data);
-		force_assert(a->sorted);
+		assert(a->data);
+		assert(a->sorted);
 		for (j = a->used; j < a->size; j++) a->data[j] = NULL;
 	} else if (a->size == a->used) {
 		a->size  += 16;
 		a->data   = realloc(a->data,   sizeof(*a->data)   * a->size);
 		a->sorted = realloc(a->sorted, sizeof(*a->sorted) * a->size);
-		force_assert(a->data);
-		force_assert(a->sorted);
+		assert(a->data);
+		assert(a->sorted);
 		for (j = a->used; j < a->size; j++) a->data[j] = NULL;
 	}
 
 	ndx = (int) a->used;
 
-	/* make sure there is nothing here */
-	if (a->data[ndx]) a->data[ndx]->free(a->data[ndx]);
-
 	a->data[a->used++] = str;
 
 	if (pos != ndx &&
 	    ((pos < 0) ||
-	     buffer_caseless_compare(CONST_BUF_LEN(str->key), CONST_BUF_LEN(a->data[a->sorted[pos]]->key)) > 0)) {
+	     buffer_caseless_compare(str->key->ptr, str->key->used, a->data[a->sorted[pos]]->key->ptr, a->data[a->sorted[pos]]->key->used) > 0)) {
 		pos++;
 	}
 
-	/* move everything on step to the right */
+	/* move everything one step to the right */
 	if (pos != ndx) {
 		memmove(a->sorted + (pos + 1), a->sorted + (pos), (ndx - pos) * sizeof(*a->sorted));
 	}
@@ -354,30 +359,30 @@ int main (int argc, char **argv) {
 	a = array_init();
 
 	ds = data_string_init();
-	buffer_copy_string_len(ds->key, CONST_STR_LEN("abc"));
-	buffer_copy_string_len(ds->value, CONST_STR_LEN("alfrag"));
+	buffer_copy_string(ds->key, "abc");
+	buffer_copy_string(ds->value, "alfrag");
 
 	array_insert_unique(a, (data_unset *)ds);
 
 	ds = data_string_init();
-	buffer_copy_string_len(ds->key, CONST_STR_LEN("abc"));
-	buffer_copy_string_len(ds->value, CONST_STR_LEN("hameplman"));
+	buffer_copy_string(ds->key, "abc");
+	buffer_copy_string(ds->value, "hameplman");
 
 	array_insert_unique(a, (data_unset *)ds);
 
 	ds = data_string_init();
-	buffer_copy_string_len(ds->key, CONST_STR_LEN("123"));
-	buffer_copy_string_len(ds->value, CONST_STR_LEN("alfrag"));
+	buffer_copy_string(ds->key, "123");
+	buffer_copy_string(ds->value, "alfrag");
 
 	array_insert_unique(a, (data_unset *)ds);
 
 	dc = data_count_init();
-	buffer_copy_string_len(dc->key, CONST_STR_LEN("def"));
+	buffer_copy_string(dc->key, "def");
 
 	array_insert_unique(a, (data_unset *)dc);
 
 	dc = data_count_init();
-	buffer_copy_string_len(dc->key, CONST_STR_LEN("def"));
+	buffer_copy_string(dc->key, "def");
 
 	array_insert_unique(a, (data_unset *)dc);
 
