@@ -1,55 +1,59 @@
 #ifndef _BASE_H_
 #define _BASE_H_
 
-#ifdef HAVE_CONFIG_H
-# include "config.h"
-#endif
-#include "settings.h"
-
 #include <sys/types.h>
-#include <sys/time.h>
 #include <sys/stat.h>
 
-#include <limits.h>
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
+#include <limits.h>
 #ifdef HAVE_STDINT_H
 # include <stdint.h>
 #endif
-
 #ifdef HAVE_INTTYPES_H
 # include <inttypes.h>
 #endif
 
+#include "settings.h"
+
+/* include glib.h before our buffer.h and array.h to make sure their parameter-names
+ * don't clash with our type-names */
+#include <glib.h>
+
 #include "buffer.h"
 #include "array.h"
 #include "chunk.h"
+#include "filter.h"
 #include "keyvalue.h"
+#include "settings.h"
 #include "fdevent.h"
 #include "sys-socket.h"
-#include "splaytree.h"
-#include "etag.h"
-
+#include "http_req.h"
 
 #if defined HAVE_LIBSSL && defined HAVE_OPENSSL_SSL_H
 # define USE_OPENSSL
 # include <openssl/ssl.h>
-# if ! defined OPENSSL_NO_TLSEXT && ! defined SSL_CTRL_SET_TLSEXT_HOSTNAME
-#  define OPENSSL_NO_TLSEXT
-# endif
 #endif
 
-#ifdef HAVE_FAM_H
-# include <fam.h>
+#ifdef HAVE_SYS_INOTIFY_H
+# include <sys/inotify.h>
 #endif
 
+#ifdef USE_LINUX_AIO_SENDFILE
+# include <libaio.h>
+#endif
+
+#ifdef USE_POSIX_AIO
+# include <aio.h>
+#endif
+
+/** some compat */
 #ifndef O_BINARY
 # define O_BINARY 0
 #endif
 
-#ifndef O_LARGEFILE
-# define O_LARGEFILE 0
-#endif
-
 #ifndef SIZE_MAX
 # ifdef SIZE_T_MAX
 #  define SIZE_MAX SIZE_T_MAX
@@ -65,7 +69,7 @@
 #ifdef __APPLE__
 #include <crt_externs.h>
 #define environ (* _NSGetEnviron())
-#else
+#elif !defined(_WIN32)
 extern char **environ;
 #endif
 
@@ -76,7 +80,8 @@ typedef int socklen_t;
 
 /* solaris and NetBSD 1.3.x again */
 #if (!defined(HAVE_STDINT_H)) && (!defined(HAVE_INTTYPES_H)) && (!defined(uint32_t))
-# define uint32_t u_int32_t
+/* # define uint32_t u_int32_t */
+typedef unsigned __int32 uint32_t;
 #endif
 
 
@@ -84,6 +89,8 @@ typedef int socklen_t;
 # define SHUT_WR 1
 #endif
 
+#include "settings.h"
+
 typedef enum { T_CONFIG_UNSET,
 		T_CONFIG_STRING,
 		T_CONFIG_SHORT,
@@ -124,25 +131,6 @@ typedef struct {
 	short factor;
 } fcgi_connections;
 
-
-typedef union {
-#ifdef HAVE_IPV6
-	struct sockaddr_in6 ipv6;
-#endif
-	struct sockaddr_in ipv4;
-#ifdef HAVE_SYS_UN_H
-	struct sockaddr_un un;
-#endif
-	struct sockaddr plain;
-} sock_addr;
-
-/* fcgi_response_header contains ... */
-#define HTTP_STATUS         BV(0)
-#define HTTP_CONNECTION     BV(1)
-#define HTTP_CONTENT_LENGTH BV(2)
-#define HTTP_DATE           BV(3)
-#define HTTP_LOCATION       BV(4)
-
 typedef struct {
 	/** HEADER */
 	/* the request-line */
@@ -154,19 +142,12 @@ typedef struct {
 	http_method_t  http_method;
 	http_version_t http_version;
 
-	buffer *request_line;
-
-	/* strings to the header */
-	buffer *http_host; /* not alloced */
-	const char   *http_range;
-	const char   *http_content_type;
-	const char   *http_if_modified_since;
-	const char   *http_if_none_match;
+	buffer *http_host;
 
 	array  *headers;
 
 	/* CONTENT */
-	size_t content_length; /* returned by strtoul() */
+	off_t   content_length; /* returned by strtoul() */
 
 	/* internal representation */
 	int     accept_encoding;
@@ -177,7 +158,7 @@ typedef struct {
 
 typedef struct {
 	off_t   content_length;
-	int     keep_alive;               /* used by  the subrequests in proxy, cgi and fcgi to say the subrequest was keep-alive or not */
+	int     keep_alive;               /* used by the subrequests in proxy, cgi and fcgi to say whether the subrequest was keep-alive or not */
 
 	array  *headers;
 
@@ -187,15 +168,11 @@ typedef struct {
 } response;
 
 typedef struct {
-	buffer *scheme; /* scheme without colon or slashes ( "http" or "https" ) */
-
-	/* authority with optional portnumber ("site.name" or "site.name:8080" ) NOTE: without "username:password@" */
+	buffer *scheme;
 	buffer *authority;
-
-	/* path including leading slash ("/" or "/index.html") - urldecoded, and sanitized  ( buffer_path_simplify() && buffer_urldecode_path() ) */
 	buffer *path;
-	buffer *path_raw; /* raw path, as sent from client. no urldecoding or path simplifying */
-	buffer *query; /* querystring ( everything after "?", ie: in "/index.php?foo=1", query is "foo=1" ) */
+	buffer *path_raw;
+	buffer *query;
 } request_uri;
 
 typedef struct {
@@ -216,28 +193,34 @@ typedef struct {
 
 	time_t stat_ts;
 
+	enum {
+		STAT_CACHE_ENTRY_UNSET, 
+		STAT_CACHE_ENTRY_ASYNC_STAT, 
+		STAT_CACHE_ENTRY_STAT_FINISHED
+	} state;
+
 #ifdef HAVE_LSTAT
 	char is_symlink;
 #endif
 
-#ifdef HAVE_FAM_H
-	int    dir_version;
+#if defined(HAVE_SYS_INOTIFY_H)
+	int    dir_version; /* when this entry was created the dir had this version */
+	int    dir_ndx;
 #endif
 
 	buffer *content_type;
 } stat_cache_entry;
 
 typedef struct {
-	splay_tree *files; /* the nodes of the tree are stat_cache_entry's */
+	GHashTable *files; /* a HashTable of stat_cache_entries for the files */
+	GHashTable *dirs;  /* a HashTable of stat_cache_entries for the dirs */
 
-	buffer *dir_name; /* for building the dirname from the filename */
-#ifdef HAVE_FAM_H
-	splay_tree *dirs; /* the nodes of the tree are fam_dir_entry */
+	buffer *dir_name;  /* for building the dirname from the filename */
+	buffer *hash_key;  /* tmp-buf for building the hash-key */
 
-	FAMConnection fam;
-	int    fam_fcce_ndx;
+#if defined(HAVE_SYS_INOTIFY_H)
+	iosocket *sock;    /* socket to the inotify fd (this should be in a backend struct */
 #endif
-	buffer *hash_key;  /* temp-store for the hash-key */
 } stat_cache;
 
 typedef struct {
@@ -255,6 +238,7 @@ typedef struct {
 	unsigned short max_keep_alive_idle;
 	unsigned short max_read_idle;
 	unsigned short max_write_idle;
+	unsigned short max_connection_idle;
 	unsigned short use_xattr;
 	unsigned short follow_symlink;
 	unsigned short range_requests;
@@ -266,34 +250,15 @@ typedef struct {
 	unsigned short log_request_handling;
 	unsigned short log_response_header;
 	unsigned short log_condition_handling;
-	unsigned short log_ssl_noise;
-	unsigned short log_timeouts;
+	unsigned short log_condition_cache_handling;
 
 
 	/* server wide */
 	buffer *ssl_pemfile;
 	buffer *ssl_ca_file;
-	buffer *ssl_cipher_list;
-	buffer *ssl_dh_file;
-	buffer *ssl_ec_curve;
-	unsigned short ssl_honor_cipher_order; /* determine SSL cipher in server-preferred order, not client-order */
-	unsigned short ssl_empty_fragments; /* whether to not set SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS */
-	unsigned short ssl_use_sslv2;
-	unsigned short ssl_use_sslv3;
-	unsigned short ssl_verifyclient;
-	unsigned short ssl_verifyclient_enforce;
-	unsigned short ssl_verifyclient_depth;
-	buffer *ssl_verifyclient_username;
-	unsigned short ssl_verifyclient_export_cert;
-	unsigned short ssl_disable_client_renegotiation;
-
-	unsigned short use_ipv6, set_v6only; /* set_v6only is only a temporary option */
-	unsigned short defer_accept;
-	unsigned short ssl_enabled; /* only interesting for setting up listening sockets. don't use at runtime */
+	unsigned short use_ipv6;
+	unsigned short is_ssl;
 	unsigned short allow_http11;
-	unsigned short etag_use_inode;
-	unsigned short etag_use_mtime;
-	unsigned short etag_use_size;
 	unsigned short force_lowercase_filenames; /* if the FS is case-insensitive, force all files to lower-case */
 	unsigned int max_request_size;
 
@@ -306,38 +271,35 @@ typedef struct {
 	/* server-wide traffic-shaper
 	 *
 	 * each context has the counter which is inited once
-	 * a second by the global_kbytes_per_second config-var
+	 * per second by the global_kbytes_per_second config-var
 	 *
 	 * as soon as global_kbytes_per_second gets below 0
 	 * the connected conns are "offline" a little bit
 	 *
 	 * the problem:
-	 * we somehow have to loose our "we are writable" signal
+	 * we somehow have to lose our "we are writable" signal
 	 * on the way.
 	 *
 	 */
 	off_t *global_bytes_per_second_cnt_ptr; /*  */
 
 #ifdef USE_OPENSSL
-	SSL_CTX *ssl_ctx; /* not patched */
-	/* SNI per host: with COMP_SERVER_SOCKET, COMP_HTTP_SCHEME, COMP_HTTP_HOST */
-	EVP_PKEY *ssl_pemfile_pkey;
-	X509 *ssl_pemfile_x509;
-	STACK_OF(X509_NAME) *ssl_ca_file_cert_names;
+	SSL_CTX *ssl_ctx;
 #endif
 } specific_config;
 
 /* the order of the items should be the same as they are processed
  * read before write as we use this later */
 typedef enum {
-	CON_STATE_CONNECT,
-	CON_STATE_REQUEST_START,
-	CON_STATE_READ,
-	CON_STATE_REQUEST_END,
-	CON_STATE_READ_POST,
-	CON_STATE_HANDLE_REQUEST,
-	CON_STATE_RESPONSE_START,
-	CON_STATE_WRITE,
+	CON_STATE_CONNECT,         /** we are wait for a connect */
+	CON_STATE_REQUEST_START,   /** after the connect, the request is initialized, keep-alive starts here again */
+	CON_STATE_READ_REQUEST_HEADER,   /** loop in the read-request-header until the full header is received */
+	CON_STATE_VALIDATE_REQUEST_HEADER,   /** validate the request-header */
+	CON_STATE_HANDLE_REQUEST_HEADER, /** find a handler for the request */
+	CON_STATE_READ_REQUEST_CONTENT,  /** forward the request content to the handler */
+	CON_STATE_HANDLE_RESPONSE_HEADER, /** the backend bounces the response back to the client */
+	CON_STATE_WRITE_RESPONSE_HEADER,
+	CON_STATE_WRITE_RESPONSE_CONTENT,
 	CON_STATE_RESPONSE_END,
 	CON_STATE_ERROR,
 	CON_STATE_CLOSE
@@ -349,8 +311,6 @@ typedef struct {
 	int patterncount;
 	int matches[3 * 10];
 	buffer *comp_value; /* just a pointer */
-	
-	comp_key_t comp_type;
 } cond_cache_t;
 
 typedef struct {
@@ -373,23 +333,23 @@ typedef struct {
 				      * this is self-protection
 				      */
 
-	int fd;                      /* the FD for this connection */
-	int fde_ndx;                 /* index for the fdevent-handler */
+	iosocket *sock;
 	int ndx;                     /* reverse mapping to server->connection[ndx] */
 
 	/* fd states */
 	int is_readable;
 	int is_writable;
 
-	int keep_alive;              /* only request.c can enable it, all other just disable */
-	int keep_alive_idle;         /* remember max_keep_alive_idle from config */
+	int     keep_alive;          /* only request.c can enable it, all others just disable */
 
 	int file_started;
-	int file_finished;
 
-	chunkqueue *write_queue;      /* a large queue for low-level write ( HTTP response ) [ file, mem ] */
-	chunkqueue *read_queue;       /* a small queue for low-level read ( HTTP request ) [ mem ] */
-	chunkqueue *request_content_queue; /* takes request-content into tempfile if necessary [ tempfile, mem ]*/
+	chunkqueue *send;            /* the response-content before filters are applied */
+	chunkqueue *recv;            /* the request-content, without encoding */
+
+	filter_chain *send_filters;  /* the chain of filters to apply to response-content. */
+	chunkqueue *send_raw;        /* the full response (HTTP-Header + compression + chunking ) */
+	chunkqueue *recv_raw;        /* the full request (HTTP-Header + chunking ) */
 
 	int traffic_limit_reached;
 
@@ -405,8 +365,8 @@ typedef struct {
 
 	/* request */
 	buffer *parse_request;
-	unsigned int parsed_response; /* bitfield which contains the important header-fields of the parsed response header */
 
+	http_req *http_req;
 	request  request;
 	request_uri uri;
 	physical physical;
@@ -414,6 +374,7 @@ typedef struct {
 
 	size_t header_len;
 
+	buffer *authed_user;
 	array  *environment; /* used to pass lighttpd internal stuff to the FastCGI/CGI apps, setenv does that */
 
 	/* response */
@@ -435,19 +396,7 @@ typedef struct {
 	int error_handler_saved_status;
 	int in_error_handler;
 
-	struct server_socket *srv_socket;   /* reference to the server-socket */
-
-#ifdef USE_OPENSSL
-	SSL *ssl;
-# ifndef OPENSSL_NO_TLSEXT
-	buffer *tlsext_server_name;
-# endif
-	unsigned int renegotiations; /* count of SSL_CB_HANDSHAKE_START */
-#endif
-	/* etag handling */
-	etag_flags_t etag_flags;
-
-	int conditional_is_valid[COMP_LAST_ELEMENT]; 
+	void *srv_socket;   /* reference to the server-socket (typecast to server_socket) */
 } connection;
 
 typedef struct {
@@ -487,15 +436,23 @@ typedef struct {
 	size_t size;
 } buffer_plugin;
 
+typedef enum {
+	NETWORK_STATUS_UNSET,
+	NETWORK_STATUS_SUCCESS,
+	NETWORK_STATUS_FATAL_ERROR,
+	NETWORK_STATUS_CONNECTION_CLOSE,
+	NETWORK_STATUS_WAIT_FOR_EVENT,
+	NETWORK_STATUS_WAIT_FOR_AIO_EVENT,
+	NETWORK_STATUS_WAIT_FOR_FD,
+	NETWORK_STATUS_INTERRUPTED
+} network_status_t;
+
 typedef struct {
 	unsigned short port;
 	buffer *bindhost;
 
-	buffer *errorlog_file;
-	unsigned short errorlog_use_syslog;
-	buffer *breakagelog_file;
-
 	unsigned short dont_daemonize;
+	unsigned short daemonize_on_shutdown;
 	buffer *changeroot;
 	buffer *username;
 	buffer *groupname;
@@ -509,6 +466,8 @@ typedef struct {
 	array *modules;
 	array *upload_tempdirs;
 
+	unsigned short use_noatime;
+
 	unsigned short max_worker;
 	unsigned short max_fds;
 	unsigned short max_conns;
@@ -519,20 +478,46 @@ typedef struct {
 
 	enum { STAT_CACHE_ENGINE_UNSET,
 			STAT_CACHE_ENGINE_NONE,
-			STAT_CACHE_ENGINE_SIMPLE
-#ifdef HAVE_FAM_H
-			, STAT_CACHE_ENGINE_FAM
-#endif
+			STAT_CACHE_ENGINE_SIMPLE,
+			STAT_CACHE_ENGINE_FAM,
+			STAT_CACHE_ENGINE_INOTIFY
 	} stat_cache_engine;
 	unsigned short enable_cores;
-	unsigned short reject_expect_100_with_417;
+
+	buffer *errorlog_file;
+	unsigned short errorlog_use_syslog;
+
+	unsigned short max_stat_threads;
+	unsigned short max_read_threads;
 } server_config;
 
-typedef struct server_socket {
-	sock_addr addr;
-	int       fd;
-	int       fde_ndx;
+typedef enum {
+	NETWORK_BACKEND_UNSET,
 
+	NETWORK_BACKEND_WRITE,
+	NETWORK_BACKEND_WRITEV,
+
+	NETWORK_BACKEND_LINUX_SENDFILE,
+	NETWORK_BACKEND_LINUX_AIO_SENDFILE,
+	NETWORK_BACKEND_POSIX_AIO,
+	NETWORK_BACKEND_GTHREAD_AIO,
+
+	NETWORK_BACKEND_FREEBSD_SENDFILE,
+	NETWORK_BACKEND_SOLARIS_SENDFILEV,
+
+	NETWORK_BACKEND_WIN32_SEND,
+	NETWORK_BACKEND_WIN32_TRANSMITFILE,
+
+} network_backend_t;
+
+
+typedef struct {
+	sock_addr addr;
+	iosocket *sock;
+
+	buffer *ssl_pemfile;
+	buffer *ssl_ca_file;
+	unsigned short use_ipv6;
 	unsigned short is_ssl;
 
 	buffer *srv_token;
@@ -552,11 +537,6 @@ typedef struct {
 typedef struct server {
 	server_socket_array srv_sockets;
 
-	/* the errorlog */
-	int errorlog_fd;
-	enum { ERRORLOG_FILE, ERRORLOG_FD, ERRORLOG_SYSLOG, ERRORLOG_PIPE } errorlog_mode;
-	buffer *errorlog_buf;
-
 	fdevents *ev, *ev_ins;
 
 	buffer_plugin plugins;
@@ -571,8 +551,6 @@ typedef struct server {
 	int ssl_is_init;
 
 	int max_fds;    /* max possible fds */
-	int cur_fds;    /* currently used fds */
-	int want_fds;   /* waiting fds */
 	int sockets_disabled;
 
 	size_t max_conns;
@@ -603,9 +581,6 @@ typedef struct server {
 	time_t last_generated_debug_ts;
 	time_t startup_ts;
 
-	char entropy[8]; /* from /dev/[u]random if possible, otherwise rand() */
-	char is_real_entropy; /* whether entropy is from /dev/[u]random */
-
 	buffer *ts_debug_str;
 	buffer *ts_date_str;
 
@@ -618,40 +593,48 @@ typedef struct server {
 
 	server_config  srvconf;
 
-	short int config_deprecated;
-	short int config_unsupported;
+	short unsigned config_deprecated;
+	short unsigned config_unsupported;
 
 	connections *conns;
 	connections *joblist;
+	connections *joblist_prev;
 	connections *fdwaitqueue;
 
 	stat_cache  *stat_cache;
 
-	/**
-	 * The status array can carry all the status information you want
-	 * the key to the array is <module-prefix>.<name>
-	 * and the values are counters
-	 *
-	 * example:
-	 *   fastcgi.backends        = 10
-	 *   fastcgi.active-backends = 6
-	 *   fastcgi.backend.<key>.load = 24
-	 *   fastcgi.backend.<key>....
-	 *
-	 *   fastcgi.backend.<key>.disconnects = ...
-	 */
-	array *status;
-
 	fdevent_handler_t event_handler;
 
-	int (* network_backend_write)(struct server *srv, connection *con, int fd, chunkqueue *cq, off_t max_bytes);
+	network_status_t (* network_backend_write)(struct server *srv, connection *con, iosocket *sock, chunkqueue *cq);
+	network_status_t (* network_backend_read)(struct server *srv, connection *con, iosocket *sock, chunkqueue *cq);
 #ifdef USE_OPENSSL
-	int (* network_ssl_backend_write)(struct server *srv, connection *con, SSL *ssl, chunkqueue *cq, off_t max_bytes);
+	network_status_t (* network_ssl_backend_write)(struct server *srv, connection *con, iosocket *sock, chunkqueue *cq);
+	network_status_t (* network_ssl_backend_read)(struct server *srv, connection *con, iosocket *sock, chunkqueue *cq);
 #endif
 
+#ifdef HAVE_PWD_H
 	uid_t uid;
 	gid_t gid;
-} server;
+#endif
+#ifdef USE_GTHREAD
+#ifdef USE_LINUX_AIO_SENDFILE
+	io_context_t linux_io_ctx;
 
+	struct iocb *linux_io_iocbs;
+
+#endif
+#ifdef USE_POSIX_AIO
+	struct aiocb *posix_aio_iocbs;
+#endif
+
+	GAsyncQueue *stat_queue; /* send a stat_job into this queue and joblist_queue will get a wakeup when the stat is finished */
+	GAsyncQueue *joblist_queue;
+	GAsyncQueue *aio_write_queue;
+#endif
+	network_backend_t network_backend;
+	int is_shutdown;
+} server;
+
+int server_out_of_fds(server *srv, connection *con);
 
 #endif
