@@ -1,5 +1,3 @@
-#include "buffer.h"
-
 #include <stdlib.h>
 #include <string.h>
 
@@ -7,11 +5,8 @@
 #include <assert.h>
 #include <ctype.h>
 
-#if defined HAVE_STDINT_H
-# include <stdint.h>
-#elif defined HAVE_INTTYPES_H
-# include <inttypes.h>
-#endif
+#include "buffer.h"
+
 
 static const char hex_chars[] = "0123456789abcdef";
 
@@ -25,7 +20,7 @@ buffer* buffer_init(void) {
 	buffer *b;
 
 	b = malloc(sizeof(*b));
-	force_assert(b);
+	assert(b);
 
 	b->ptr = NULL;
 	b->size = 0;
@@ -60,8 +55,6 @@ void buffer_reset(buffer *b) {
 		free(b->ptr);
 		b->ptr = NULL;
 		b->size = 0;
-	} else if (b->size) {
-		b->ptr[0] = '\0';
 	}
 
 	b->used = 0;
@@ -70,7 +63,7 @@ void buffer_reset(buffer *b) {
 
 /**
  *
- * allocate (if neccessary) enough space for 'size' bytes and
+ * allocate (if necessary) enough space for 'size' bytes and
  * set the 'used' counter to 0
  *
  */
@@ -86,11 +79,11 @@ int buffer_prepare_copy(buffer *b, size_t size) {
 
 		b->size = size;
 
-		/* always allocate a multiply of BUFFER_PIECE_SIZE */
+		/* always allocate a multiple of BUFFER_PIECE_SIZE */
 		b->size += BUFFER_PIECE_SIZE - (b->size % BUFFER_PIECE_SIZE);
 
 		b->ptr = malloc(b->size);
-		force_assert(b->ptr);
+		assert(b->ptr);
 	}
 	b->used = 0;
 	return 0;
@@ -98,7 +91,7 @@ int buffer_prepare_copy(buffer *b, size_t size) {
 
 /**
  *
- * increase the internal buffer (if neccessary) to append another 'size' byte
+ * increase the internal buffer (if necessary) to append another 'size' byte
  * ->used isn't changed
  *
  */
@@ -109,20 +102,20 @@ int buffer_prepare_append(buffer *b, size_t size) {
 	if (0 == b->size) {
 		b->size = size;
 
-		/* always allocate a multiply of BUFFER_PIECE_SIZE */
+		/* always allocate a multiple of BUFFER_PIECE_SIZE */
 		b->size += BUFFER_PIECE_SIZE - (b->size % BUFFER_PIECE_SIZE);
 
 		b->ptr = malloc(b->size);
 		b->used = 0;
-		force_assert(b->ptr);
+		assert(b->ptr);
 	} else if (b->used + size > b->size) {
 		b->size += size;
 
-		/* always allocate a multiply of BUFFER_PIECE_SIZE */
+		/* always allocate a multiple of BUFFER_PIECE_SIZE */
 		b->size += BUFFER_PIECE_SIZE - (b->size % BUFFER_PIECE_SIZE);
 
 		b->ptr = realloc(b->ptr, b->size);
-		force_assert(b->ptr);
+		assert(b->ptr);
 	}
 	return 0;
 }
@@ -164,7 +157,7 @@ int buffer_copy_string_buffer(buffer *b, const buffer *src) {
 	if (!src) return -1;
 
 	if (src->used == 0) {
-		buffer_reset(b);
+		b->used = 0;
 		return 0;
 	}
 	return buffer_copy_string_len(b, src->ptr, src->used - 1);
@@ -192,7 +185,6 @@ int buffer_append_string_rfill(buffer *b, const char *s, size_t maxlen) {
 	if (!s || !b) return -1;
 
 	s_len = strlen(s);
-	if (s_len > maxlen)  s_len = maxlen;
 	buffer_prepare_append(b, maxlen + 1);
 	if (b->used == 0)
 		b->used++;
@@ -211,7 +203,7 @@ int buffer_append_string_rfill(buffer *b, const char *s, size_t maxlen) {
  * append a string to the end of the buffer
  *
  * the resulting buffer is terminated with a '\0'
- * s is treated as a un-terminated string (a \0 is handled a normal character)
+ * s is treated as an un-terminated string (a \0 is handled as a normal character)
  *
  * @param b a buffer
  * @param s the string
@@ -289,7 +281,7 @@ int buffer_append_long_hex(buffer *b, unsigned long value) {
 	return 0;
 }
 
-int LI_ltostr(char *buf, long val) {
+int ltostr(char *buf, long val) {
 	char swap;
 	char *end;
 	int len = 1;
@@ -328,7 +320,7 @@ int buffer_append_long(buffer *b, long val) {
 	if (b->used == 0)
 		b->used++;
 
-	b->used += LI_ltostr(b->ptr + (b->used - 1), val);
+	b->used += ltostr(b->ptr + (b->used - 1), val);
 	return 0;
 }
 
@@ -409,6 +401,75 @@ char hex2int(unsigned char hex) {
 }
 
 
+/**
+ * init the ptr buffer
+ *
+ */
+buffer_ptr *buffer_ptr_init(buffer_ptr_free_t freer)
+{
+	buffer_ptr *l = calloc(1, sizeof(buffer_ptr));
+	l->free = freer;
+
+	return l;
+}
+
+/**
+ * free the buffer_array
+ *
+ */
+void buffer_ptr_free(buffer_ptr *l)
+{
+	if (NULL != l) {
+		buffer_ptr_clear(l);
+		free(l);
+	}
+}
+
+void buffer_ptr_clear(buffer_ptr *l)
+{
+	assert(NULL != l);
+
+	if (l->free && l->used) {
+		size_t i;
+		for (i = 0; i < l->used; i ++) {
+			l->free(l->ptr[i]);
+		}
+	}
+
+	if (l->ptr) {
+		free(l->ptr);
+		l->ptr = NULL;
+	}
+	l->used = 0;
+	l->size = 0;
+}
+
+void buffer_ptr_append(buffer_ptr* l, void *item)
+{
+	assert(NULL != l);
+	if (l->ptr == NULL) {
+		l->size = 16;
+		l->ptr = (void **)malloc(sizeof(void *) * l->size);
+	}
+	else if (l->used == l->size) {
+		l->size += 16;
+		l->ptr = realloc(l->ptr, sizeof(void *) * l->size);
+	}
+	l->ptr[l->used++] = item;
+}
+
+void *buffer_ptr_pop(buffer_ptr* l)
+{
+	assert(NULL != l && l->used > 0);
+	return l->ptr[--l->used];
+}
+
+void *buffer_ptr_top(buffer_ptr* l)
+{
+	assert(NULL != l && l->used > 0);
+	return l->ptr[l->used-1];
+}
+
 /**
  * init the buffer
  *
@@ -419,7 +480,7 @@ buffer_array* buffer_array_init(void) {
 
 	b = malloc(sizeof(*b));
 
-	force_assert(b);
+	assert(b);
 	b->ptr = NULL;
 	b->size = 0;
 	b->used = 0;
@@ -463,14 +524,14 @@ buffer *buffer_array_append_get_buffer(buffer_array *b) {
 	if (b->size == 0) {
 		b->size = 16;
 		b->ptr = malloc(sizeof(*b->ptr) * b->size);
-		force_assert(b->ptr);
+		assert(b->ptr);
 		for (i = 0; i < b->size; i++) {
 			b->ptr[i] = NULL;
 		}
 	} else if (b->size == b->used) {
 		b->size += 16;
 		b->ptr = realloc(b->ptr, sizeof(*b->ptr) * b->size);
-		force_assert(b->ptr);
+		assert(b->ptr);
 		for (i = b->used; i < b->size; i++) {
 			b->ptr[i] = NULL;
 		}
@@ -511,12 +572,11 @@ buffer *buffer_init_string(const char *str) {
 }
 
 int buffer_is_empty(buffer *b) {
-	if (!b) return 1;
 	return (b->used == 0);
 }
 
 /**
- * check if two buffer contain the same data
+ * check if two buffers contain the same data
  *
  * HISTORY: this function was pretty much optimized, but didn't handled
  * alignment properly.
@@ -526,7 +586,7 @@ int buffer_is_equal(buffer *a, buffer *b) {
 	if (a->used != b->used) return 0;
 	if (a->used == 0) return 1;
 
-	return (0 == strcmp(a->ptr, b->ptr));
+	return (0 == strncmp(a->ptr, b->ptr, a->used - 1));
 }
 
 int buffer_is_equal_string(buffer *a, const char *s, size_t b_len) {
@@ -538,38 +598,55 @@ int buffer_is_equal_string(buffer *a, const char *s, size_t b_len) {
 	return buffer_is_equal(a, &b);
 }
 
-/* buffer_is_equal_caseless_string(b, CONST_STR_LEN("value")) */
-int buffer_is_equal_caseless_string(buffer *a, const char *s, size_t b_len) {
-	if (a->used != b_len + 1) return 0;
-
-	return (0 == strcasecmp(a->ptr, s));
-}
-
+/* simple-assumption:
+ *
+ * most parts are equal and doing a case conversion takes time
+ *
+ */
 int buffer_caseless_compare(const char *a, size_t a_len, const char *b, size_t b_len) {
-	size_t const len = (a_len < b_len) ? a_len : b_len;
-	size_t i;
+	size_t ndx = 0, max_ndx;
+	size_t *al, *bl;
+	size_t mask = sizeof(*al) - 1;
 
-	for (i = 0; i < len; ++i) {
-		unsigned char ca = a[i], cb = b[i];
-		if (ca == cb) continue;
+	al = (size_t *)a;
+	bl = (size_t *)b;
 
-		/* always lowercase for transitive results */
-#if 1
-		if (ca >= 'A' && ca <= 'Z') ca |= 32;
-		if (cb >= 'A' && cb <= 'Z') cb |= 32;
-#else
-		/* try to produce code without branching (jumps) */
-		ca |= ((unsigned char)(ca - (unsigned char)'A') <= (unsigned char)('Z' - 'A')) ? 32 : 0;
-		cb |= ((unsigned char)(cb - (unsigned char)'A') <= (unsigned char)('Z' - 'A')) ? 32 : 0;
-#endif
+	/* is the alignment correct? */
+	if ( ((size_t)al & mask) == 0 &&
+	     ((size_t)bl & mask) == 0 ) {
+
+		max_ndx = ((a_len < b_len) ? a_len : b_len) & ~mask;
+
+		for (; ndx < max_ndx; ndx += sizeof(*al)) {
+			if (*al != *bl) break;
+			al++; bl++;
+
+		}
 
-		if (ca == cb) continue;
-		return ca - cb;
 	}
-	if (a_len == b_len) return 0;
-	return a_len - b_len;
+
+	a = (char *)al;
+	b = (char *)bl;
+
+	max_ndx = ((a_len < b_len) ? a_len : b_len);
+
+	for (; ndx < max_ndx; ndx++) {
+		char a1 = *a++, b1 = *b++;
+
+		if (a1 != b1) {
+			if ((a1 >= 'A' && a1 <= 'Z') && (b1 >= 'a' && b1 <= 'z'))
+				a1 |= 32;
+			else if ((a1 >= 'a' && a1 <= 'z') && (b1 >= 'A' && b1 <= 'Z'))
+				b1 |= 32;
+			if ((a1 - b1) != 0) return (a1 - b1);
+
+		}
+	}
+
+	return 0;
 }
 
+
 /**
  * check if the rightmost bytes of the string are equal.
  *
@@ -611,19 +688,18 @@ int buffer_copy_string_hex(buffer *b, const char *in, size_t in_len) {
 	return 0;
 }
 
-/* everything except: ! ( ) * - . 0-9 A-Z _ a-z */
-static const char encoded_chars_rel_uri_part[] = {
+const char encoded_chars_rel_uri_part[] = {
 	/*
 	0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
 	*/
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  00 -  0F control chars */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  10 -  1F */
 	1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 1,  /*  20 -  2F space " # $ % & ' + , / */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,  /*  30 -  3F : ; < = > ? */
-	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  40 -  4F @ */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0,  /*  50 -  5F [ \ ] ^ */
-	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  60 -  6F ` */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,  /*  70 -  7F { | } ~ DEL */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,  /*  30 -  3F : ; = ? @ < > */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  40 -  4F */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  50 -  5F */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  60 -  6F */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,  /*  70 -  7F DEL */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  80 -  8F */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  90 -  9F */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  A0 -  AF */
@@ -634,19 +710,18 @@ static const char encoded_chars_rel_uri_part[] = {
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  F0 -  FF */
 };
 
-/* everything except: ! ( ) * - . / 0-9 A-Z _ a-z */
-static const char encoded_chars_rel_uri[] = {
+const char encoded_chars_rel_uri[] = {
 	/*
 	0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
 	*/
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  00 -  0F control chars */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  10 -  1F */
-	1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0,  /*  20 -  2F space " # $ % & ' + , */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,  /*  30 -  3F : ; < = > ? */
-	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  40 -  4F @ */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0,  /*  50 -  5F [ \ ] ^ */
-	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  60 -  6F ` */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,  /*  70 -  7F { | } ~ DEL */
+	1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0,  /*  20 -  2F space " # $ % & ' + , / */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,  /*  30 -  3F : ; = ? @ < > */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  40 -  4F */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  50 -  5F */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  60 -  6F */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,  /*  70 -  7F DEL */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  80 -  8F */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  90 -  9F */
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  A0 -  AF */
@@ -657,7 +732,7 @@ static const char encoded_chars_rel_uri[] = {
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  F0 -  FF */
 };
 
-static const char encoded_chars_html[] = {
+const char encoded_chars_html[] = {
 	/*
 	0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
 	*/
@@ -679,7 +754,7 @@ static const char encoded_chars_html[] = {
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  F0 -  FF */
 };
 
-static const char encoded_chars_minimal_xml[] = {
+const char encoded_chars_minimal_xml[] = {
 	/*
 	0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
 	*/
@@ -701,7 +776,7 @@ static const char encoded_chars_minimal_xml[] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  F0 -  FF */
 };
 
-static const char encoded_chars_hex[] = {
+const char encoded_chars_hex[] = {
 	/*
 	0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
 	*/
@@ -723,29 +798,6 @@ static const char encoded_chars_hex[] = {
 	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /*  F0 -  FF */
 };
 
-static const char encoded_chars_http_header[] = {
-	/*
-	0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
-	*/
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,  /*  00 -  0F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  10 -  1F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  20 -  2F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  30 -  3F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  40 -  4F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  50 -  5F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  60 -  6F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  70 -  7F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  80 -  8F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  90 -  9F */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  A0 -  AF */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  B0 -  BF */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  C0 -  CF */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  D0 -  DF */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  E0 -  EF */
-	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /*  F0 -  FF */
-};
-
-
 
 int buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_encoding_t encoding) {
 	unsigned char *ds, *d;
@@ -753,10 +805,9 @@ int buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_
 	const char *map = NULL;
 
 	if (!s || !b) return -1;
+	if (b->used == 0) return -1;
 
-	if (b->ptr[b->used - 1] != '\0') {
-		SEGFAULT();
-	}
+	if (b->ptr[b->used - 1] != '\0') return -1;
 
 	if (s_len == 0) return 0;
 
@@ -776,16 +827,13 @@ int buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_
 	case ENCODING_HEX:
 		map = encoded_chars_hex;
 		break;
-	case ENCODING_HTTP_HEADER:
-		map = encoded_chars_http_header;
-		break;
 	case ENCODING_UNSET:
-		break;
+		return buffer_append_string_len(b, s, s_len);
 	}
 
-	force_assert(map != NULL);
+	assert(map != NULL);
 
-	/* count to-be-encoded-characters */
+	/* count to-be-encoded characters */
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
 		if (map[*ds]) {
 			switch(encoding) {
@@ -797,7 +845,6 @@ int buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_
 			case ENCODING_MINIMAL_XML:
 				d_len += 6;
 				break;
-			case ENCODING_HTTP_HEADER:
 			case ENCODING_HEX:
 				d_len += 2;
 				break;
@@ -833,10 +880,6 @@ int buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_
 				d[d_len++] = hex_chars[((*ds) >> 4) & 0x0F];
 				d[d_len++] = hex_chars[(*ds) & 0x0F];
 				break;
-			case ENCODING_HTTP_HEADER:
-				d[d_len++] = *ds;
-				d[d_len++] = '\t';
-				break;
 			case ENCODING_UNSET:
 				break;
 			}
@@ -854,7 +897,7 @@ int buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer_
 }
 
 
-/* decodes url-special-chars inplace.
+/* decodes url-special chars in-place.
  * replaces non-printable characters with '_'
  */
 
@@ -880,7 +923,7 @@ static int buffer_urldecode_internal(buffer *url, int is_query) {
 				if (low != 0xFF) {
 					high = (high << 4) | low;
 
-					/* map control-characters out */
+					/* map out control characters */
 					if (high < 32 || high == 127) high = '_';
 
 					*dst = high;
@@ -916,7 +959,7 @@ int buffer_urldecode_query(buffer *url) {
  * /abc/./xyz       gets  /abc/xyz
  * /abc//xyz        gets  /abc/xyz
  *
- * NOTE: src and dest can point to the same buffer, in which case,
+ * NOTE: src and dest can point to the same buffer, in which case
  *       the operation is performed in-place.
  */
 
@@ -939,18 +982,6 @@ int buffer_path_simplify(buffer *dest, buffer *src)
 	start = dest->ptr;
 	out   = dest->ptr;
 	slash = dest->ptr;
-
-
-#if defined(__WIN32) || defined(__CYGWIN__)
-	/* cygwin is treating \ and / the same, so we have to that too
-	 */
-
-	for (walk = src->ptr; *walk; walk++) {
-		if (*walk == '\\') *walk = '/';
-	}
-	walk = src->ptr;
-#endif
-
 	while (*walk == ' ') {
 		walk++;
 	}
@@ -1030,6 +1061,31 @@ int light_isalnum(int c) {
 	return light_isdigit(c) || light_isalpha(c);
 }
 
+#undef BUFFER_CTYPE_FUNC
+#define BUFFER_CTYPE_FUNC(type) \
+	int buffer_is##type(buffer *b) { \
+		size_t i, len; \
+		if (b->used < 2) return 0; \
+		/* strlen */ \
+		len = b->used - 1; \
+		/* c-string only */ \
+		if (b->ptr[len] != '\0') { \
+			return 0; \
+		} \
+		/* check on the whole string */ \
+		for (i = 0; i < len; i ++) { \
+			if (!light_is##type(b->ptr[i])) { \
+				return 0; \
+			} \
+		} \
+		return 1; \
+	}
+
+BUFFER_CTYPE_FUNC(digit)
+BUFFER_CTYPE_FUNC(xdigit)
+BUFFER_CTYPE_FUNC(alpha)
+BUFFER_CTYPE_FUNC(alnum)
+
 int buffer_to_lower(buffer *b) {
 	char *c;
 
@@ -1059,9 +1115,42 @@ int buffer_to_upper(buffer *b) {
 	return 0;
 }
 
-void log_failed_assert(const char *filename, unsigned int line, const char *msg) {
-	/* can't use buffer here; could lead to recursive assertions */
-	fprintf(stderr, "%s.%d: %s\n", filename, line, msg);
-	fflush(stderr);
-	abort();
+buffer_pool *buffer_pool_init() {
+	buffer_pool *bp;
+
+	bp = calloc(1, sizeof(*bp));
+
+	return bp;
 }
+
+void buffer_pool_free(buffer_pool *bp) {
+	if (!bp) return;
+
+	ARRAY_STATIC_FREE(bp, buffer, b, buffer_free(b));
+
+	free(bp);
+
+	return;
+}
+
+buffer *buffer_pool_get(buffer_pool *bp) {
+	buffer *b;
+
+	if (bp->used == 0) {
+		return buffer_init();
+	}
+
+	b = bp->ptr[--bp->used];
+
+	buffer_reset(b);
+
+	return b;
+}
+
+void buffer_pool_append(buffer_pool *bp, buffer *b) {
+	ARRAY_STATIC_PREPARE_APPEND(bp);
+
+	bp->ptr[bp->used++] = b;
+}
+
+
