@@ -6,6 +6,7 @@
 #endif
 
 #include <sys/types.h>
+#include <stdlib.h>
 
 #if defined HAVE_STDINT_H
 #include <stdint.h>
@@ -13,6 +14,12 @@
 #include <inttypes.h>
 #endif
 
-uint32_t generate_crc32c(char *string, size_t length);
+#ifdef _WIN32
+#define uint32_t unsigned __int32
+#endif
+
+#include "settings.h"
+
+LI_API uint32_t generate_crc32c(char *string, size_t length);
 
 #endif
