@@ -1,19 +1,14 @@
-#include "server.h"
-#include "log.h"
-#include "http_auth.h"
-#include "inet_ntop_cache.h"
-#include "stream.h"
-#include "base64.h"
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 #ifdef HAVE_CRYPT_H
 # include <crypt.h>
 #elif defined(__linux__)
 /* linux needs _XOPEN_SOURCE */
-# define _XOPEN_SOURCE
 #endif
 
-#if defined(HAVE_LIBCRYPT) && !defined(HAVE_CRYPT)
-/* always assume crypt() is present if we have -lcrypt */
+#ifdef HAVE_LIBCRYPT
 # define HAVE_CRYPT
 #endif
 
@@ -26,26 +21,23 @@
 #include <string.h>
 #include <time.h>
 #include <errno.h>
-#include <unistd.h>
 #include <ctype.h>
 
-#include "md5.h"
+#include "server.h"
+#include "log.h"
+#include "http_auth.h"
+#include "http_auth_digest.h"
+#include "stream.h"
+
+#include "sys-strings.h"
+#include "sys-files.h"
 
 #ifdef USE_OPENSSL
-#include <openssl/sha.h>
+# include <openssl/md5.h>
+#else
+# include "md5.h"
 #endif
 
-#include "safe_memclear.h"
-
-#define HASHLEN 16
-#define HASHHEXLEN 32
-typedef unsigned char HASH[HASHLEN];
-typedef char HASHHEX[HASHHEXLEN+1];
-
-static void CvtHex(const HASH Bin, char Hex[33]) {
-	li_tohex(Hex, (const char*) Bin, 16);
-}
-
 /**
  * the $apr1$ handling is taken from apache 1.3.x
  */
@@ -63,16 +55,97 @@ static void CvtHex(const HASH Bin, char Hex[33]) {
 
 handler_t auth_ldap_init(server *srv, mod_auth_plugin_config *s);
 
+static const char base64_pad = '=';
+
+static const short base64_reverse_table[256] = {
+	        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
+		        52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
+		        -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
+		        15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
+		        -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
+		        41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+		        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
+};
+
+
+static unsigned char * base64_decode(buffer *out, const char *in) {
+	unsigned char *result;
+	int ch, j = 0, k;
+	size_t i;
+
+	size_t in_len = strlen(in);
+
+	buffer_prepare_copy(out, in_len);
+
+	result = (unsigned char *)out->ptr;
+
+	ch = in[0];
+	/* run through the whole string, converting as we go */
+	for (i = 0; i < in_len; i++) {
+		ch = in[i];
+
+		if (ch == '\0') break;
+
+		if (ch == base64_pad) break;
+
+		ch = base64_reverse_table[ch];
+		if (ch < 0) continue;
+
+		switch(i % 4) {
+		case 0:
+			result[j] = ch << 2;
+			break;
+		case 1:
+			result[j++] |= ch >> 4;
+			result[j] = (ch & 0x0f) << 4;
+			break;
+		case 2:
+			result[j++] |= ch >>2;
+			result[j] = (ch & 0x03) << 6;
+			break;
+		case 3:
+			result[j++] |= ch;
+			break;
+		}
+	}
+	k = j;
+	/* mop things up if we ended on a boundary */
+	if (ch == base64_pad) {
+		switch(i % 4) {
+		case 0:
+		case 1:
+			return NULL;
+		case 2:
+			k++;
+		case 3:
+			result[k++] = 0;
+		}
+	}
+	result[k] = '\0';
+
+	out->used = k;
+
+	return result;
+}
+
 static int http_auth_get_password(server *srv, mod_auth_plugin_data *p, buffer *username, buffer *realm, buffer *password) {
 	int ret = -1;
 
-	if (buffer_is_empty(username) || buffer_is_empty(realm)) return -1;
+	if (!username->used|| !realm->used) return -1;
 
 	if (p->conf.auth_backend == AUTH_BACKEND_HTDIGEST) {
 		stream f;
 		char * f_line;
 
-		if (buffer_string_is_empty(p->conf.auth_htdigest_userfile)) return -1;
+		if (buffer_is_empty(p->conf.auth_htdigest_userfile)) return -1;
 
 		if (0 != stream_open(&f, p->conf.auth_htdigest_userfile)) {
 			log_error_write(srv, __FILE__, __LINE__, "sbss", "opening digest-userfile", p->conf.auth_htdigest_userfile, "failed:", strerror(errno));
@@ -126,8 +199,8 @@ static int http_auth_get_password(server *srv, mod_auth_plugin_data *p, buffer *
 				pwd_len = f.size - (f_pwd - f.start);
 			}
 
-			if (buffer_string_length(username) == u_len &&
-			    (buffer_string_length(realm) == r_len) &&
+			if (username->used - 1 == u_len &&
+			    (realm->used - 1 == r_len) &&
 			    (0 == strncmp(username->ptr, f_user, u_len)) &&
 			    (0 == strncmp(realm->ptr, f_realm, r_len))) {
 				/* found */
@@ -153,7 +226,7 @@ static int http_auth_get_password(server *srv, mod_auth_plugin_data *p, buffer *
 
 		auth_fn = (p->conf.auth_backend == AUTH_BACKEND_HTPASSWD) ? p->conf.auth_htpasswd_userfile : p->conf.auth_plain_userfile;
 
-		if (buffer_string_is_empty(auth_fn)) return -1;
+		if (buffer_is_empty(auth_fn)) return -1;
 
 		if (0 != stream_open(&f, auth_fn)) {
 			log_error_write(srv, __FILE__, __LINE__, "sbss",
@@ -196,7 +269,7 @@ static int http_auth_get_password(server *srv, mod_auth_plugin_data *p, buffer *
 				pwd_len = f.size - (f_pwd - f.start);
 			}
 
-			if (buffer_string_length(username) == u_len &&
+			if (username->used - 1 == u_len &&
 			    (0 == strncmp(username->ptr, f_user, u_len))) {
 				/* found */
 
@@ -222,18 +295,36 @@ static int http_auth_get_password(server *srv, mod_auth_plugin_data *p, buffer *
 	return ret;
 }
 
-int http_auth_match_rules(server *srv, array *req, const char *username, const char *group, const char *host) {
+static int http_auth_match_rules(server *srv, mod_auth_plugin_data *p, const char *url, const char *username, const char *group, const char *host) {
 	const char *r = NULL, *rules = NULL;
+	size_t i;
 	int username_len;
 	data_string *require;
+	array *req;
 
 	UNUSED(group);
 	UNUSED(host);
 
-	require = (data_string *)array_get_element(req, "require");
+	/* check what has to be match to fullfil the request */
+	/* search auth-directives for path */
+	for (i = 0; i < p->conf.auth_require->used; i++) {
+		if (p->conf.auth_require->data[i]->key->used == 0) continue;
+
+		if (0 == strncmp(url, p->conf.auth_require->data[i]->key->ptr, p->conf.auth_require->data[i]->key->used - 1)) {
+			break;
+		}
+	}
+
+	if (i == p->conf.auth_require->used) {
+		return -1;
+	}
+
+	req = ((data_array *)(p->conf.auth_require->data[i]))->value;
+
+	require = (data_string *)array_get_element(req, CONST_STR_LEN("require"));
 
 	/* if we get here, the user we got a authed user */
-	if (0 == strcmp(require->value->ptr, "valid-user")) {
+	if (buffer_is_equal_string(require->value, CONST_STR_LEN("valid-user"))) {
 		return 0;
 	}
 
@@ -339,185 +430,166 @@ int http_auth_match_rules(server *srv, array *req, const char *username, const c
 
 static void to64(char *s, unsigned long v, int n)
 {
-	static const unsigned char itoa64[] =         /* 0 ... 63 => ASCII - 64 */
-		"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
+    static unsigned char itoa64[] =         /* 0 ... 63 => ASCII - 64 */
+        "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
 
-	while (--n >= 0) {
-		*s++ = itoa64[v&0x3f];
-		v >>= 6;
-	}
+    while (--n >= 0) {
+        *s++ = itoa64[v&0x3f];
+        v >>= 6;
+    }
 }
 
 static void apr_md5_encode(const char *pw, const char *salt, char *result, size_t nbytes) {
-	/*
-	 * Minimum size is 8 bytes for salt, plus 1 for the trailing NUL,
-	 * plus 4 for the '$' separators, plus the password hash itself.
-	 * Let's leave a goodly amount of leeway.
-	 */
+    /*
+     * Minimum size is 8 bytes for salt, plus 1 for the trailing NUL,
+     * plus 4 for the '$' separators, plus the password hash itself.
+     * Let's leave a goodly amount of leeway.
+     */
 
-	char passwd[120], *p;
-	const char *sp, *ep;
-	unsigned char final[APR_MD5_DIGESTSIZE];
-	ssize_t sl, pl, i;
-	li_MD5_CTX ctx, ctx1;
-	unsigned long l;
+    char passwd[120], *p;
+    const char *sp, *ep;
+    unsigned char final[APR_MD5_DIGESTSIZE];
+    ssize_t sl, pl, i;
+    MD5_CTX ctx, ctx1;
+    unsigned long l;
 
-	/*
-	 * Refine the salt first.  It's possible we were given an already-hashed
-	 * string as the salt argument, so extract the actual salt value from it
-	 * if so.  Otherwise just use the string up to the first '$' as the salt.
-	 */
-	sp = salt;
+    /*
+     * Refine the salt first.  It's possible we were given an already-hashed
+     * string as the salt argument, so extract the actual salt value from it
+     * if so.  Otherwise just use the string up to the first '$' as the salt.
+     */
+    sp = salt;
 
-	/*
-	 * If it starts with the magic string, then skip that.
-	 */
-	if (!strncmp(sp, APR1_ID, strlen(APR1_ID))) {
-		sp += strlen(APR1_ID);
-	}
+    /*
+     * If it starts with the magic string, then skip that.
+     */
+    if (!strncmp(sp, APR1_ID, strlen(APR1_ID))) {
+        sp += strlen(APR1_ID);
+    }
 
-	/*
-	 * It stops at the first '$' or 8 chars, whichever comes first
-	 */
-	for (ep = sp; (*ep != '\0') && (*ep != '$') && (ep < (sp + 8)); ep++) {
-		continue;
-	}
+    /*
+     * It stops at the first '$' or 8 chars, whichever comes first
+     */
+    for (ep = sp; (*ep != '\0') && (*ep != '$') && (ep < (sp + 8)); ep++) {
+        continue;
+    }
 
-	/*
-	 * Get the length of the true salt
-	 */
-	sl = ep - sp;
+    /*
+     * Get the length of the true salt
+     */
+    sl = ep - sp;
 
-	/*
-	 * 'Time to make the doughnuts..'
-	 */
-	li_MD5_Init(&ctx);
+    /*
+     * 'Time to make the doughnuts..'
+     */
+    MD5_Init(&ctx);
 
-	/*
-	 * The password first, since that is what is most unknown
-	 */
-	li_MD5_Update(&ctx, pw, strlen(pw));
+    /*
+     * The password first, since that is what is most unknown
+     */
+    MD5_Update(&ctx, pw, strlen(pw));
 
-	/*
-	 * Then our magic string
-	 */
-	li_MD5_Update(&ctx, APR1_ID, strlen(APR1_ID));
+    /*
+     * Then our magic string
+     */
+    MD5_Update(&ctx, APR1_ID, strlen(APR1_ID));
 
-	/*
-	 * Then the raw salt
-	 */
-	li_MD5_Update(&ctx, sp, sl);
+    /*
+     * Then the raw salt
+     */
+    MD5_Update(&ctx, sp, sl);
 
-	/*
-	 * Then just as many characters of the MD5(pw, salt, pw)
-	 */
-	li_MD5_Init(&ctx1);
-	li_MD5_Update(&ctx1, pw, strlen(pw));
-	li_MD5_Update(&ctx1, sp, sl);
-	li_MD5_Update(&ctx1, pw, strlen(pw));
-	li_MD5_Final(final, &ctx1);
-	for (pl = strlen(pw); pl > 0; pl -= APR_MD5_DIGESTSIZE) {
-		li_MD5_Update(
-			&ctx, final,
-			(pl > APR_MD5_DIGESTSIZE) ? APR_MD5_DIGESTSIZE : pl);
-	}
+    /*
+     * Then just as many characters of the MD5(pw, salt, pw)
+     */
+    MD5_Init(&ctx1);
+    MD5_Update(&ctx1, pw, strlen(pw));
+    MD5_Update(&ctx1, sp, sl);
+    MD5_Update(&ctx1, pw, strlen(pw));
+    MD5_Final(final, &ctx1);
+    for (pl = strlen(pw); pl > 0; pl -= APR_MD5_DIGESTSIZE) {
+        MD5_Update(&ctx, final,
+                      (pl > APR_MD5_DIGESTSIZE) ? APR_MD5_DIGESTSIZE : pl);
+    }
 
-	/*
-	 * Don't leave anything around in vm they could use.
-	 */
-	memset(final, 0, sizeof(final));
+    /*
+     * Don't leave anything around in vm they could use.
+     */
+    memset(final, 0, sizeof(final));
 
-	/*
-	 * Then something really weird...
-	 */
-	for (i = strlen(pw); i != 0; i >>= 1) {
-		if (i & 1) {
-			li_MD5_Update(&ctx, final, 1);
-		}
-		else {
-			li_MD5_Update(&ctx, pw, 1);
-		}
-	}
+    /*
+     * Then something really weird...
+     */
+    for (i = strlen(pw); i != 0; i >>= 1) {
+        if (i & 1) {
+            MD5_Update(&ctx, final, 1);
+        }
+        else {
+            MD5_Update(&ctx, pw, 1);
+        }
+    }
 
-	/*
-	 * Now make the output string.  We know our limitations, so we
-	 * can use the string routines without bounds checking.
-	 */
-	strcpy(passwd, APR1_ID);
-	strncat(passwd, sp, sl);
-	strcat(passwd, "$");
+    /*
+     * Now make the output string.  We know our limitations, so we
+     * can use the string routines without bounds checking.
+     */
+    strcpy(passwd, APR1_ID);
+    strncat(passwd, sp, sl);
+    strcat(passwd, "$");
 
-	li_MD5_Final(final, &ctx);
+    MD5_Final(final, &ctx);
 
-	/*
-	 * And now, just to make sure things don't run too fast..
-	 * On a 60 Mhz Pentium this takes 34 msec, so you would
-	 * need 30 seconds to build a 1000 entry dictionary...
-	 */
-	for (i = 0; i < 1000; i++) {
-		li_MD5_Init(&ctx1);
-		if (i & 1) {
-			li_MD5_Update(&ctx1, pw, strlen(pw));
-		}
-		else {
-			li_MD5_Update(&ctx1, final, APR_MD5_DIGESTSIZE);
-		}
-		if (i % 3) {
-			li_MD5_Update(&ctx1, sp, sl);
-		}
+    /*
+     * And now, just to make sure things don't run too fast..
+     * On a 60 Mhz Pentium this takes 34 msec, so you would
+     * need 30 seconds to build a 1000 entry dictionary...
+     */
+    for (i = 0; i < 1000; i++) {
+        MD5_Init(&ctx1);
+        if (i & 1) {
+            MD5_Update(&ctx1, pw, strlen(pw));
+        }
+        else {
+            MD5_Update(&ctx1, final, APR_MD5_DIGESTSIZE);
+        }
+        if (i % 3) {
+            MD5_Update(&ctx1, sp, sl);
+        }
 
-		if (i % 7) {
-			li_MD5_Update(&ctx1, pw, strlen(pw));
-		}
+        if (i % 7) {
+            MD5_Update(&ctx1, pw, strlen(pw));
+        }
 
-		if (i & 1) {
-			li_MD5_Update(&ctx1, final, APR_MD5_DIGESTSIZE);
-		}
-		else {
-			li_MD5_Update(&ctx1, pw, strlen(pw));
-		}
-		li_MD5_Final(final,&ctx1);
-	}
+        if (i & 1) {
+            MD5_Update(&ctx1, final, APR_MD5_DIGESTSIZE);
+        }
+        else {
+            MD5_Update(&ctx1, pw, strlen(pw));
+        }
+        MD5_Final(final,&ctx1);
+    }
 
-	p = passwd + strlen(passwd);
+    p = passwd + strlen(passwd);
 
-	l = (final[ 0]<<16) | (final[ 6]<<8) | final[12]; to64(p, l, 4); p += 4;
-	l = (final[ 1]<<16) | (final[ 7]<<8) | final[13]; to64(p, l, 4); p += 4;
-	l = (final[ 2]<<16) | (final[ 8]<<8) | final[14]; to64(p, l, 4); p += 4;
-	l = (final[ 3]<<16) | (final[ 9]<<8) | final[15]; to64(p, l, 4); p += 4;
-	l = (final[ 4]<<16) | (final[10]<<8) | final[ 5]; to64(p, l, 4); p += 4;
-	l =                    final[11]                ; to64(p, l, 2); p += 2;
-	*p = '\0';
+    l = (final[ 0]<<16) | (final[ 6]<<8) | final[12]; to64(p, l, 4); p += 4;
+    l = (final[ 1]<<16) | (final[ 7]<<8) | final[13]; to64(p, l, 4); p += 4;
+    l = (final[ 2]<<16) | (final[ 8]<<8) | final[14]; to64(p, l, 4); p += 4;
+    l = (final[ 3]<<16) | (final[ 9]<<8) | final[15]; to64(p, l, 4); p += 4;
+    l = (final[ 4]<<16) | (final[10]<<8) | final[ 5]; to64(p, l, 4); p += 4;
+    l =                    final[11]                ; to64(p, l, 2); p += 2;
+    *p = '\0';
 
-	/*
-	 * Don't leave anything around in vm they could use.
-	 */
-	safe_memclear(final, sizeof(final));
+    /*
+     * Don't leave anything around in vm they could use.
+     */
+    memset(final, 0, sizeof(final));
 
 	/* FIXME
 	 */
 #define apr_cpystrn strncpy
-	apr_cpystrn(result, passwd, nbytes - 1);
+    apr_cpystrn(result, passwd, nbytes - 1);
 }
 
-#ifdef USE_OPENSSL
-static void apr_sha_encode(const char *pw, char *result, size_t nbytes) {
-	unsigned char digest[20];
-	size_t base64_written;
-
-	SHA1((const unsigned char*) pw, strlen(pw), digest);
-
-	memset(result, 0, nbytes);
-
-	/* need 5 bytes for "{SHA}", 28 for base64 (3 bytes -> 4 bytes) of SHA1 (20 bytes), 1 terminating */
-	if (nbytes < 5 + 28 + 1) return;
-
-	memcpy(result, "{SHA}", 5);
-	base64_written = li_to_base64(result + 5, nbytes - 5, digest, 20, BASE64_STANDARD);
-	force_assert(base64_written == 28);
-	result[5 + base64_written] = '\0'; /* terminate string */
-}
-#endif
 
 /**
  *
@@ -537,21 +609,21 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 		 * user:realm:md5(user:realm:password)
 		 */
 
-		li_MD5_CTX Md5Ctx;
+		MD5_CTX Md5Ctx;
 		HASH HA1;
 		char a1[256];
 
-		li_MD5_Init(&Md5Ctx);
-		li_MD5_Update(&Md5Ctx, CONST_BUF_LEN(username));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, CONST_BUF_LEN(realm));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *)pw, strlen(pw));
-		li_MD5_Final(HA1, &Md5Ctx);
+		MD5_Init(&Md5Ctx);
+		MD5_Update(&Md5Ctx, (unsigned char *)username->ptr, username->used - 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)realm->ptr, realm->used - 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)pw, strlen(pw));
+		MD5_Final(HA1, &Md5Ctx);
 
 		CvtHex(HA1, a1);
 
-		if (0 == strcmp(password->ptr, a1)) {
+		if (buffer_is_equal_string(password, a1, strlen(a1))) {
 			return 0;
 		}
 	} else if (p->conf.auth_backend == AUTH_BACKEND_HTPASSWD) {
@@ -562,40 +634,65 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 			 */
 			apr_md5_encode(pw, password->ptr, sample, sizeof(sample));
 			return (strcmp(sample, password->ptr) == 0) ? 0 : 1;
-#ifdef USE_OPENSSL
-		} else if (0 == strncmp(password->ptr, "{SHA}", 5)) {
-			apr_sha_encode(pw, sample, sizeof(sample));
-			return (strcmp(sample, password->ptr) == 0) ? 0 : 1;
-#endif
 		} else {
-#if defined(HAVE_CRYPT_R) || defined(HAVE_CRYPT)
-			char *crypted;
-#if defined(HAVE_CRYPT_R)
-			struct crypt_data crypt_tmp_data;
-			crypt_tmp_data.initialized = 0;
-#endif
+#ifdef HAVE_CRYPT
+		char salt[32];
+		char *crypted;
+		size_t salt_len = 0;
+		/*
+		 * htpasswd format
+		 *
+		 * user:crypted password
+		 */
 
-			/* a simple DES password is 2 + 11 characters. everything else should be longer. */
-			if (buffer_string_length(password) < 13) {
-				return -1;
-			}
+		/*
+		 *  Algorithm      Salt
+		 *  CRYPT_STD_DES   2-character (Default)
+		 *  CRYPT_EXT_DES   9-character
+		 *  CRYPT_MD5       12-character beginning with $1$
+		 *  CRYPT_BLOWFISH  16-character beginning with $2$
+		 */
 
-#if defined(HAVE_CRYPT_R)
-			if (0 == (crypted = crypt_r(pw, password->ptr, &crypt_tmp_data))) {
-#else
-			if (0 == (crypted = crypt(pw, password->ptr))) {
-#endif
-				/* crypt failed. */
-				return -1;
-			}
-
-			if (0 == strcmp(password->ptr, crypted)) {
-				return 0;
-			}
-#endif
+		if (password->used < 13 + 1) {
+			fprintf(stderr, "%s.%d\n", __FILE__, __LINE__);
+			return -1;
 		}
+
+		if (password->used == 13 + 1) {
+			/* a simple DES password is 2 + 11 characters */
+			salt_len = 2;
+		} else if (password->ptr[0] == '$' && password->ptr[2] == '$') {
+			char *dollar = NULL;
+
+			if (NULL == (dollar = strchr(password->ptr + 3, '$'))) {
+				fprintf(stderr, "%s.%d\n", __FILE__, __LINE__);
+				return -1;
+			}
+
+			salt_len = dollar - password->ptr;
+		}
+
+		if (salt_len > sizeof(salt) - 1) {
+			fprintf(stderr, "%s.%d\n", __FILE__, __LINE__);
+			return -1;
+		}
+
+		strncpy(salt, password->ptr, salt_len);
+
+		salt[salt_len] = '\0';
+
+		crypted = crypt(pw, salt);
+
+		if (buffer_is_equal_string(password, crypted, strlen(crypted))) {
+			return 0;
+		} else {
+			fprintf(stderr, "%s.%d\n", __FILE__, __LINE__);
+		}
+
+#endif
+	}
 	} else if (p->conf.auth_backend == AUTH_BACKEND_PLAIN) {
-		if (0 == strcmp(password->ptr, pw)) {
+		if (buffer_is_equal_string(password, pw, strlen(pw))) {
 			return 0;
 		}
 	} else if (p->conf.auth_backend == AUTH_BACKEND_LDAP) {
@@ -605,7 +702,7 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 		char *dn;
 		int ret;
 		char *attrs[] = { LDAP_NO_ATTRS, NULL };
-		size_t i, len;
+		size_t i;
 
 		/* for now we stay synchronous */
 
@@ -624,20 +721,14 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 		 * a unpleasant way
 		 */
 
-		len = buffer_string_length(username);
-		for (i = 0; i < len; i++) {
+		for (i = 0; i < username->used - 1; i++) {
 			char c = username->ptr[i];
 
 			if (!isalpha(c) &&
-			    !isdigit(c) &&
-			    (c != ' ') &&
-			    (c != '@') &&
-			    (c != '-') &&
-			    (c != '_') &&
-			    (c != '.') ) {
+			    !isdigit(c)) {
 
 				log_error_write(srv, __FILE__, __LINE__, "sbd",
-					"ldap: invalid character (- _.@a-zA-Z0-9 allowed) in username:", username, i);
+					"ldap: invalid character (a-zA-Z0-9 allowed) in username:", username, i);
 
 				return -1;
 			}
@@ -647,33 +738,26 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 			return -1;
 
 		/* build filter */
-		buffer_copy_buffer(p->ldap_filter, p->conf.ldap_filter_pre);
+		buffer_copy_string_buffer(p->ldap_filter, p->conf.ldap_filter_pre);
 		buffer_append_string_buffer(p->ldap_filter, username);
 		buffer_append_string_buffer(p->ldap_filter, p->conf.ldap_filter_post);
 
 
 		/* 2. */
-		if (p->anon_conf->ldap == NULL ||
-		    LDAP_SUCCESS != (ret = ldap_search_s(p->anon_conf->ldap, p->conf.auth_ldap_basedn->ptr, LDAP_SCOPE_SUBTREE, p->ldap_filter->ptr, attrs, 0, &lm))) {
+		if (p->conf.ldap == NULL ||
+		    LDAP_SUCCESS != (ret = ldap_search_s(p->conf.ldap, p->conf.auth_ldap_basedn->ptr, LDAP_SCOPE_SUBTREE, p->ldap_filter->ptr, attrs, 0, &lm))) {
+			if (auth_ldap_init(srv, &p->conf) != HANDLER_GO_ON)
+				return -1;
+			if (LDAP_SUCCESS != (ret = ldap_search_s(p->conf.ldap, p->conf.auth_ldap_basedn->ptr, LDAP_SCOPE_SUBTREE, p->ldap_filter->ptr, attrs, 0, &lm))) {
 
-			/* try again; the ldap library sometimes fails for the first call but reconnects */
-			if (p->anon_conf->ldap == NULL || ret != LDAP_SERVER_DOWN ||
-			    LDAP_SUCCESS != (ret = ldap_search_s(p->anon_conf->ldap, p->conf.auth_ldap_basedn->ptr, LDAP_SCOPE_SUBTREE, p->ldap_filter->ptr, attrs, 0, &lm))) {
+			log_error_write(srv, __FILE__, __LINE__, "sssb",
+					"ldap:", ldap_err2string(ret), "filter:", p->ldap_filter);
 
-				if (auth_ldap_init(srv, p->anon_conf) != HANDLER_GO_ON)
-					return -1;
-
-				if (NULL == p->anon_conf->ldap) return -1;
-
-				if (LDAP_SUCCESS != (ret = ldap_search_s(p->anon_conf->ldap, p->conf.auth_ldap_basedn->ptr, LDAP_SCOPE_SUBTREE, p->ldap_filter->ptr, attrs, 0, &lm))) {
-					log_error_write(srv, __FILE__, __LINE__, "sssb",
-							"ldap:", ldap_err2string(ret), "filter:", p->ldap_filter);
-					return -1;
-				}
+			return -1;
 			}
 		}
 
-		if (NULL == (first = ldap_first_entry(p->anon_conf->ldap, lm))) {
+		if (NULL == (first = ldap_first_entry(p->conf.ldap, lm))) {
 			log_error_write(srv, __FILE__, __LINE__, "s", "ldap ...");
 
 			ldap_msgfree(lm);
@@ -681,7 +765,7 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 			return -1;
 		}
 
-		if (NULL == (dn = ldap_get_dn(p->anon_conf->ldap, first))) {
+		if (NULL == (dn = ldap_get_dn(p->conf.ldap, first))) {
 			log_error_write(srv, __FILE__, __LINE__, "s", "ldap ...");
 
 			ldap_msgfree(lm);
@@ -737,52 +821,45 @@ static int http_auth_basic_password_compare(server *srv, mod_auth_plugin_data *p
 	return -1;
 }
 
-int http_auth_basic_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, const char *realm_str) {
+int http_auth_basic_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, buffer *url, const char *realm_str) {
 	buffer *username, *password;
 	char *pw;
 
 	data_string *realm;
 
-	realm = (data_string *)array_get_element(req, "realm");
+	realm = (data_string *)array_get_element(req, CONST_STR_LEN("realm"));
 
 	username = buffer_init();
+	password = buffer_init();
 
-	if (!buffer_append_base64_decode(username, realm_str, strlen(realm_str), BASE64_STANDARD)) {
-		log_error_write(srv, __FILE__, __LINE__, "sb", "decodeing base64-string failed", username);
-
-		buffer_free(username);
-		return 0;
-	}
+	base64_decode(username, realm_str);
 
 	/* r2 == user:password */
 	if (NULL == (pw = strchr(username->ptr, ':'))) {
+		buffer_free(username);
+
 		log_error_write(srv, __FILE__, __LINE__, "sb", ": is missing in", username);
 
-		buffer_free(username);
 		return 0;
 	}
 
-	buffer_string_set_length(username, pw - username->ptr);
-	pw++;
+	*pw++ = '\0';
+
+	username->used = pw - username->ptr;
 
-	password = buffer_init();
 	/* copy password to r1 */
 	if (http_auth_get_password(srv, p, username, realm->value, password)) {
 		buffer_free(username);
 		buffer_free(password);
 
-		if (AUTH_BACKEND_UNSET == p->conf.auth_backend) {
-			log_error_write(srv, __FILE__, __LINE__, "s", "auth.backend is not set");
-		} else {
-			log_error_write(srv, __FILE__, __LINE__, "ss", "get_password failed, IP:", inet_ntop_cache_get_ip(srv, &(con->dst_addr)));
-		}
+		log_error_write(srv, __FILE__, __LINE__, "s", "get_password failed");
 
 		return 0;
 	}
 
 	/* password doesn't match */
 	if (http_auth_basic_password_compare(srv, p, req, username, realm->value, password, pw)) {
-		log_error_write(srv, __FILE__, __LINE__, "sbsBss", "password doesn't match for", con->uri.path, "username:", username, ", IP:", inet_ntop_cache_get_ip(srv, &(con->dst_addr)));
+		log_error_write(srv, __FILE__, __LINE__, "sbb", "password doesn't match for", con->uri.path, username);
 
 		buffer_free(username);
 		buffer_free(password);
@@ -791,7 +868,7 @@ int http_auth_basic_check(server *srv, connection *con, mod_auth_plugin_data *p,
 	}
 
 	/* value is our allow-rules */
-	if (http_auth_match_rules(srv, req, username->ptr, NULL, NULL)) {
+	if (http_auth_match_rules(srv, p, url->ptr, username->ptr, NULL, NULL)) {
 		buffer_free(username);
 		buffer_free(password);
 
@@ -801,7 +878,7 @@ int http_auth_basic_check(server *srv, connection *con, mod_auth_plugin_data *p,
 	}
 
 	/* remember the username */
-	buffer_copy_buffer(p->auth_user, username);
+	buffer_copy_string_buffer(p->auth_user, username);
 
 	buffer_free(username);
 	buffer_free(password);
@@ -815,27 +892,26 @@ typedef struct {
 	char **ptr;
 } digest_kv;
 
-/* return values: -1: error/bad request, 0: failed, 1: success */
-int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, const char *realm_str) {
+int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, buffer *url, const char *realm_str) {
 	char a1[256];
 	char a2[256];
 
-	char *username = NULL;
-	char *realm = NULL;
-	char *nonce = NULL;
-	char *uri = NULL;
-	char *algorithm = NULL;
-	char *qop = NULL;
-	char *cnonce = NULL;
-	char *nc = NULL;
-	char *respons = NULL;
+	char *username;
+	char *realm;
+	char *nonce;
+	char *uri;
+	char *algorithm;
+	char *qop;
+	char *cnonce;
+	char *nc;
+	char *respons;
 
 	char *e, *c;
 	const char *m = NULL;
 	int i;
 	buffer *password, *b, *username_buf, *realm_buf;
 
-	li_MD5_CTX Md5Ctx;
+	MD5_CTX Md5Ctx;
 	HASH HA1;
 	HASH HA2;
 	HASH RespHash;
@@ -869,9 +945,15 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 	dkv[6].ptr = &cnonce;
 	dkv[7].ptr = &nc;
 	dkv[8].ptr = &respons;
+	dkv[9].ptr = NULL;
 
 	UNUSED(req);
 
+	for (i = 0; dkv[i].key; i++) {
+		*(dkv[i].ptr) = NULL;
+	}
+
+
 	if (p->conf.auth_backend != AUTH_BACKEND_HTDIGEST &&
 	    p->conf.auth_backend != AUTH_BACKEND_PLAIN) {
 		log_error_write(srv, __FILE__, __LINE__, "s",
@@ -886,7 +968,7 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 	for (c = b->ptr; *c; c++) {
 		/* skip whitespaces */
 		while (*c == ' ' || *c == '\t') c++;
-		if (!*c) break;
+		if (!c) break;
 
 		for (i = 0; dkv[i].key; i++) {
 			if ((0 == strncmp(c, dkv[i].key, dkv[i].key_len))) {
@@ -917,7 +999,7 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 		log_error_write(srv, __FILE__, __LINE__, "ss", "realm", realm);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "nonce", nonce);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "uri", uri);
-		log_error_write(srv, __FILE__, __LINE__, "ss", "algorithm", algorithm);
+		log_error_write(srv, __FILE__, __LINE__, "ss", "algorigthm", algorithm);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "qop", qop);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "cnonce", cnonce);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "nc", nc);
@@ -935,29 +1017,6 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 
 		log_error_write(srv, __FILE__, __LINE__, "s",
 				"digest: missing field");
-
-		buffer_free(b);
-		return -1;
-	}
-
-	/**
-	 * protect the md5-sess against missing cnonce and nonce
-	 */
-	if (algorithm &&
-	    0 == strcasecmp(algorithm, "md5-sess") &&
-	    (!nonce || !cnonce)) {
-		log_error_write(srv, __FILE__, __LINE__, "s",
-				"digest: (md5-sess: missing field");
-
-		buffer_free(b);
-		return -1;
-	}
-
-	if (qop && strcasecmp(qop, "auth-int") == 0) {
-		log_error_write(srv, __FILE__, __LINE__, "s",
-				"digest: qop=auth-int not supported");
-
-		buffer_free(b);
 		return -1;
 	}
 
@@ -980,13 +1039,13 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 
 	if (p->conf.auth_backend == AUTH_BACKEND_PLAIN) {
 		/* generate password from plain-text */
-		li_MD5_Init(&Md5Ctx);
-		li_MD5_Update(&Md5Ctx, (unsigned char *)username, strlen(username));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *)realm, strlen(realm));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, CONST_BUF_LEN(password));
-		li_MD5_Final(HA1, &Md5Ctx);
+		MD5_Init(&Md5Ctx);
+		MD5_Update(&Md5Ctx, (unsigned char *)username, strlen(username));
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)realm, strlen(realm));
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)password->ptr, password->used - 1);
+		MD5_Final(HA1, &Md5Ctx);
 	} else if (p->conf.auth_backend == AUTH_BACKEND_HTDIGEST) {
 		/* HA1 */
 		/* transform the 32-byte-hex-md5 to a 16-byte-md5 */
@@ -1003,50 +1062,45 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 
 	if (algorithm &&
 	    strcasecmp(algorithm, "md5-sess") == 0) {
-		li_MD5_Init(&Md5Ctx);
-		/* Errata ID 1649: http://www.rfc-editor.org/errata_search.php?rfc=2617 */
-		CvtHex(HA1, a1);
-		li_MD5_Update(&Md5Ctx, (unsigned char *)a1, 32);
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *)nonce, strlen(nonce));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *)cnonce, strlen(cnonce));
-		li_MD5_Final(HA1, &Md5Ctx);
+		MD5_Init(&Md5Ctx);
+		MD5_Update(&Md5Ctx, (unsigned char *)HA1, 16);
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)nonce, strlen(nonce));
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)cnonce, strlen(cnonce));
+		MD5_Final(HA1, &Md5Ctx);
 	}
 
 	CvtHex(HA1, a1);
 
 	/* calculate H(A2) */
-	li_MD5_Init(&Md5Ctx);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)m, strlen(m));
-	li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-	li_MD5_Update(&Md5Ctx, (unsigned char *)uri, strlen(uri));
-	/* qop=auth-int not supported, already checked above */
-/*
+	MD5_Init(&Md5Ctx);
+	MD5_Update(&Md5Ctx, (unsigned char *)m, strlen(m));
+	MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+	MD5_Update(&Md5Ctx, (unsigned char *)uri, strlen(uri));
 	if (qop && strcasecmp(qop, "auth-int") == 0) {
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *) [body checksum], HASHHEXLEN);
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)"", HASHHEXLEN);
 	}
-*/
-	li_MD5_Final(HA2, &Md5Ctx);
+	MD5_Final(HA2, &Md5Ctx);
 	CvtHex(HA2, HA2Hex);
 
 	/* calculate response */
-	li_MD5_Init(&Md5Ctx);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)a1, HASHHEXLEN);
-	li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-	li_MD5_Update(&Md5Ctx, (unsigned char *)nonce, strlen(nonce));
-	li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
+	MD5_Init(&Md5Ctx);
+	MD5_Update(&Md5Ctx, (unsigned char *)a1, HASHHEXLEN);
+	MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+	MD5_Update(&Md5Ctx, (unsigned char *)nonce, strlen(nonce));
+	MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
 	if (qop && *qop) {
-		li_MD5_Update(&Md5Ctx, (unsigned char *)nc, strlen(nc));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *)cnonce, strlen(cnonce));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
-		li_MD5_Update(&Md5Ctx, (unsigned char *)qop, strlen(qop));
-		li_MD5_Update(&Md5Ctx, CONST_STR_LEN(":"));
+		MD5_Update(&Md5Ctx, (unsigned char *)nc, strlen(nc));
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)cnonce, strlen(cnonce));
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
+		MD5_Update(&Md5Ctx, (unsigned char *)qop, strlen(qop));
+		MD5_Update(&Md5Ctx, (unsigned char *)":", 1);
 	};
-	li_MD5_Update(&Md5Ctx, (unsigned char *)HA2Hex, HASHHEXLEN);
-	li_MD5_Final(RespHash, &Md5Ctx);
+	MD5_Update(&Md5Ctx, (unsigned char *)HA2Hex, HASHHEXLEN);
+	MD5_Final(RespHash, &Md5Ctx);
 	CvtHex(RespHash, a2);
 
 	if (0 != strcmp(a2, respons)) {
@@ -1057,15 +1111,15 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 				"digest: digest mismatch", a2, respons);
 		}
 
-		log_error_write(srv, __FILE__, __LINE__, "ssss",
-				"digest: auth failed for ", username, ": wrong password, IP:", inet_ntop_cache_get_ip(srv, &(con->dst_addr)));
+		log_error_write(srv, __FILE__, __LINE__, "sss",
+				"digest: auth failed for", username, "wrong password");
 
 		buffer_free(b);
 		return 0;
 	}
 
 	/* value is our allow-rules */
-	if (http_auth_match_rules(srv, req, username, NULL, NULL)) {
+	if (http_auth_match_rules(srv, p, url->ptr, username, NULL, NULL)) {
 		buffer_free(b);
 
 		log_error_write(srv, __FILE__, __LINE__, "s",
@@ -1089,24 +1143,23 @@ int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p
 
 int http_auth_digest_generate_nonce(server *srv, mod_auth_plugin_data *p, buffer *fn, char out[33]) {
 	HASH h;
-	li_MD5_CTX Md5Ctx;
-	char hh[LI_ITOSTRING_LENGTH];
+	MD5_CTX Md5Ctx;
+	char hh[32];
 
 	UNUSED(p);
 
 	/* generate shared-secret */
-	li_MD5_Init(&Md5Ctx);
-	li_MD5_Update(&Md5Ctx, CONST_BUF_LEN(fn));
-	li_MD5_Update(&Md5Ctx, CONST_STR_LEN("+"));
+	MD5_Init(&Md5Ctx);
+	MD5_Update(&Md5Ctx, (unsigned char *)fn->ptr, fn->used - 1);
+	MD5_Update(&Md5Ctx, (unsigned char *)"+", 1);
 
 	/* we assume sizeof(time_t) == 4 here, but if not it ain't a problem at all */
-	li_itostr(hh, srv->cur_ts);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
-	li_MD5_Update(&Md5Ctx, (unsigned char *)srv->entropy, sizeof(srv->entropy));
-	li_itostr(hh, rand());
-	li_MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
+	ltostr(hh, srv->cur_ts);
+	MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
+	ltostr(hh, rand());
+	MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
 
-	li_MD5_Final(h, &Md5Ctx);
+	MD5_Final(h, &Md5Ctx);
 
 	CvtHex(h, out);
 
