@@ -35,6 +35,8 @@ typedef struct {
 	buffer *auth_ldap_bindpw;
 	buffer *auth_ldap_filter;
 	buffer *auth_ldap_cafile;
+	buffer *auth_ldap_cert;
+	buffer *auth_ldap_key;
 	unsigned short auth_ldap_starttls;
 	unsigned short auth_ldap_allow_empty_pw;
 
@@ -63,12 +65,11 @@ typedef struct {
 
 	mod_auth_plugin_config **config_storage;
 
-	mod_auth_plugin_config conf, *anon_conf; /* this is only used as long as no handler_ctx is setup */
+	mod_auth_plugin_config conf; /* this is only used as long as no handler_ctx is setup */
 } mod_auth_plugin_data;
 
-int http_auth_basic_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, const char *realm_str);
-int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, const char *realm_str);
+int http_auth_basic_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, buffer *url, const char *realm_str);
+int http_auth_digest_check(server *srv, connection *con, mod_auth_plugin_data *p, array *req, buffer *url, const char *realm_str);
 int http_auth_digest_generate_nonce(server *srv, mod_auth_plugin_data *p, buffer *fn, char hh[33]);
-int http_auth_match_rules(server *srv, array *req, const char *username, const char *group, const char *host);
 
 #endif
