@@ -2,6 +2,6 @@
 #define _INET_NTOP_CACHE_H_
 
 #include "base.h"
-const char * inet_ntop_cache_get_ip(server *srv, sock_addr *addr);
+LI_API const char * inet_ntop_cache_get_ip(server *srv, sock_addr *addr);
 
 #endif
