@@ -1,12 +1,14 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
 
 #include "plugin.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
+#include "sys-strings.h"
 
 typedef struct {
 	array *access_deny;
@@ -23,6 +25,8 @@ typedef struct {
 INIT_FUNC(mod_access_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	return p;
@@ -40,8 +44,6 @@ FREE_FUNC(mod_access_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
-
 			array_free(s->access_deny);
 
 			free(s);
@@ -63,10 +65,9 @@ SETDEFAULTS_FUNC(mod_access_set_defaults) {
 		{ NULL,                          NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
 	};
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
@@ -76,7 +77,7 @@ SETDEFAULTS_FUNC(mod_access_set_defaults) {
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 	}
@@ -84,13 +85,11 @@ SETDEFAULTS_FUNC(mod_access_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_access_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(access_deny);
+	PATCH_OPTION(access_deny);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -105,72 +104,48 @@ static int mod_access_patch_connection(server *srv, connection *con, plugin_data
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("url.access-deny"))) {
-				PATCH(access_deny);
+				PATCH_OPTION(access_deny);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
-/**
- * URI handler
- *
- * we will get called twice:
- * - after the clean up of the URL and 
- * - after the pathinfo checks are done
- *
- * this handles the issue of trailing slashes
- */
 URIHANDLER_FUNC(mod_access_uri_handler) {
 	plugin_data *p = p_d;
 	int s_len;
 	size_t k;
 
-	if (buffer_is_empty(con->uri.path)) return HANDLER_GO_ON;
+	if (con->uri.path->used == 0) return HANDLER_GO_ON;
 
 	mod_access_patch_connection(srv, con, p);
 
-	s_len = buffer_string_length(con->uri.path);
-
-	if (con->conf.log_request_handling) {
-		log_error_write(srv, __FILE__, __LINE__, "s",
-				"-- mod_access_uri_handler called");
-	}
+	s_len = con->uri.path->used - 1;
 
 	for (k = 0; k < p->conf.access_deny->used; k++) {
 		data_string *ds = (data_string *)p->conf.access_deny->data[k];
-		int ct_len = buffer_string_length(ds->value);
-		int denied = 0;
-
+		int ct_len = ds->value->used - 1;
 
 		if (ct_len > s_len) continue;
-		if (buffer_is_empty(ds->value)) continue;
+
+		if (ds->value->used == 0) continue;
 
 		/* if we have a case-insensitive FS we have to lower-case the URI here too */
 
 		if (con->conf.force_lowercase_filenames) {
 			if (0 == strncasecmp(con->uri.path->ptr + s_len - ct_len, ds->value->ptr, ct_len)) {
-				denied = 1;
+				con->http_status = 403;
+
+				return HANDLER_FINISHED;
 			}
 		} else {
 			if (0 == strncmp(con->uri.path->ptr + s_len - ct_len, ds->value->ptr, ct_len)) {
-				denied = 1;
+				con->http_status = 403;
+
+				return HANDLER_FINISHED;
 			}
 		}
-
-		if (denied) {
-			con->http_status = 403;
-			con->mode = DIRECT;
-
-			if (con->conf.log_request_handling) {
-	 			log_error_write(srv, __FILE__, __LINE__, "sb", 
-					"url denied as we match:", ds->value);
-			}
-
-			return HANDLER_FINISHED;
-		}
 	}
 
 	/* not found */
@@ -178,15 +153,13 @@ URIHANDLER_FUNC(mod_access_uri_handler) {
 }
 
 
-int mod_access_plugin_init(plugin *p);
-int mod_access_plugin_init(plugin *p) {
+LI_EXPORT int mod_access_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("access");
 
 	p->init        = mod_access_init;
 	p->set_defaults = mod_access_set_defaults;
-	p->handle_uri_clean = mod_access_uri_handler;
-	p->handle_subrequest_start  = mod_access_uri_handler;
+	p->handle_uri_clean  = mod_access_uri_handler;
 	p->cleanup     = mod_access_free;
 
 	p->data        = NULL;
