@@ -1,3 +1,22 @@
+/*
+ * make sure _GNU_SOURCE is defined
+ */
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
+
+#include <sys/types.h>
+#include <sys/stat.h>
+
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <fcntl.h> /* only the defines on windows */
+#include <errno.h>
+#include <time.h>
+
+#include <stdio.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -7,19 +26,7 @@
 #include "inet_ntop_cache.h"
 
 #include "sys-socket.h"
-
-#include <sys/types.h>
-#include <sys/stat.h>
-
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <fcntl.h>
-#include <unistd.h>
-#include <errno.h>
-#include <time.h>
-
-#include <stdio.h>
+#include "sys-files.h"
 
 #ifdef HAVE_SYSLOG_H
 # include <syslog.h>
@@ -69,7 +76,7 @@ typedef struct {
  *
  */
 
-static const format_mapping fmap[] =
+const format_mapping fmap[] =
 {
 	{ '%', FORMAT_PERCENT },
 	{ 'h', FORMAT_REMOTE_HOST },
@@ -124,20 +131,17 @@ typedef struct {
 
 typedef struct {
 	buffer *access_logfile;
-	int    log_access_fd;
-	buffer *access_logbuffer; /* each logfile has a separate buffer */
-
-	unsigned short use_syslog; /* syslog has global buffer */
-	unsigned short syslog_level;
-
 	buffer *format;
+	unsigned short use_syslog;
 
+
+	int    log_access_fd;
 	time_t last_generated_accesslog_ts;
 	time_t *last_generated_accesslog_ts_ptr;
 
+
+	buffer *access_logbuffer;
 	buffer *ts_accesslog_str;
-	buffer *ts_accesslog_fmt_str;
-	unsigned short append_tz_offset;
 
 	format_fields *parsed_format;
 } plugin_config;
@@ -147,106 +151,37 @@ typedef struct {
 
 	plugin_config **config_storage;
 	plugin_config conf;
-
-	buffer *syslog_logbuffer; /* syslog has global buffer. no caching, always written directly */
 } plugin_data;
 
 INIT_FUNC(mod_accesslog_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
-	p->syslog_logbuffer = buffer_init();
 
 	return p;
 }
 
-static void accesslog_write_all(server *srv, const buffer *filename, int fd, const void* buf, size_t count) {
-	if (-1 == write_all(fd, buf, count)) {
-		log_error_write(srv, __FILE__, __LINE__, "sbs",
-			"writing access log entry failed:", filename, strerror(errno));
-	}
-}
-
-static void accesslog_append_escaped(buffer *dest, buffer *str) {
-	char *ptr, *start, *end;
-
-	/* replaces non-printable chars with \xHH where HH is the hex representation of the byte */
-	/* exceptions: " => \", \ => \\, whitespace chars => \n \t etc. */
-	if (buffer_string_is_empty(str)) return;
-	buffer_string_prepare_append(dest, buffer_string_length(str));
-
-	for (ptr = start = str->ptr, end = str->ptr + buffer_string_length(str); ptr < end; ptr++) {
-		unsigned char const c = (unsigned char) *ptr;
-		if (c >= ' ' && c <= '~' && c != '"' && c != '\\') {
-			/* nothing to change, add later as one block */
-		} else {
-			/* copy previous part */
-			if (start < ptr) {
-				buffer_append_string_len(dest, start, ptr - start);
-			}
-			start = ptr + 1;
-
-			switch (c) {
-			case '"':
-				BUFFER_APPEND_STRING_CONST(dest, "\\\"");
-				break;
-			case '\\':
-				BUFFER_APPEND_STRING_CONST(dest, "\\\\");
-				break;
-			case '\b':
-				BUFFER_APPEND_STRING_CONST(dest, "\\b");
-				break;
-			case '\n':
-				BUFFER_APPEND_STRING_CONST(dest, "\\n");
-				break;
-			case '\r':
-				BUFFER_APPEND_STRING_CONST(dest, "\\r");
-				break;
-			case '\t':
-				BUFFER_APPEND_STRING_CONST(dest, "\\t");
-				break;
-			case '\v':
-				BUFFER_APPEND_STRING_CONST(dest, "\\v");
-				break;
-			default: {
-					/* non printable char => \xHH */
-					char hh[5] = {'\\','x',0,0,0};
-					char h = c / 16;
-					hh[2] = (h > 9) ? (h - 10 + 'A') : (h + '0');
-					h = c % 16;
-					hh[3] = (h > 9) ? (h - 10 + 'A') : (h + '0');
-					buffer_append_string_len(dest, &hh[0], 4);
-				}
-				break;
-			}
-		}
-	}
-
-	if (start < end) {
-		buffer_append_string_len(dest, start, end - start);
-	}
-}
-
-static int accesslog_parse_format(server *srv, format_fields *fields, buffer *format) {
+int accesslog_parse_format(server *srv, format_fields *fields, buffer *format) {
 	size_t i, j, k = 0, start = 0;
 
-	if (buffer_is_empty(format)) return -1;
+	for (i = 0; i < format->used - 1; i++) {
 
-	for (i = 0; i < buffer_string_length(format); i++) {
 		switch(format->ptr[i]) {
 		case '%':
-			if (i > 0 && start != i) {
-				/* copy the string before this % */
+			if (start != i) {
+				/* copy the string */
 				if (fields->size == 0) {
 					fields->size = 16;
 					fields->used = 0;
-					fields->ptr = malloc(fields->size * sizeof(format_field * ));
+					fields->ptr = malloc(fields->size * sizeof(format_fields * ));
 				} else if (fields->used == fields->size) {
 					fields->size += 16;
-					fields->ptr = realloc(fields->ptr, fields->size * sizeof(format_field * ));
+					fields->ptr = realloc(fields->ptr, fields->size * sizeof(format_fields * ));
 				}
 
-				fields->ptr[fields->used] = malloc(sizeof(format_field));
+				fields->ptr[fields->used] = malloc(sizeof(format_fields));
 				fields->ptr[fields->used]->type = FIELD_STRING;
 				fields->ptr[fields->used]->string = buffer_init();
 
@@ -255,34 +190,30 @@ static int accesslog_parse_format(server *srv, format_fields *fields, buffer *fo
 				fields->used++;
 			}
 
+
 			/* we need a new field */
 
 			if (fields->size == 0) {
 				fields->size = 16;
 				fields->used = 0;
-				fields->ptr = malloc(fields->size * sizeof(format_field * ));
+				fields->ptr = malloc(fields->size * sizeof(format_fields * ));
 			} else if (fields->used == fields->size) {
 				fields->size += 16;
-				fields->ptr = realloc(fields->ptr, fields->size * sizeof(format_field * ));
+				fields->ptr = realloc(fields->ptr, fields->size * sizeof(format_fields * ));
 			}
 
 			/* search for the terminating command */
 			switch (format->ptr[i+1]) {
 			case '>':
 			case '<':
-				/* after the } has to be a character */
-				if (format->ptr[i+2] == '\0') {
-					log_error_write(srv, __FILE__, __LINE__, "s", "%< and %> have to be followed by a format-specifier");
-					return -1;
-				}
-
+				/* only for s */
 
 				for (j = 0; fmap[j].key != '\0'; j++) {
 					if (fmap[j].key != format->ptr[i+2]) continue;
 
 					/* found key */
 
-					fields->ptr[fields->used] = malloc(sizeof(format_field));
+					fields->ptr[fields->used] = malloc(sizeof(format_fields));
 					fields->ptr[fields->used]->type = FIELD_FORMAT;
 					fields->ptr[fields->used]->field = fmap[j].type;
 					fields->ptr[fields->used]->string = NULL;
@@ -293,34 +224,26 @@ static int accesslog_parse_format(server *srv, format_fields *fields, buffer *fo
 				}
 
 				if (fmap[j].key == '\0') {
-					log_error_write(srv, __FILE__, __LINE__, "s", "%< and %> have to be followed by a valid format-specifier");
+					log_error_write(srv, __FILE__, __LINE__, "ss", "config: ", "failed");
 					return -1;
 				}
 
 				start = i + 3;
-				i = start - 1; /* skip the string */
 
 				break;
 			case '{':
 				/* go forward to } */
 
-				for (k = i+2; k < buffer_string_length(format); k++) {
+				for (k = i+2; k < format->used - 1; k++) {
 					if (format->ptr[k] == '}') break;
 				}
 
-				if (k == buffer_string_length(format)) {
-					log_error_write(srv, __FILE__, __LINE__, "s", "%{ has to be terminated by a }");
+				if (k == format->used - 1) {
+					log_error_write(srv, __FILE__, __LINE__, "ss", "config: ", "failed");
 					return -1;
 				}
-
-				/* after the } has to be a character */
 				if (format->ptr[k+1] == '\0') {
-					log_error_write(srv, __FILE__, __LINE__, "s", "%{...} has to be followed by a format-specifier");
-					return -1;
-				}
-
-				if (k == i + 2) {
-					log_error_write(srv, __FILE__, __LINE__, "s", "%{...} has to be contain a string");
+					log_error_write(srv, __FILE__, __LINE__, "ss", "config: ", "failed");
 					return -1;
 				}
 
@@ -329,7 +252,7 @@ static int accesslog_parse_format(server *srv, format_fields *fields, buffer *fo
 
 					/* found key */
 
-					fields->ptr[fields->used] = malloc(sizeof(format_field));
+					fields->ptr[fields->used] = malloc(sizeof(format_fields));
 					fields->ptr[fields->used]->type = FIELD_FORMAT;
 					fields->ptr[fields->used]->field = fmap[j].type;
 					fields->ptr[fields->used]->string = buffer_init();
@@ -342,27 +265,20 @@ static int accesslog_parse_format(server *srv, format_fields *fields, buffer *fo
 				}
 
 				if (fmap[j].key == '\0') {
-					log_error_write(srv, __FILE__, __LINE__, "s", "%{...} has to be followed by a valid format-specifier");
+					log_error_write(srv, __FILE__, __LINE__, "ss", "config: ", "failed");
 					return -1;
 				}
 
 				start = k + 2;
-				i = start - 1; /* skip the string */
 
 				break;
 			default:
-				/* after the % has to be a character */
-				if (format->ptr[i+1] == '\0') {
-					log_error_write(srv, __FILE__, __LINE__, "s", "% has to be followed by a format-specifier");
-					return -1;
-				}
-
 				for (j = 0; fmap[j].key != '\0'; j++) {
 					if (fmap[j].key != format->ptr[i+1]) continue;
 
 					/* found key */
 
-					fields->ptr[fields->used] = malloc(sizeof(format_field));
+					fields->ptr[fields->used] = malloc(sizeof(format_fields));
 					fields->ptr[fields->used]->type = FIELD_FORMAT;
 					fields->ptr[fields->used]->field = fmap[j].type;
 					fields->ptr[fields->used]->string = NULL;
@@ -373,12 +289,11 @@ static int accesslog_parse_format(server *srv, format_fields *fields, buffer *fo
 				}
 
 				if (fmap[j].key == '\0') {
-					log_error_write(srv, __FILE__, __LINE__, "s", "% has to be followed by a valid format-specifier");
+					log_error_write(srv, __FILE__, __LINE__, "ss", "config: ", "failed");
 					return -1;
 				}
 
 				start = i + 2;
-				i = start - 1; /* skip the string */
 
 				break;
 			}
@@ -392,13 +307,13 @@ static int accesslog_parse_format(server *srv, format_fields *fields, buffer *fo
 		if (fields->size == 0) {
 			fields->size = 16;
 			fields->used = 0;
-			fields->ptr = malloc(fields->size * sizeof(format_field * ));
+			fields->ptr = malloc(fields->size * sizeof(format_fields * ));
 		} else if (fields->used == fields->size) {
 			fields->size += 16;
-			fields->ptr = realloc(fields->ptr, fields->size * sizeof(format_field * ));
+			fields->ptr = realloc(fields->ptr, fields->size * sizeof(format_fields * ));
 		}
 
-		fields->ptr[fields->used] = malloc(sizeof(format_field));
+		fields->ptr[fields->used] = malloc(sizeof(format_fields));
 		fields->ptr[fields->used]->type = FIELD_STRING;
 		fields->ptr[fields->used]->string = buffer_init();
 
@@ -421,18 +336,23 @@ FREE_FUNC(mod_accesslog_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
+			if (!s) continue;
 
-			if (!buffer_string_is_empty(s->access_logbuffer)) {
-				if (s->log_access_fd != -1) {
-					accesslog_write_all(srv, s->access_logfile, s->log_access_fd, CONST_BUF_LEN(s->access_logbuffer));
+			if (s->access_logbuffer->used) {
+				if (s->use_syslog) {
+# ifdef HAVE_SYSLOG_H
+					if (s->access_logbuffer->used > 2) {
+						syslog(LOG_INFO, "%*s", s->access_logbuffer->used - 2, s->access_logbuffer->ptr);
+					}
+# endif
+				} else if (s->log_access_fd != -1) {
+					write(s->log_access_fd, s->access_logbuffer->ptr, s->access_logbuffer->used - 1);
 				}
 			}
 
 			if (s->log_access_fd != -1) close(s->log_access_fd);
 
 			buffer_free(s->ts_accesslog_str);
-			buffer_free(s->ts_accesslog_fmt_str);
 			buffer_free(s->access_logbuffer);
 			buffer_free(s->format);
 			buffer_free(s->access_logfile);
@@ -453,7 +373,6 @@ FREE_FUNC(mod_accesslog_free) {
 		free(p->config_storage);
 	}
 
-	if (p->syslog_logbuffer) buffer_free(p->syslog_logbuffer);
 	free(p);
 
 	return HANDLER_GO_ON;
@@ -467,16 +386,14 @@ SETDEFAULTS_FUNC(log_access_open) {
 		{ "accesslog.filename",             NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
 		{ "accesslog.use-syslog",           NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },
 		{ "accesslog.format",               NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
-		{ "accesslog.syslog-level",         NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },
 		{ NULL,                             NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
 	};
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
@@ -484,35 +401,30 @@ SETDEFAULTS_FUNC(log_access_open) {
 		s->format = buffer_init();
 		s->access_logbuffer = buffer_init();
 		s->ts_accesslog_str = buffer_init();
-		s->ts_accesslog_fmt_str = buffer_init();
 		s->log_access_fd = -1;
 		s->last_generated_accesslog_ts = 0;
 		s->last_generated_accesslog_ts_ptr = &(s->last_generated_accesslog_ts);
-		s->syslog_level = LOG_INFO;
 
 
 		cv[0].destination = s->access_logfile;
 		cv[1].destination = &(s->use_syslog);
 		cv[2].destination = s->format;
-		cv[3].destination = &(s->syslog_level);
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 
-		if (i == 0 && buffer_string_is_empty(s->format)) {
+		if (i == 0 && buffer_is_empty(s->format)) {
 			/* set a default logfile string */
 
-			buffer_copy_string_len(s->format, CONST_STR_LEN("%h %V %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\""));
+			buffer_copy_string(s->format, "%h %V %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\"");
 		}
 
 		/* parse */
 
-		if (!buffer_is_empty(s->format)) {
-			size_t j, count;
-
+		if (s->format->used) {
 			s->parsed_format = calloc(1, sizeof(*(s->parsed_format)));
 
 			if (-1 == accesslog_parse_format(srv, s->parsed_format, s->format)) {
@@ -522,29 +434,6 @@ SETDEFAULTS_FUNC(log_access_open) {
 
 				return HANDLER_ERROR;
 			}
-
-			/* make sure they didn't try to send the timestamp in twice...
-			 * also, save the format string in a different variable (this
-			 * will save a few conditionals later)
-			 */
-			count = 0;
-			for (j = 0; j < s->parsed_format->used; j++) {
-				if (FIELD_FORMAT == s->parsed_format->ptr[j]->type) {
-					if (FORMAT_TIMESTAMP == s->parsed_format->ptr[j]->field) {
-						if (!buffer_string_is_empty(s->parsed_format->ptr[j]->string)) {
-							buffer_copy_string(s->ts_accesslog_fmt_str, s->parsed_format->ptr[j]->string->ptr);
-						}
-
-						if (++count > 1) {
-							log_error_write(srv, __FILE__, __LINE__, "sb",
-								"you may not use the timestamp twice in the same access log:", s->format);
-
-							return HANDLER_ERROR;
-						}
-					}
-				}
-			}
-
 #if 0
 			/* debugging */
 			for (j = 0; j < s->parsed_format->used; j++) {
@@ -565,26 +454,79 @@ SETDEFAULTS_FUNC(log_access_open) {
 #endif
 		}
 
-		s->append_tz_offset = 0;
-		if (buffer_string_is_empty(s->ts_accesslog_fmt_str)) {
-#if defined(HAVE_STRUCT_TM_GMTOFF)
-			BUFFER_COPY_STRING_CONST(s->ts_accesslog_fmt_str, "[%d/%b/%Y:%H:%M:%S ");
-			s->append_tz_offset = 1;
-#else
-			BUFFER_COPY_STRING_CONST(s->ts_accesslog_fmt_str, "[%d/%b/%Y:%H:%M:%S +0000]");
-#endif
-		}
-
 		if (s->use_syslog) {
 			/* ignore the next checks */
 			continue;
 		}
 
-		if (buffer_string_is_empty(s->access_logfile)) continue;
+		if (buffer_is_empty(s->access_logfile)) continue;
+
+		if (s->access_logfile->ptr[0] == '|') {
+#ifdef HAVE_FORK
+			/* create write pipe and spawn process */
+
+			int to_log_fds[2];
+			pid_t pid;
+
+			if (pipe(to_log_fds)) {
+				log_error_write(srv, __FILE__, __LINE__, "ss", "pipe failed: ", strerror(errno));
+				return HANDLER_ERROR;
+			}
+
+			/* fork, execve */
+			switch (pid = fork()) {
+			case 0:
+				/* child */
+
+				close(STDIN_FILENO);
+				dup2(to_log_fds[0], STDIN_FILENO);
+				close(to_log_fds[0]);
+				/* not needed */
+				close(to_log_fds[1]);
+
+				/* we don't need the client socket */
+				for (i = 3; i < 256; i++) {
+					close(i);
+				}
+
+				/* exec the log-process (skip the | )
+				 *
+				 */
+
+				execl("/bin/sh", "sh", "-c", s->access_logfile->ptr + 1, NULL);
+
+				log_error_write(srv, __FILE__, __LINE__, "sss",
+						"spawning log-process failed: ", strerror(errno),
+						s->access_logfile->ptr + 1);
+
+				exit(-1);
+				break;
+			case -1:
+				/* error */
+				log_error_write(srv, __FILE__, __LINE__, "ss", "fork failed: ", strerror(errno));
+				break;
+			default:
+				close(to_log_fds[0]);
+
+				s->log_access_fd = to_log_fds[1];
+
+				break;
+			}
+#else
+			return -1;
+#endif
+		} else if (-1 == (s->log_access_fd =
+				  open(s->access_logfile->ptr, O_APPEND | O_WRONLY | O_CREAT | O_LARGEFILE, 0644))) {
+
+			log_error_write(srv, __FILE__, __LINE__, "ssb",
+					"opening access-log failed:",
+					strerror(errno), s->access_logfile);
 
-		if (-1 == (s->log_access_fd = open_logfile_or_pipe(srv, s->access_logfile->ptr)))
 			return HANDLER_ERROR;
-
+		}
+#ifndef _WIN32
+		fcntl(s->log_access_fd, F_SETFD, FD_CLOEXEC);
+#endif
 	}
 
 	return HANDLER_GO_ON;
@@ -599,19 +541,26 @@ SIGHUP_FUNC(log_access_cycle) {
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s = p->config_storage[i];
 
-		if (!buffer_string_is_empty(s->access_logbuffer)) {
-			if (s->log_access_fd != -1) {
-				accesslog_write_all(srv, s->access_logfile, s->log_access_fd, CONST_BUF_LEN(s->access_logbuffer));
+		if (s->access_logbuffer->used) {
+			if (s->use_syslog) {
+#ifdef HAVE_SYSLOG_H
+				if (s->access_logbuffer->used > 2) {
+					/* syslog appends a \n on its own */
+					syslog(LOG_INFO, "%*s", s->access_logbuffer->used - 2, s->access_logbuffer->ptr);
+				}
+#endif
+			} else if (s->log_access_fd != -1) {
+				write(s->log_access_fd, s->access_logbuffer->ptr, s->access_logbuffer->used - 1);
 			}
 
 			buffer_reset(s->access_logbuffer);
 		}
 
-		if (s->use_syslog == 0
-			&& !buffer_string_is_empty(s->access_logfile)
-			&& s->access_logfile->ptr[0] != '|') {
+		if (s->use_syslog == 0 &&
+		    !buffer_is_empty(s->access_logfile) &&
+		    s->access_logfile->ptr[0] != '|') {
 
-			if (-1 != s->log_access_fd) close(s->log_access_fd);
+			close(s->log_access_fd);
 
 			if (-1 == (s->log_access_fd =
 				   open(s->access_logfile->ptr, O_APPEND | O_WRONLY | O_CREAT | O_LARGEFILE, 0644))) {
@@ -620,30 +569,24 @@ SIGHUP_FUNC(log_access_cycle) {
 
 				return HANDLER_ERROR;
 			}
-			fd_close_on_exec(s->log_access_fd);
 		}
 	}
 
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_accesslog_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(access_logfile);
-	PATCH(format);
-	PATCH(log_access_fd);
-	PATCH(last_generated_accesslog_ts_ptr);
-	PATCH(access_logbuffer);
-	PATCH(ts_accesslog_str);
-	PATCH(ts_accesslog_fmt_str);
-	PATCH(append_tz_offset);
-	PATCH(parsed_format);
-	PATCH(use_syslog);
-	PATCH(syslog_level);
+	PATCH_OPTION(access_logfile);
+	PATCH_OPTION(format);
+	PATCH_OPTION(log_access_fd);
+	PATCH_OPTION(last_generated_accesslog_ts_ptr);
+	PATCH_OPTION(access_logbuffer);
+	PATCH_OPTION(ts_accesslog_str);
+	PATCH_OPTION(parsed_format);
+	PATCH_OPTION(use_syslog);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -658,27 +601,22 @@ static int mod_accesslog_patch_connection(server *srv, connection *con, plugin_d
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("accesslog.filename"))) {
-				PATCH(access_logfile);
-				PATCH(log_access_fd);
-				PATCH(access_logbuffer);
+				PATCH_OPTION(access_logfile);
+				PATCH_OPTION(log_access_fd);
+				PATCH_OPTION(last_generated_accesslog_ts_ptr);
+				PATCH_OPTION(access_logbuffer);
+				PATCH_OPTION(ts_accesslog_str);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("accesslog.format"))) {
-				PATCH(format);
-				PATCH(parsed_format);
-				PATCH(last_generated_accesslog_ts_ptr);
-				PATCH(ts_accesslog_str);
-				PATCH(ts_accesslog_fmt_str);
-				PATCH(append_tz_offset);
+				PATCH_OPTION(format);
+				PATCH_OPTION(parsed_format);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("accesslog.use-syslog"))) {
-				PATCH(use_syslog);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("accesslog.syslog-level"))) {
-				PATCH(syslog_level);
+				PATCH_OPTION(use_syslog);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 REQUESTDONE_FUNC(log_access_write) {
 	plugin_data *p = p_d;
@@ -690,17 +628,9 @@ REQUESTDONE_FUNC(log_access_write) {
 
 	mod_accesslog_patch_connection(srv, con, p);
 
-	/* No output device, nothing to do */
-	if (!p->conf.use_syslog && p->conf.log_access_fd == -1) return HANDLER_GO_ON;
-
-	if (p->conf.use_syslog) {
-		b = p->syslog_logbuffer;
-	} else {
-		b = p->conf.access_logbuffer;
-	}
-
-	if (buffer_is_empty(b)) {
-		buffer_string_set_length(b, 0);
+	b = p->conf.access_logbuffer;
+	if (b->used == 0) {
+		buffer_copy_string(b, "");
 	}
 
 	for (j = 0; j < p->conf.parsed_format->used; j++) {
@@ -719,38 +649,38 @@ REQUESTDONE_FUNC(log_access_write) {
 					long scd, hrs, min;
 #endif
 
-					buffer_string_prepare_copy(p->conf.ts_accesslog_str, 255);
+					buffer_prepare_copy(p->conf.ts_accesslog_str, 255);
 #if defined(HAVE_STRUCT_TM_GMTOFF)
 # ifdef HAVE_LOCALTIME_R
 					localtime_r(&(srv->cur_ts), &tm);
-					buffer_append_strftime(p->conf.ts_accesslog_str, p->conf.ts_accesslog_fmt_str->ptr, &tm);
-# else /* HAVE_LOCALTIME_R */
-					buffer_append_strftime(p->conf.ts_accesslog_str, p->conf.ts_accesslog_fmt_str->ptr, localtime(&(srv->cur_ts)));
-# endif /* HAVE_LOCALTIME_R */
+					strftime(p->conf.ts_accesslog_str->ptr, p->conf.ts_accesslog_str->size - 1, "[%d/%b/%Y:%H:%M:%S ", &tm);
+# else
+					strftime(p->conf.ts_accesslog_str->ptr, p->conf.ts_accesslog_str->size - 1, "[%d/%b/%Y:%H:%M:%S ", localtime_r(&(srv->cur_ts)));
+# endif
+					p->conf.ts_accesslog_str->used = strlen(p->conf.ts_accesslog_str->ptr) + 1;
 
-					if (p->conf.append_tz_offset) {
-						buffer_append_string_len(p->conf.ts_accesslog_str, tm.tm_gmtoff >= 0 ? "+" : "-", 1);
+					buffer_append_string(p->conf.ts_accesslog_str, tm.tm_gmtoff >= 0 ? "+" : "-");
 
-						scd = abs(tm.tm_gmtoff);
-						hrs = scd / 3600;
-						min = (scd % 3600) / 60;
+					scd = abs(tm.tm_gmtoff);
+					hrs = scd / 3600;
+					min = (scd % 3600) / 60;
 
-						/* hours */
-						if (hrs < 10) buffer_append_string_len(p->conf.ts_accesslog_str, CONST_STR_LEN("0"));
-						buffer_append_int(p->conf.ts_accesslog_str, hrs);
+					/* hours */
+					if (hrs < 10) buffer_append_string(p->conf.ts_accesslog_str, "0");
+					buffer_append_long(p->conf.ts_accesslog_str, hrs);
 
-						if (min < 10) buffer_append_string_len(p->conf.ts_accesslog_str, CONST_STR_LEN("0"));
-						buffer_append_int(p->conf.ts_accesslog_str, min);
-						buffer_append_string_len(p->conf.ts_accesslog_str, CONST_STR_LEN("]"));
-					}
-#else /* HAVE_STRUCT_TM_GMTOFF */
-# ifdef HAVE_GMTIME_R
+					if (min < 10) buffer_append_string(p->conf.ts_accesslog_str, "0");
+					buffer_append_long(p->conf.ts_accesslog_str, min);
+					BUFFER_APPEND_STRING_CONST(p->conf.ts_accesslog_str, "]");
+#else
+#ifdef HAVE_GMTIME_R
 					gmtime_r(&(srv->cur_ts), &tm);
-					buffer_append_strftime(p->conf.ts_accesslog_str, p->conf.ts_accesslog_fmt_str->ptr, &tm);
-# else /* HAVE_GMTIME_R */
-					buffer_append_strftime(p->conf.ts_accesslog_str, p->conf.ts_accesslog_fmt_str->ptr, gmtime(&(srv->cur_ts)));
-# endif /* HAVE_GMTIME_R */
-#endif /* HAVE_STRUCT_TM_GMTOFF */
+					strftime(p->conf.ts_accesslog_str->ptr, p->conf.ts_accesslog_str->size - 1, "[%d/%b/%Y:%H:%M:%S +0000]", &tm);
+#else
+					strftime(p->conf.ts_accesslog_str->ptr, p->conf.ts_accesslog_str->size - 1, "[%d/%b/%Y:%H:%M:%S +0000]", gmtime(&(srv->cur_ts)));
+#endif
+					p->conf.ts_accesslog_str->used = strlen(p->conf.ts_accesslog_str->ptr) + 1;
+#endif
 
 					*(p->conf.last_generated_accesslog_ts_ptr) = srv->cur_ts;
 					newts = 1;
@@ -768,127 +698,107 @@ REQUESTDONE_FUNC(log_access_write) {
 				break;
 			case FORMAT_REMOTE_IDENT:
 				/* ident */
-				buffer_append_string_len(b, CONST_STR_LEN("-"));
+				BUFFER_APPEND_STRING_CONST(b, "-");
 				break;
 			case FORMAT_REMOTE_USER:
-				if (NULL != (ds = (data_string *)array_get_element(con->environment, "REMOTE_USER")) && !buffer_string_is_empty(ds->value)) {
-					accesslog_append_escaped(b, ds->value);
+				if (con->authed_user->used > 1) {
+					buffer_append_string_buffer(b, con->authed_user);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_REQUEST_LINE:
-				if (!buffer_string_is_empty(con->request.request_line)) {
-					accesslog_append_escaped(b, con->request.request_line);
-				}
+				buffer_append_string(b, get_http_method_name(con->request.http_method));
+				buffer_append_string(b, " ");
+				buffer_append_string_buffer(b, con->request.orig_uri);
+				buffer_append_string(b, " ");
+				buffer_append_string(b, get_http_version_name(con->request.http_version));
+
 				break;
 			case FORMAT_STATUS:
-				buffer_append_int(b, con->http_status);
+				buffer_append_long(b, con->http_status);
 				break;
 
 			case FORMAT_BYTES_OUT_NO_HEADER:
 				if (con->bytes_written > 0) {
-					buffer_append_int(b,
+					buffer_append_off_t(b,
 							    con->bytes_written - con->bytes_header <= 0 ? 0 : con->bytes_written - con->bytes_header);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_HEADER:
-				if (NULL != (ds = (data_string *)array_get_element(con->request.headers, p->conf.parsed_format->ptr[j]->string->ptr))) {
-					accesslog_append_escaped(b, ds->value);
+				if (NULL != (ds = (data_string *)array_get_element(con->request.headers, CONST_BUF_LEN(p->conf.parsed_format->ptr[j]->string)))) {
+					buffer_append_string_buffer(b, ds->value);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_RESPONSE_HEADER:
-				if (NULL != (ds = (data_string *)array_get_element(con->response.headers, p->conf.parsed_format->ptr[j]->string->ptr))) {
-					accesslog_append_escaped(b, ds->value);
+				if (NULL != (ds = (data_string *)array_get_element(con->response.headers, CONST_BUF_LEN(p->conf.parsed_format->ptr[j]->string)))) {
+					buffer_append_string_buffer(b, ds->value);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
-				}
-				break;
-			case FORMAT_ENV:
-				if (NULL != (ds = (data_string *)array_get_element(con->environment, p->conf.parsed_format->ptr[j]->string->ptr))) {
-					accesslog_append_escaped(b, ds->value);
-				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_FILENAME:
-				if (!buffer_string_is_empty(con->physical.path)) {
+				if (con->physical.path->used > 1) {
 					buffer_append_string_buffer(b, con->physical.path);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_BYTES_OUT:
 				if (con->bytes_written > 0) {
-					buffer_append_int(b, con->bytes_written);
+					buffer_append_off_t(b, con->bytes_written);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_BYTES_IN:
 				if (con->bytes_read > 0) {
-					buffer_append_int(b, con->bytes_read);
+					buffer_append_off_t(b, con->bytes_read);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_TIME_USED:
-				buffer_append_int(b, srv->cur_ts - con->request_start);
+				buffer_append_long(b, srv->cur_ts - con->request_start);
 				break;
 			case FORMAT_SERVER_NAME:
-				if (!buffer_string_is_empty(con->server_name)) {
+				if (con->server_name->used > 1) {
 					buffer_append_string_buffer(b, con->server_name);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_HTTP_HOST:
-				if (!buffer_string_is_empty(con->uri.authority)) {
-					accesslog_append_escaped(b, con->uri.authority);
+				if (con->uri.authority->used > 1) {
+					buffer_append_string_buffer(b, con->uri.authority);
 				} else {
-					buffer_append_string_len(b, CONST_STR_LEN("-"));
+					BUFFER_APPEND_STRING_CONST(b, "-");
 				}
 				break;
 			case FORMAT_REQUEST_PROTOCOL:
-				buffer_append_string_len(b,
-					con->request.http_version == HTTP_VERSION_1_1 ? "HTTP/1.1" : "HTTP/1.0", 8);
+				buffer_append_string(b,
+						     con->request.http_version == HTTP_VERSION_1_1 ? "HTTP/1.1" : "HTTP/1.0");
 				break;
 			case FORMAT_REQUEST_METHOD:
 				buffer_append_string(b, get_http_method_name(con->request.http_method));
 				break;
-			case FORMAT_PERCENT:
-				buffer_append_string_len(b, CONST_STR_LEN("%"));
-				break;
 			case FORMAT_SERVER_PORT:
-				{
-					const char *colon;
-					buffer *srvtoken = ((server_socket*)(con->srv_socket))->srv_token;
-					if (srvtoken->ptr[0] == '[') {
-						colon = strstr(srvtoken->ptr, "]:");
-					} else {
-						colon = strchr(srvtoken->ptr, ':');
-					}
-					if (colon) {
-						buffer_append_string(b, colon+1);
-					} else {
-						buffer_append_int(b, srv->srvconf.port);
-					}
-				}
+				buffer_append_long(b, srv->srvconf.port);
 				break;
 			case FORMAT_QUERY_STRING:
-				accesslog_append_escaped(b, con->uri.query);
+				buffer_append_string_buffer(b, con->uri.query);
 				break;
 			case FORMAT_URL:
-				accesslog_append_escaped(b, con->uri.path_raw);
+				buffer_append_string_buffer(b, con->uri.path_raw);
 				break;
 			case FORMAT_CONNECTION_STATUS:
 				switch(con->keep_alive) {
-				case 0: buffer_append_string_len(b, CONST_STR_LEN("-")); break;
-				default: buffer_append_string_len(b, CONST_STR_LEN("+")); break;
+				case 0: buffer_append_string(b, "-"); break;
+				default: buffer_append_string(b, "+"); break;
 				}
 				break;
 			default:
@@ -897,6 +807,7 @@ REQUESTDONE_FUNC(log_access_write) {
 				 { 'A', FORMAT_LOCAL_ADDR },
 				 { 'C', FORMAT_COOKIE },
 				 { 'D', FORMAT_TIME_USED_MS },
+				 { 'e', FORMAT_ENV },
 				 */
 
 				break;
@@ -907,22 +818,21 @@ REQUESTDONE_FUNC(log_access_write) {
 		}
 	}
 
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
 	if (p->conf.use_syslog ||  /* syslog doesn't cache */
-	    (!buffer_string_is_empty(p->conf.access_logfile) && p->conf.access_logfile->ptr[0] == '|') || /* pipes don't cache */
+	    (p->conf.access_logfile->used && p->conf.access_logfile->ptr[0] != '|') || /* pipes don't cache */
 	    newts ||
-	    buffer_string_length(b) >= BUFFER_MAX_REUSE_SIZE) {
+	    b->used > BUFFER_MAX_REUSE_SIZE) {
 		if (p->conf.use_syslog) {
 #ifdef HAVE_SYSLOG_H
-			if (!buffer_string_is_empty(b)) {
+			if (b->used > 2) {
 				/* syslog appends a \n on its own */
-				buffer_string_set_length(b, buffer_string_length(b) - 1);
-				syslog(p->conf.syslog_level, "%s", b->ptr);
+				syslog(LOG_INFO, "%*s", b->used - 2, b->ptr);
 			}
 #endif
 		} else if (p->conf.log_access_fd != -1) {
-			accesslog_write_all(srv, p->conf.access_logfile, p->conf.log_access_fd, CONST_BUF_LEN(b));
+			write(p->conf.log_access_fd, b->ptr, b->used - 1);
 		}
 		buffer_reset(b);
 	}
@@ -931,8 +841,7 @@ REQUESTDONE_FUNC(log_access_write) {
 }
 
 
-int mod_accesslog_plugin_init(plugin *p);
-int mod_accesslog_plugin_init(plugin *p) {
+LI_EXPORT int mod_accesslog_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("accesslog");
 
@@ -940,8 +849,8 @@ int mod_accesslog_plugin_init(plugin *p) {
 	p->set_defaults= log_access_open;
 	p->cleanup     = mod_accesslog_free;
 
-	p->handle_request_done  = log_access_write;
-	p->handle_sighup        = log_access_cycle;
+	p->handle_response_done  = log_access_write;
+	p->handle_sighup         = log_access_cycle;
 
 	p->data        = NULL;
 
