@@ -1,13 +1,14 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <stdio.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
 
 #include "plugin.h"
-
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <stdio.h>
+#include "sys-strings.h"
 
 /* plugin config for all request/connections */
 typedef struct {
@@ -26,6 +27,8 @@ typedef struct {
 INIT_FUNC(mod_alias_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 
@@ -45,8 +48,6 @@ FREE_FUNC(mod_alias_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
-
 			array_free(s->alias);
 
 			free(s);
@@ -72,10 +73,9 @@ SETDEFAULTS_FUNC(mod_alias_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
@@ -84,7 +84,7 @@ SETDEFAULTS_FUNC(mod_alias_set_defaults) {
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 		if (s->alias->used >= 2) {
@@ -96,16 +96,17 @@ SETDEFAULTS_FUNC(mod_alias_set_defaults) {
 				for (k = j + 1; k < a->used; k ++) {
 					const buffer *key = a->data[a->sorted[k]]->key;
 
-					if (buffer_string_length(key) < buffer_string_length(prefix)) {
+					if (key->used < prefix->used) {
 						break;
 					}
-					if (memcmp(key->ptr, prefix->ptr, buffer_string_length(prefix)) != 0) {
+					if (memcmp(key->ptr, prefix->ptr, prefix->used - 1) != 0) {
 						break;
 					}
 					/* ok, they have same prefix. check position */
 					if (a->sorted[j] < a->sorted[k]) {
-						log_error_write(srv, __FILE__, __LINE__, "SBSBS",
-							"url.alias: `", key, "' will never match as `", prefix, "' matched first");
+						fprintf(stderr, "url.alias: `%s' will never match as `%s' matched first\n",
+								key->ptr,
+								prefix->ptr);
 						return HANDLER_ERROR;
 					}
 				}
@@ -116,13 +117,11 @@ SETDEFAULTS_FUNC(mod_alias_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_alias_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(alias);
+	PATCH_OPTION(alias);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -137,14 +136,13 @@ static int mod_alias_patch_connection(server *srv, connection *con, plugin_data
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("alias.url"))) {
-				PATCH(alias);
+				PATCH_OPTION(alias);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 PHYSICALPATH_FUNC(mod_alias_physical_handler) {
 	plugin_data *p = p_d;
@@ -152,32 +150,31 @@ PHYSICALPATH_FUNC(mod_alias_physical_handler) {
 	char *uri_ptr;
 	size_t k;
 
-	if (buffer_is_empty(con->physical.path)) return HANDLER_GO_ON;
+	if (con->physical.path->used == 0) return HANDLER_GO_ON;
 
 	mod_alias_patch_connection(srv, con, p);
 
 	/* not to include the tailing slash */
-	basedir_len = buffer_string_length(con->physical.basedir);
-	if ('/' == con->physical.basedir->ptr[basedir_len-1]) --basedir_len;
-	uri_len = buffer_string_length(con->physical.path) - basedir_len;
+	basedir_len = (con->physical.basedir->used - 1) - 1;
+	uri_len = con->physical.path->used - 1 - basedir_len;
 	uri_ptr = con->physical.path->ptr + basedir_len;
 
 	for (k = 0; k < p->conf.alias->used; k++) {
 		data_string *ds = (data_string *)p->conf.alias->data[k];
-		int alias_len = buffer_string_length(ds->key);
+		int alias_len = ds->key->used - 1;
 
 		if (alias_len > uri_len) continue;
-		if (buffer_is_empty(ds->key)) continue;
+		if (ds->key->used == 0) continue;
 
 		if (0 == (con->conf.force_lowercase_filenames ?
 					strncasecmp(uri_ptr, ds->key->ptr, alias_len) :
 					strncmp(uri_ptr, ds->key->ptr, alias_len))) {
 			/* matched */
 
-			buffer_copy_buffer(con->physical.basedir, ds->value);
-			buffer_copy_buffer(srv->tmp_buf, ds->value);
+			buffer_copy_string_buffer(con->physical.basedir, ds->value);
+			buffer_copy_string_buffer(srv->tmp_buf, ds->value);
 			buffer_append_string(srv->tmp_buf, uri_ptr + alias_len);
-			buffer_copy_buffer(con->physical.path, srv->tmp_buf);
+			buffer_copy_string_buffer(con->physical.path, srv->tmp_buf);
 
 			return HANDLER_GO_ON;
 		}
@@ -189,8 +186,7 @@ PHYSICALPATH_FUNC(mod_alias_physical_handler) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_alias_plugin_init(plugin *p);
-int mod_alias_plugin_init(plugin *p) {
+LI_EXPORT int mod_alias_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("alias");
 
