@@ -1,3 +1,13 @@
+#include <sys/types.h>
+#include <sys/stat.h>
+
+#include <fcntl.h>
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <errno.h>
+#include <time.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -9,18 +19,6 @@
 #include "crc32.h"
 #include "etag.h"
 
-#include <sys/types.h>
-#include <sys/stat.h>
-
-#include <assert.h>
-#include <fcntl.h>
-#include <unistd.h>
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <errno.h>
-#include <time.h>
-
 #if defined HAVE_ZLIB_H && defined HAVE_LIBZ
 # define USE_ZLIB
 # include <zlib.h>
@@ -34,6 +32,7 @@
 #endif
 
 #include "sys-mmap.h"
+#include "sys-files.h"
 
 /* request: accept-encoding */
 #define HTTP_ACCEPT_ENCODING_IDENTITY BV(0)
@@ -41,18 +40,15 @@
 #define HTTP_ACCEPT_ENCODING_DEFLATE  BV(2)
 #define HTTP_ACCEPT_ENCODING_COMPRESS BV(3)
 #define HTTP_ACCEPT_ENCODING_BZIP2    BV(4)
-#define HTTP_ACCEPT_ENCODING_X_GZIP   BV(5)
-#define HTTP_ACCEPT_ENCODING_X_BZIP2  BV(6)
 
 #ifdef __WIN32
-# define mkdir(x,y) mkdir(x)
+#define mkdir(x,y) mkdir(x)
 #endif
 
 typedef struct {
 	buffer *compress_cache_dir;
 	array  *compress;
 	off_t   compress_max_filesize; /** max filesize in kb */
-	int     allowed_encodings;
 } plugin_config;
 
 typedef struct {
@@ -67,6 +63,8 @@ typedef struct {
 INIT_FUNC(mod_compress_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->ofn = buffer_init();
@@ -90,7 +88,7 @@ FREE_FUNC(mod_compress_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
+			if (!s) continue;
 
 			array_free(s->compress);
 			buffer_free(s->compress_cache_dir);
@@ -106,48 +104,26 @@ FREE_FUNC(mod_compress_free) {
 	return HANDLER_GO_ON;
 }
 
-/* 0 on success, -1 for error */
-static int mkdir_recursive(char *dir) {
-	char *p = dir;
+void mkdir_recursive(const char *dir) {
+
+	char dir_copy[256];
+	char *p = dir_copy;
 
 	if (!dir || !dir[0])
-		return 0;
+		return;
+
+	strncpy(dir_copy, dir, sizeof(dir_copy) / sizeof(dir_copy[0]));
 
 	while ((p = strchr(p + 1, '/')) != NULL) {
 
 		*p = '\0';
-		if ((mkdir(dir, 0700) != 0) && (errno != EEXIST)) {
-			*p = '/';
-			return -1;
-		}
+		if ((mkdir(dir_copy, 0700) != 0) && (errno != EEXIST))
+			return;
 
 		*p++ = '/';
-		if (!*p) return 0; /* Ignore trailing slash */
 	}
 
-	return (mkdir(dir, 0700) != 0) && (errno != EEXIST) ? -1 : 0;
-}
-
-/* 0 on success, -1 for error */
-static int mkdir_for_file(char *filename) {
-	char *p = filename;
-
-	if (!filename || !filename[0])
-		return -1;
-
-	while ((p = strchr(p + 1, '/')) != NULL) {
-
-		*p = '\0';
-		if ((mkdir(filename, 0700) != 0) && (errno != EEXIST)) {
-			*p = '/';
-			return -1;
-		}
-
-		*p++ = '/';
-		if (!*p) return -1; /* Unexpected trailing slash in filename */
-	}
-
-	return 0;
+	mkdir(dir, 0700);
 }
 
 SETDEFAULTS_FUNC(mod_compress_setdefaults) {
@@ -158,80 +134,44 @@ SETDEFAULTS_FUNC(mod_compress_setdefaults) {
 		{ "compress.cache-dir",             NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
 		{ "compress.filetype",              NULL, T_CONFIG_ARRAY, T_CONFIG_SCOPE_CONNECTION },
 		{ "compress.max-filesize",          NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },
-		{ "compress.allowed-encodings",     NULL, T_CONFIG_ARRAY, T_CONFIG_SCOPE_CONNECTION },
 		{ NULL,                             NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
 	};
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
-		array  *encodings_arr = array_init();
 
 		s = calloc(1, sizeof(plugin_config));
 		s->compress_cache_dir = buffer_init();
 		s->compress = array_init();
 		s->compress_max_filesize = 0;
-		s->allowed_encodings = 0;
 
 		cv[0].destination = s->compress_cache_dir;
 		cv[1].destination = s->compress;
 		cv[2].destination = &(s->compress_max_filesize);
-		cv[3].destination = encodings_arr; /* temp array for allowed encodings list */
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 
-		if (encodings_arr->used) {
-			size_t j = 0;
-			for (j = 0; j < encodings_arr->used; j++) {
-				data_string *ds = (data_string *)encodings_arr->data[j];
-#ifdef USE_ZLIB
-				if (NULL != strstr(ds->value->ptr, "gzip"))
-					s->allowed_encodings |= HTTP_ACCEPT_ENCODING_GZIP | HTTP_ACCEPT_ENCODING_X_GZIP;
-				if (NULL != strstr(ds->value->ptr, "x-gzip"))
-					s->allowed_encodings |= HTTP_ACCEPT_ENCODING_X_GZIP;
-				if (NULL != strstr(ds->value->ptr, "deflate"))
-					s->allowed_encodings |= HTTP_ACCEPT_ENCODING_DEFLATE;
-				/*
-				if (NULL != strstr(ds->value->ptr, "compress"))
-					s->allowed_encodings |= HTTP_ACCEPT_ENCODING_COMPRESS;
-				*/
-#endif
-#ifdef USE_BZ2LIB
-				if (NULL != strstr(ds->value->ptr, "bzip2"))
-					s->allowed_encodings |= HTTP_ACCEPT_ENCODING_BZIP2 | HTTP_ACCEPT_ENCODING_X_BZIP2;
-				if (NULL != strstr(ds->value->ptr, "x-bzip2"))
-					s->allowed_encodings |= HTTP_ACCEPT_ENCODING_X_BZIP2;
-#endif
-			}
-		} else {
-			/* default encodings */
-			s->allowed_encodings = 0
-#ifdef USE_ZLIB
-				| HTTP_ACCEPT_ENCODING_GZIP | HTTP_ACCEPT_ENCODING_X_GZIP | HTTP_ACCEPT_ENCODING_DEFLATE
-#endif
-#ifdef USE_BZ2LIB
-				| HTTP_ACCEPT_ENCODING_BZIP2 | HTTP_ACCEPT_ENCODING_X_BZIP2
-#endif
-				;
-		}
-
-		array_free(encodings_arr);
-
-		if (!buffer_string_is_empty(s->compress_cache_dir)) {
+		if (!buffer_is_empty(s->compress_cache_dir)) {
 			struct stat st;
-			mkdir_recursive(s->compress_cache_dir->ptr);
-
 			if (0 != stat(s->compress_cache_dir->ptr, &st)) {
-				log_error_write(srv, __FILE__, __LINE__, "sbs", "can't stat compress.cache-dir",
-						s->compress_cache_dir, strerror(errno));
 
-				return HANDLER_ERROR;
+				log_error_write(srv, __FILE__, __LINE__, "sbs", "can't stat compress.cache-dir, attempting to create",
+						s->compress_cache_dir, strerror(errno));
+				mkdir_recursive(s->compress_cache_dir->ptr);
+
+				if (0 != stat(s->compress_cache_dir->ptr, &st)) {
+
+					log_error_write(srv, __FILE__, __LINE__, "sbs", "can't stat compress.cache-dir, create failed",
+									s->compress_cache_dir, strerror(errno));
+
+					return HANDLER_ERROR;
+				}
 			}
 		}
 	}
@@ -245,7 +185,6 @@ static int deflate_file_to_buffer_gzip(server *srv, connection *con, plugin_data
 	unsigned char *c;
 	unsigned long crc;
 	z_stream z;
-	size_t outlen;
 
 	UNUSED(srv);
 	UNUSED(con);
@@ -268,7 +207,7 @@ static int deflate_file_to_buffer_gzip(server *srv, connection *con, plugin_data
 	z.total_in = 0;
 
 
-	buffer_string_prepare_copy(p->b, (z.avail_in * 1.1) + 12 + 18);
+	buffer_prepare_copy(p->b, (z.avail_in * 1.1) + 12 + 18);
 
 	/* write gzip header */
 
@@ -284,9 +223,9 @@ static int deflate_file_to_buffer_gzip(server *srv, connection *con, plugin_data
 	c[8] = 0x00; /* extra flags */
 	c[9] = 0x03; /* UNIX */
 
-	outlen = 10;
-	z.next_out = (unsigned char *)p->b->ptr + outlen;
-	z.avail_out = p->b->size - outlen - 9;
+	p->b->used = 10;
+	z.next_out = (unsigned char *)p->b->ptr + p->b->used;
+	z.avail_out = p->b->size - p->b->used - 8;
 	z.total_out = 0;
 
 	if (Z_STREAM_END != deflate(&z, Z_FINISH)) {
@@ -295,11 +234,11 @@ static int deflate_file_to_buffer_gzip(server *srv, connection *con, plugin_data
 	}
 
 	/* trailer */
-	outlen += z.total_out;
+	p->b->used += z.total_out;
 
 	crc = generate_crc32c(start, st_size);
 
-	c = (unsigned char *)p->b->ptr + outlen;
+	c = (unsigned char *)p->b->ptr + p->b->used;
 
 	c[0] = (crc >>  0) & 0xff;
 	c[1] = (crc >>  8) & 0xff;
@@ -309,8 +248,7 @@ static int deflate_file_to_buffer_gzip(server *srv, connection *con, plugin_data
 	c[5] = (z.total_in >>  8) & 0xff;
 	c[6] = (z.total_in >> 16) & 0xff;
 	c[7] = (z.total_in >> 24) & 0xff;
-	outlen += 8;
-	buffer_commit(p->b, outlen);
+	p->b->used += 8;
 
 	if (Z_OK != deflateEnd(&z)) {
 		return -1;
@@ -342,10 +280,10 @@ static int deflate_file_to_buffer_deflate(server *srv, connection *con, plugin_d
 	z.avail_in = st_size;
 	z.total_in = 0;
 
-	buffer_string_prepare_copy(p->b, (z.avail_in * 1.1) + 12);
+	buffer_prepare_copy(p->b, (z.avail_in * 1.1) + 12);
 
 	z.next_out = (unsigned char *)p->b->ptr;
-	z.avail_out = p->b->size - 1;
+	z.avail_out = p->b->size;
 	z.total_out = 0;
 
 	if (Z_STREAM_END != deflate(&z, Z_FINISH)) {
@@ -353,13 +291,13 @@ static int deflate_file_to_buffer_deflate(server *srv, connection *con, plugin_d
 		return -1;
 	}
 
+	/* trailer */
+	p->b->used += z.total_out;
+
 	if (Z_OK != deflateEnd(&z)) {
 		return -1;
 	}
 
-	/* trailer */
-	buffer_commit(p->b, z.total_out);
-
 	return 0;
 }
 
@@ -388,10 +326,10 @@ static int deflate_file_to_buffer_bzip2(server *srv, connection *con, plugin_dat
 	bz.total_in_lo32 = 0;
 	bz.total_in_hi32 = 0;
 
-	buffer_string_prepare_copy(p->b, (bz.avail_in * 1.1) + 12);
+	buffer_prepare_copy(p->b, (bz.avail_in * 1.1) + 12);
 
 	bz.next_out = p->b->ptr;
-	bz.avail_out = p->b->size - 1;
+	bz.avail_out = p->b->size;
 	bz.total_out_lo32 = 0;
 	bz.total_out_hi32 = 0;
 
@@ -400,15 +338,15 @@ static int deflate_file_to_buffer_bzip2(server *srv, connection *con, plugin_dat
 		return -1;
 	}
 
-	if (BZ_OK != BZ2_bzCompressEnd(&bz)) {
-		return -1;
-	}
-
 	/* file is too large for now */
 	if (bz.total_out_hi32) return -1;
 
 	/* trailer */
-	buffer_commit(p->b, bz.total_out_lo32);
+	p->b->used = bz.total_out_lo32;
+
+	if (BZ_OK != BZ2_bzCompressEnd(&bz)) {
+		return -1;
+	}
 
 	return 0;
 }
@@ -416,10 +354,13 @@ static int deflate_file_to_buffer_bzip2(server *srv, connection *con, plugin_dat
 
 static int deflate_file_to_file(server *srv, connection *con, plugin_data *p, buffer *fn, stat_cache_entry *sce, int type) {
 	int ifd, ofd;
-	int ret;
+	int ret = -1;
 	void *start;
 	const char *filename = fn->ptr;
 	ssize_t r;
+	stat_cache_entry *compressed_sce = NULL;
+
+	if (buffer_is_empty(p->conf.compress_cache_dir)) return -1;
 
 	/* overflow */
 	if ((off_t)(sce->st.st_size * 1.1) < sce->st.st_size) return -1;
@@ -432,26 +373,44 @@ static int deflate_file_to_file(server *srv, connection *con, plugin_data *p, bu
 	if (sce->st.st_size > 128 * 1024 * 1024) return -1;
 
 	buffer_reset(p->ofn);
-	buffer_copy_buffer(p->ofn, p->conf.compress_cache_dir);
-	buffer_append_slash(p->ofn);
+	buffer_copy_string_buffer(p->ofn, p->conf.compress_cache_dir);
+	PATHNAME_APPEND_SLASH(p->ofn);
 
-	if (0 == strncmp(con->physical.path->ptr, con->physical.doc_root->ptr, buffer_string_length(con->physical.doc_root))) {
-		buffer_append_string(p->ofn, con->physical.path->ptr + buffer_string_length(con->physical.doc_root));
+	if (0 == strncmp(con->physical.path->ptr, con->physical.doc_root->ptr, con->physical.doc_root->used-1)) {
+		size_t offset = p->ofn->used - 1;
+		char *dir, *nextdir;
+
+		buffer_append_string(p->ofn, con->physical.path->ptr + con->physical.doc_root->used - 1);
+
+		buffer_copy_string_buffer(p->b, p->ofn);
+
+		/* mkdir -p ... */
+		for (dir = p->b->ptr + offset; NULL != (nextdir = strchr(dir, '/')); dir = nextdir + 1) {
+			*nextdir = '\0';
+
+			if (-1 == mkdir(p->b->ptr, 0700)) {
+				if (errno != EEXIST) {
+					log_error_write(srv, __FILE__, __LINE__, "sbss", "creating cache-directory", p->b, "failed", strerror(errno));
+
+					return -1;
+				}
+			}
+
+			*nextdir = '/';
+		}
 	} else {
 		buffer_append_string_buffer(p->ofn, con->uri.path);
 	}
 
 	switch(type) {
 	case HTTP_ACCEPT_ENCODING_GZIP:
-	case HTTP_ACCEPT_ENCODING_X_GZIP:
-		buffer_append_string_len(p->ofn, CONST_STR_LEN("-gzip-"));
+		buffer_append_string(p->ofn, "-gzip-");
 		break;
 	case HTTP_ACCEPT_ENCODING_DEFLATE:
-		buffer_append_string_len(p->ofn, CONST_STR_LEN("-deflate-"));
+		buffer_append_string(p->ofn, "-deflate-");
 		break;
 	case HTTP_ACCEPT_ENCODING_BZIP2:
-	case HTTP_ACCEPT_ENCODING_X_BZIP2:
-		buffer_append_string_len(p->ofn, CONST_STR_LEN("-bzip2-"));
+		buffer_append_string(p->ofn, "-bzip2-");
 		break;
 	default:
 		log_error_write(srv, __FILE__, __LINE__, "sd", "unknown compression type", type);
@@ -460,79 +419,53 @@ static int deflate_file_to_file(server *srv, connection *con, plugin_data *p, bu
 
 	buffer_append_string_buffer(p->ofn, sce->etag);
 
-	if (-1 == mkdir_for_file(p->ofn->ptr)) {
-		log_error_write(srv, __FILE__, __LINE__, "sb", "couldn't create directory for file", p->ofn);
-		return -1;
+
+	if (HANDLER_ERROR != stat_cache_get_entry(srv, con, p->ofn, &compressed_sce)) {
+		/* file exists */
+
+		chunkqueue_append_file(con->send, p->ofn, 0, compressed_sce->st.st_size);
+		con->send->is_closed = 1;
+
+		return 0;
 	}
 
 	if (-1 == (ofd = open(p->ofn->ptr, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, 0600))) {
 		if (errno == EEXIST) {
 			/* cache-entry exists */
-#if 0
-			log_error_write(srv, __FILE__, __LINE__, "bs", p->ofn, "compress-cache hit");
-#endif
-			buffer_copy_buffer(con->physical.path, p->ofn);
 
-			return 0;
 		}
 
-		log_error_write(srv, __FILE__, __LINE__, "sbss", "creating cachefile", p->ofn, "failed", strerror(errno));
+		log_error_write(srv, __FILE__, __LINE__, "sbss",
+				"creating cachefile", p->ofn,
+				"failed", strerror(errno));
 
 		return -1;
 	}
-#if 0
-	log_error_write(srv, __FILE__, __LINE__, "bs", p->ofn, "compress-cache miss");
-#endif
+
 	if (-1 == (ifd = open(filename, O_RDONLY | O_BINARY))) {
-		log_error_write(srv, __FILE__, __LINE__, "sbss", "opening plain-file", fn, "failed", strerror(errno));
+		log_error_write(srv, __FILE__, __LINE__, "sbss",
+				"opening plain-file", fn,
+				"failed", strerror(errno));
 
 		close(ofd);
 
-		/* Remove the incomplete cache file, so that later hits aren't served from it */
-		if (-1 == unlink(p->ofn->ptr)) {
-			log_error_write(srv, __FILE__, __LINE__, "sbss", "unlinking incomplete cachefile", p->ofn, "failed:", strerror(errno));
-		}
-
 		return -1;
 	}
 
-#ifdef USE_MMAP
+
 	if (MAP_FAILED == (start = mmap(NULL, sce->st.st_size, PROT_READ, MAP_SHARED, ifd, 0))) {
-		log_error_write(srv, __FILE__, __LINE__, "sbss", "mmaping", fn, "failed", strerror(errno));
+		log_error_write(srv, __FILE__, __LINE__, "sbss",
+				"mmaping", fn,
+				"failed", strerror(errno));
 
 		close(ofd);
 		close(ifd);
-
-		/* Remove the incomplete cache file, so that later hits aren't served from it */
-		if (-1 == unlink(p->ofn->ptr)) {
-			log_error_write(srv, __FILE__, __LINE__, "sbss", "unlinking incomplete cachefile", p->ofn, "failed:", strerror(errno));
-		}
-
 		return -1;
 	}
-#else
-	start = malloc(sce->st.st_size);
-	if (NULL == start || sce->st.st_size != read(ifd, start, sce->st.st_size)) {
-		log_error_write(srv, __FILE__, __LINE__, "sbss", "reading", fn, "failed", strerror(errno));
 
-		close(ofd);
-		close(ifd);
-		free(start);
-
-		/* Remove the incomplete cache file, so that later hits aren't served from it */
-		if (-1 == unlink(p->ofn->ptr)) {
-			log_error_write(srv, __FILE__, __LINE__, "sbss", "unlinking incomplete cachefile", p->ofn, "failed:", strerror(errno));
-		}
-
-		return -1;
-	}
-#endif
-
-	ret = -1;
 	switch(type) {
 #ifdef USE_ZLIB
 	case HTTP_ACCEPT_ENCODING_GZIP:
-	case HTTP_ACCEPT_ENCODING_X_GZIP:
 		ret = deflate_file_to_buffer_gzip(srv, con, p, start, sce->st.st_size, sce->st.st_mtime);
 		break;
 	case HTTP_ACCEPT_ENCODING_DEFLATE:
@@ -541,42 +474,33 @@ static int deflate_file_to_file(server *srv, connection *con, plugin_data *p, bu
 #endif
 #ifdef USE_BZ2LIB
 	case HTTP_ACCEPT_ENCODING_BZIP2:
-	case HTTP_ACCEPT_ENCODING_X_BZIP2:
 		ret = deflate_file_to_buffer_bzip2(srv, con, p, start, sce->st.st_size);
 		break;
 #endif
+	default:
+		ret = -1;
+		break;
 	}
 
-	if (ret == 0) {
-		r = write(ofd, CONST_BUF_LEN(p->b));
-		if (-1 == r) {
-			log_error_write(srv, __FILE__, __LINE__, "sbss", "writing cachefile", p->ofn, "failed:", strerror(errno));
-			ret = -1;
-		} else if ((size_t)r != buffer_string_length(p->b)) {
-			log_error_write(srv, __FILE__, __LINE__, "sbs", "writing cachefile", p->ofn, "failed: not enough bytes written");
-			ret = -1;
-		}
-	}
-
-#ifdef USE_MMAP
-	munmap(start, sce->st.st_size);
-#else
-	free(start);
-#endif
-
-	close(ofd);
-	close(ifd);
-
-	if (ret != 0) {
-		/* Remove the incomplete cache file, so that later hits aren't served from it */
-		if (-1 == unlink(p->ofn->ptr)) {
-			log_error_write(srv, __FILE__, __LINE__, "sbss", "unlinking incomplete cachefile", p->ofn, "failed:", strerror(errno));
-		}
-
+	if (-1 == (r = write(ofd, p->b->ptr, p->b->used))) {
+		munmap(start, sce->st.st_size);
+		close(ofd);
+		close(ifd);
 		return -1;
 	}
 
-	buffer_copy_buffer(con->physical.path, p->ofn);
+	if ((size_t)r != p->b->used) {
+
+	}
+
+	munmap(start, sce->st.st_size);
+	close(ofd);
+	close(ifd);
+
+	if (ret != 0) return -1;
+
+	chunkqueue_append_file(con->send, p->ofn, 0, r);
+	con->send->is_closed = 1;
 
 	return 0;
 }
@@ -585,6 +509,7 @@ static int deflate_file_to_buffer(server *srv, connection *con, plugin_data *p,
 	int ifd;
 	int ret = -1;
 	void *start;
+	buffer *b;
 
 	/* overflow */
 	if ((off_t)(sce->st.st_size * 1.1) < sce->st.st_size) return -1;
@@ -596,35 +521,25 @@ static int deflate_file_to_buffer(server *srv, connection *con, plugin_data *p,
 
 	if (sce->st.st_size > 128 * 1024 * 1024) return -1;
 
-
 	if (-1 == (ifd = open(fn->ptr, O_RDONLY | O_BINARY))) {
 		log_error_write(srv, __FILE__, __LINE__, "sbss", "opening plain-file", fn, "failed", strerror(errno));
 
 		return -1;
 	}
 
-#ifdef USE_MMAP
-	if (MAP_FAILED == (start = mmap(NULL, sce->st.st_size, PROT_READ, MAP_SHARED, ifd, 0))) {
+	start = mmap(NULL, sce->st.st_size, PROT_READ, MAP_SHARED, ifd, 0);
+
+	close(ifd);
+
+	if (MAP_FAILED == start) {
 		log_error_write(srv, __FILE__, __LINE__, "sbss", "mmaping", fn, "failed", strerror(errno));
 
-		close(ifd);
 		return -1;
 	}
-#else
-	start = malloc(sce->st.st_size);
-	if (NULL == start || sce->st.st_size != read(ifd, start, sce->st.st_size)) {
-		log_error_write(srv, __FILE__, __LINE__, "sbss", "reading", fn, "failed", strerror(errno));
-
-		close(ifd);
-		free(start);
-		return -1;
-	}
-#endif
 
 	switch(type) {
 #ifdef USE_ZLIB
 	case HTTP_ACCEPT_ENCODING_GZIP:
-	case HTTP_ACCEPT_ENCODING_X_GZIP:
 		ret = deflate_file_to_buffer_gzip(srv, con, p, start, sce->st.st_size, sce->st.st_mtime);
 		break;
 	case HTTP_ACCEPT_ENCODING_DEFLATE:
@@ -633,7 +548,6 @@ static int deflate_file_to_buffer(server *srv, connection *con, plugin_data *p,
 #endif
 #ifdef USE_BZ2LIB
 	case HTTP_ACCEPT_ENCODING_BZIP2:
-	case HTTP_ACCEPT_ENCODING_X_BZIP2:
 		ret = deflate_file_to_buffer_bzip2(srv, con, p, start, sce->st.st_size);
 		break;
 #endif
@@ -642,37 +556,29 @@ static int deflate_file_to_buffer(server *srv, connection *con, plugin_data *p,
 		break;
 	}
 
-#ifdef USE_MMAP
 	munmap(start, sce->st.st_size);
-#else
-	free(start);
-#endif
-	close(ifd);
 
 	if (ret != 0) return -1;
 
-	chunkqueue_reset(con->write_queue);
-	chunkqueue_append_buffer(con->write_queue, p->b);
+	chunkqueue_reset(con->send);
+	b = chunkqueue_get_append_buffer(con->send);
+	buffer_copy_memory(b, p->b->ptr, p->b->used + 1);
 
 	buffer_reset(con->physical.path);
 
-	con->file_finished = 1;
+	con->send->is_closed = 1;
 	con->file_started  = 1;
 
 	return 0;
 }
 
-
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_compress_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(compress_cache_dir);
-	PATCH(compress);
-	PATCH(compress_max_filesize);
-	PATCH(allowed_encodings);
+	PATCH_OPTION(compress_cache_dir);
+	PATCH_OPTION(compress);
+	PATCH_OPTION(compress_max_filesize);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -687,44 +593,23 @@ static int mod_compress_patch_connection(server *srv, connection *con, plugin_da
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("compress.cache-dir"))) {
-				PATCH(compress_cache_dir);
+				PATCH_OPTION(compress_cache_dir);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("compress.filetype"))) {
-				PATCH(compress);
+				PATCH_OPTION(compress);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("compress.max-filesize"))) {
-				PATCH(compress_max_filesize);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("compress.allowed-encodings"))) {
-				PATCH(allowed_encodings);
+				PATCH_OPTION(compress_max_filesize);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
-
-static int mod_compress_contains_encoding(const char *headervalue, const char *encoding) {
-	const char *m;
-	for ( ;; ) {
-		m = strstr(headervalue, encoding);
-		if (NULL == m) return 0;
-		if (m == headervalue || m[-1] == ' ' || m[-1] == ',') return 1;
-
-		/* only partial match, search for next value */
-		m = strchr(m, ',');
-		if (NULL == m) return 0;
-		headervalue = m + 1;
-	}
-}
 
 PHYSICALPATH_FUNC(mod_compress_physical) {
 	plugin_data *p = p_d;
 	size_t m;
 	off_t max_fsize;
 	stat_cache_entry *sce = NULL;
-	buffer *mtime = NULL;
-	buffer *content_type;
-
-	if (con->mode != DIRECT || con->http_status) return HANDLER_GO_ON;
 
 	/* only GET and POST can get compressed */
 	if (con->request.http_method != HTTP_METHOD_GET &&
@@ -732,7 +617,7 @@ PHYSICALPATH_FUNC(mod_compress_physical) {
 		return HANDLER_GO_ON;
 	}
 
-	if (buffer_string_is_empty(con->physical.path)) {
+	if (buffer_is_empty(con->physical.path)) {
 		return HANDLER_GO_ON;
 	}
 
@@ -740,50 +625,17 @@ PHYSICALPATH_FUNC(mod_compress_physical) {
 
 	max_fsize = p->conf.compress_max_filesize;
 
-	if (con->conf.log_request_handling) {
-		log_error_write(srv, __FILE__, __LINE__,  "s",  "-- handling file as static file");
-	}
-
 	if (HANDLER_ERROR == stat_cache_get_entry(srv, con, con->physical.path, &sce)) {
-		con->http_status = 403;
-
-		log_error_write(srv, __FILE__, __LINE__, "sbsb",
-				"not a regular file:", con->uri.path,
-				"->", con->physical.path);
-
-		return HANDLER_FINISHED;
-	}
-
-	/* we only handle regular files */
-#ifdef HAVE_LSTAT
-	if ((sce->is_symlink == 1) && !con->conf.follow_symlink) {
-		return HANDLER_GO_ON;
-	}
-#endif
-	if (!S_ISREG(sce->st.st_mode)) {
 		return HANDLER_GO_ON;
 	}
 
 	/* don't compress files that are too large as we need to much time to handle them */
 	if (max_fsize && (sce->st.st_size >> 10) > max_fsize) return HANDLER_GO_ON;
 
-	/* don't try to compress files less than 128 bytes
-	 *
-	 * - extra overhead for compression
-	 * - mmap() fails for st_size = 0 :)
-	 */
+	/* compressing the file might lead to larger files instead */
 	if (sce->st.st_size < 128) return HANDLER_GO_ON;
 
 	/* check if mimetype is in compress-config */
-	content_type = NULL;
-	if (sce->content_type->ptr) {
-		char *c;
-		if ( (c = strchr(sce->content_type->ptr, ';')) != NULL) {
-			content_type = srv->tmp_buf;
-			buffer_copy_string_len(content_type, sce->content_type->ptr, c - sce->content_type->ptr);
-		}
-	}
-
 	for (m = 0; m < p->conf.compress->used; m++) {
 		data_string *compress_ds = (data_string *)p->conf.compress->data[m];
 
@@ -793,112 +645,88 @@ PHYSICALPATH_FUNC(mod_compress_physical) {
 			return HANDLER_GO_ON;
 		}
 
-		if (buffer_is_equal(compress_ds->value, sce->content_type)
-		    || (content_type && buffer_is_equal(compress_ds->value, content_type))) {
+		if (buffer_is_equal(compress_ds->value, sce->content_type)) {
 			/* mimetype found */
 			data_string *ds;
 
 			/* the response might change according to Accept-Encoding */
 			response_header_insert(srv, con, CONST_STR_LEN("Vary"), CONST_STR_LEN("Accept-Encoding"));
 
-			if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Accept-Encoding"))) {
+			if (NULL != (ds = (data_string *)array_get_element(con->request.headers, CONST_STR_LEN("Accept-Encoding")))) {
 				int accept_encoding = 0;
 				char *value = ds->value->ptr;
+				int srv_encodings = 0;
 				int matched_encodings = 0;
-				int use_etag = sce->etag != NULL && sce->etag->ptr != NULL;
 
 				/* get client side support encodings */
-#ifdef USE_ZLIB
-				if (mod_compress_contains_encoding(value, "gzip")) accept_encoding |= HTTP_ACCEPT_ENCODING_GZIP;
-				if (mod_compress_contains_encoding(value, "x-gzip")) accept_encoding |= HTTP_ACCEPT_ENCODING_X_GZIP;
-				if (mod_compress_contains_encoding(value, "deflate")) accept_encoding |= HTTP_ACCEPT_ENCODING_DEFLATE;
-				if (mod_compress_contains_encoding(value, "compress")) accept_encoding |= HTTP_ACCEPT_ENCODING_COMPRESS;
-#endif
+				if (NULL != strstr(value, "gzip")) accept_encoding |= HTTP_ACCEPT_ENCODING_GZIP;
+				if (NULL != strstr(value, "deflate")) accept_encoding |= HTTP_ACCEPT_ENCODING_DEFLATE;
+				if (NULL != strstr(value, "compress")) accept_encoding |= HTTP_ACCEPT_ENCODING_COMPRESS;
+				if (NULL != strstr(value, "bzip2")) accept_encoding |= HTTP_ACCEPT_ENCODING_BZIP2;
+				if (NULL != strstr(value, "identity")) accept_encoding |= HTTP_ACCEPT_ENCODING_IDENTITY;
+
+				/* get server side supported ones */
 #ifdef USE_BZ2LIB
-				if (mod_compress_contains_encoding(value, "bzip2")) accept_encoding |= HTTP_ACCEPT_ENCODING_BZIP2;
-				if (mod_compress_contains_encoding(value, "x-bzip2")) accept_encoding |= HTTP_ACCEPT_ENCODING_X_BZIP2;
+				srv_encodings |= HTTP_ACCEPT_ENCODING_BZIP2;
+#endif
+#ifdef USE_ZLIB
+				srv_encodings |= HTTP_ACCEPT_ENCODING_GZIP;
+				srv_encodings |= HTTP_ACCEPT_ENCODING_DEFLATE;
 #endif
-				if (mod_compress_contains_encoding(value, "identity")) accept_encoding |= HTTP_ACCEPT_ENCODING_IDENTITY;
 
 				/* find matching entries */
-				matched_encodings = accept_encoding & p->conf.allowed_encodings;
+				matched_encodings = accept_encoding & srv_encodings;
 
 				if (matched_encodings) {
-					static const char dflt_gzip[] = "gzip";
-					static const char dflt_x_gzip[] = "x-gzip";
-					static const char dflt_deflate[] = "deflate";
-					static const char dflt_bzip2[] = "bzip2";
-					static const char dflt_x_bzip2[] = "x-bzip2";
+					const char *dflt_gzip = "gzip";
+					const char *dflt_deflate = "deflate";
+					const char *dflt_bzip2 = "bzip2";
 
 					const char *compression_name = NULL;
 					int compression_type = 0;
+					buffer *mtime;
 
 					mtime = strftime_cache_get(srv, sce->st.st_mtime);
+					etag_mutate(con->physical.etag, sce->etag);
 
-					/* try matching original etag of uncompressed version */
-					if (use_etag) {
-						etag_mutate(con->physical.etag, sce->etag);
-						if (HANDLER_FINISHED == http_response_handle_cachable(srv, con, mtime)) {
-							response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(sce->content_type));
-							response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
-							response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
-							return HANDLER_FINISHED;
-						}
+					response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
+					response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
+
+					/* perhaps we don't even have to compress the file as the browser still has the
+					 * current version */
+					if (HANDLER_FINISHED == http_response_handle_cachable(srv, con, mtime)) {
+						return HANDLER_FINISHED;
 					}
 
 					/* select best matching encoding */
 					if (matched_encodings & HTTP_ACCEPT_ENCODING_BZIP2) {
 						compression_type = HTTP_ACCEPT_ENCODING_BZIP2;
 						compression_name = dflt_bzip2;
-					} else if (matched_encodings & HTTP_ACCEPT_ENCODING_X_BZIP2) {
-						compression_type = HTTP_ACCEPT_ENCODING_X_BZIP2;
-						compression_name = dflt_x_bzip2;
 					} else if (matched_encodings & HTTP_ACCEPT_ENCODING_GZIP) {
 						compression_type = HTTP_ACCEPT_ENCODING_GZIP;
 						compression_name = dflt_gzip;
-					} else if (matched_encodings & HTTP_ACCEPT_ENCODING_X_GZIP) {
-						compression_type = HTTP_ACCEPT_ENCODING_X_GZIP;
-						compression_name = dflt_x_gzip;
-					} else {
-						force_assert(matched_encodings & HTTP_ACCEPT_ENCODING_DEFLATE);
+					} else if (matched_encodings & HTTP_ACCEPT_ENCODING_DEFLATE) {
 						compression_type = HTTP_ACCEPT_ENCODING_DEFLATE;
 						compression_name = dflt_deflate;
 					}
 
-					if (use_etag) {
-						/* try matching etag of compressed version */
-						buffer_copy_buffer(srv->tmp_buf, sce->etag);
-						buffer_append_string_len(srv->tmp_buf, CONST_STR_LEN("-"));
-						buffer_append_string(srv->tmp_buf, compression_name);
-						etag_mutate(con->physical.etag, srv->tmp_buf);
-					}
+					/* deflate it to file (cached) or to memory */
+					if (0 == deflate_file_to_file(srv, con, p,
+							con->physical.path, sce, compression_type) ||
+					    0 == deflate_file_to_buffer(srv, con, p,
+							con->physical.path, sce, compression_type)) {
+
+						response_header_overwrite(srv, con,
+								CONST_STR_LEN("Content-Encoding"),
+								compression_name, strlen(compression_name));
+
+						response_header_overwrite(srv, con,
+								CONST_STR_LEN("Content-Type"),
+								CONST_BUF_LEN(sce->content_type));
 
-					if (HANDLER_FINISHED == http_response_handle_cachable(srv, con, mtime)) {
-						response_header_overwrite(srv, con, CONST_STR_LEN("Content-Encoding"), compression_name, strlen(compression_name));
-						response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(sce->content_type));
-						response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
-						if (use_etag) {
-							response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
-						}
 						return HANDLER_FINISHED;
 					}
-
-					/* deflate it */
-					if (use_etag && !buffer_string_is_empty(p->conf.compress_cache_dir)) {
-						if (0 != deflate_file_to_file(srv, con, p, con->physical.path, sce, compression_type))
-							return HANDLER_GO_ON;
-					} else {
-						if (0 != deflate_file_to_buffer(srv, con, p, con->physical.path, sce, compression_type))
-							return HANDLER_GO_ON;
-					}
-					response_header_overwrite(srv, con, CONST_STR_LEN("Content-Encoding"), compression_name, strlen(compression_name));
-					response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
-					if (use_etag) {
-						response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
-					}
-					response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(sce->content_type));
-					/* let mod_staticfile handle the cached compressed files, physical path was modified */
-					return (use_etag && !buffer_string_is_empty(p->conf.compress_cache_dir)) ? HANDLER_GO_ON : HANDLER_FINISHED;
+					break;
 				}
 			}
 		}
@@ -907,14 +735,16 @@ PHYSICALPATH_FUNC(mod_compress_physical) {
 	return HANDLER_GO_ON;
 }
 
-int mod_compress_plugin_init(plugin *p);
-int mod_compress_plugin_init(plugin *p) {
+LI_EXPORT int mod_compress_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("compress");
 
 	p->init        = mod_compress_init;
 	p->set_defaults = mod_compress_setdefaults;
-	p->handle_subrequest_start  = mod_compress_physical;
+
+	/* we have to hook into the response-header settings */
+	p->handle_response_header  = mod_compress_physical;
+
 	p->cleanup     = mod_compress_free;
 
 	p->data        = NULL;
