@@ -1,3 +1,11 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <assert.h>
+#include <errno.h>
+#include <stdio.h>
+#include <time.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -7,30 +15,25 @@
 #include "response.h"
 #include "stat_cache.h"
 #include "stream.h"
+#include "etag.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <dirent.h>
-#include <assert.h>
-#include <errno.h>
-#include <stdio.h>
-#include <unistd.h>
-#include <time.h>
+#include "sys-strings.h"
 
 /**
  * this is a dirlisting for a lighttpd plugin
  */
 
-#ifdef HAVE_ATTR_ATTRIBUTES_H
+
+#ifdef HAVE_SYS_SYSLIMITS_H
+#include <sys/syslimits.h>
+#endif
+
+#ifdef HAVE_XATTR
 #include <attr/attributes.h>
 #endif
 
-#ifdef HAVE_SYS_EXTATTR_H
-#include <sys/extattr.h>
-#endif
-
-#include "version.h"
+#include "sys-files.h"
+#include "sys-strings.h"
 
 /* plugin config for all request/connections */
 
@@ -53,17 +56,13 @@ typedef struct {
 	unsigned short hide_dot_files;
 	unsigned short show_readme;
 	unsigned short hide_readme_file;
-	unsigned short encode_readme;
 	unsigned short show_header;
 	unsigned short hide_header_file;
-	unsigned short encode_header;
-	unsigned short auto_layout;
 
 	excludes_buffer *excludes;
 
 	buffer *external_css;
 	buffer *encoding;
-	buffer *set_footer;
 } plugin_config;
 
 typedef struct {
@@ -71,13 +70,14 @@ typedef struct {
 
 	buffer *tmp_buf;
 	buffer *content_charset;
+	buffer *path;
 
 	plugin_config **config_storage;
 
 	plugin_config conf;
 } plugin_data;
 
-static excludes_buffer *excludes_buffer_init(void) {
+excludes_buffer *excludes_buffer_init(void) {
 	excludes_buffer *exb;
 
 	exb = calloc(1, sizeof(*exb));
@@ -85,7 +85,7 @@ static excludes_buffer *excludes_buffer_init(void) {
 	return exb;
 }
 
-static int excludes_buffer_append(excludes_buffer *exb, buffer *string) {
+int excludes_buffer_append(excludes_buffer *exb, buffer *string) {
 #ifdef HAVE_PCRE_H
 	size_t i;
 	const char *errptr;
@@ -119,7 +119,7 @@ static int excludes_buffer_append(excludes_buffer *exb, buffer *string) {
 	}
 
 	exb->ptr[exb->used]->string = buffer_init();
-	buffer_copy_buffer(exb->ptr[exb->used]->string, string);
+	buffer_copy_string_buffer(exb->ptr[exb->used]->string, string);
 
 	exb->used++;
 
@@ -132,7 +132,7 @@ static int excludes_buffer_append(excludes_buffer *exb, buffer *string) {
 #endif
 }
 
-static void excludes_buffer_free(excludes_buffer *exb) {
+void excludes_buffer_free(excludes_buffer *exb) {
 #ifdef HAVE_PCRE_H
 	size_t i;
 
@@ -152,15 +152,18 @@ static void excludes_buffer_free(excludes_buffer *exb) {
 INIT_FUNC(mod_dirlisting_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->tmp_buf = buffer_init();
 	p->content_charset = buffer_init();
+	p->path = buffer_init();
 
 	return p;
 }
 
-/* detroy the plugin data */
+/* destroy the plugin data */
 FREE_FUNC(mod_dirlisting_free) {
 	plugin_data *p = p_d;
 
@@ -178,7 +181,6 @@ FREE_FUNC(mod_dirlisting_free) {
 			excludes_buffer_free(s->excludes);
 			buffer_free(s->external_css);
 			buffer_free(s->encoding);
-			buffer_free(s->set_footer);
 
 			free(s);
 		}
@@ -186,6 +188,7 @@ FREE_FUNC(mod_dirlisting_free) {
 	}
 
 	buffer_free(p->tmp_buf);
+	buffer_free(p->path);
 	buffer_free(p->content_charset);
 
 	free(p);
@@ -193,55 +196,75 @@ FREE_FUNC(mod_dirlisting_free) {
 	return HANDLER_GO_ON;
 }
 
+static int parse_config_entry(server *srv, plugin_config *s, array *ca, const char *option) {
+	data_unset *du;
+
+	if (NULL != (du = array_get_element(ca, option, strlen(option)))) {
+		data_array *da = (data_array *)du;
+		size_t j;
+
+		if (du->type != TYPE_ARRAY) {
+			log_error_write(srv, __FILE__, __LINE__, "sss",
+				"unexpected type for key: ", option, "array of strings");
+
+			return HANDLER_ERROR;
+		}
+
+		da = (data_array *)du;
+
+		for (j = 0; j < da->value->used; j++) {
+			if (da->value->data[j]->type != TYPE_STRING) {
+				log_error_write(srv, __FILE__, __LINE__, "sssbs",
+					"unexpected type for key: ", option, "[",
+					da->value->data[j]->key, "](string)");
+
+				return HANDLER_ERROR;
+			}
+
+			if (0 != excludes_buffer_append(s->excludes,
+				    ((data_string *)(da->value->data[j]))->value)) {
+#ifdef HAVE_PCRE_H
+				log_error_write(srv, __FILE__, __LINE__, "sb",
+						"pcre-compile failed for", ((data_string *)(da->value->data[j]))->value);
+#else
+				log_error_write(srv, __FILE__, __LINE__, "s",
+						"pcre support is missing, please install libpcre and the headers");
+#endif
+			}
+		}
+	}
+
+	return 0;
+}
+
 /* handle plugin config and check values */
 
-#define CONFIG_EXCLUDE          "dir-listing.exclude"
-#define CONFIG_ACTIVATE         "dir-listing.activate"
-#define CONFIG_HIDE_DOTFILES    "dir-listing.hide-dotfiles"
-#define CONFIG_EXTERNAL_CSS     "dir-listing.external-css"
-#define CONFIG_ENCODING         "dir-listing.encoding"
-#define CONFIG_SHOW_README      "dir-listing.show-readme"
-#define CONFIG_HIDE_README_FILE "dir-listing.hide-readme-file"
-#define CONFIG_SHOW_HEADER      "dir-listing.show-header"
-#define CONFIG_HIDE_HEADER_FILE "dir-listing.hide-header-file"
-#define CONFIG_DIR_LISTING      "server.dir-listing"
-#define CONFIG_SET_FOOTER       "dir-listing.set-footer"
-#define CONFIG_ENCODE_README    "dir-listing.encode-readme"
-#define CONFIG_ENCODE_HEADER    "dir-listing.encode-header"
-#define CONFIG_AUTO_LAYOUT      "dir-listing.auto-layout"
-
-
 SETDEFAULTS_FUNC(mod_dirlisting_set_defaults) {
 	plugin_data *p = p_d;
 	size_t i = 0;
 
 	config_values_t cv[] = {
-		{ CONFIG_EXCLUDE,          NULL, T_CONFIG_LOCAL, T_CONFIG_SCOPE_CONNECTION },   /* 0 */
-		{ CONFIG_ACTIVATE,         NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 1 */
-		{ CONFIG_HIDE_DOTFILES,    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 2 */
-		{ CONFIG_EXTERNAL_CSS,     NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 3 */
-		{ CONFIG_ENCODING,         NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 4 */
-		{ CONFIG_SHOW_README,      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 5 */
-		{ CONFIG_HIDE_README_FILE, NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 6 */
-		{ CONFIG_SHOW_HEADER,      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 7 */
-		{ CONFIG_HIDE_HEADER_FILE, NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 8 */
-		{ CONFIG_DIR_LISTING,      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 9 */
-		{ CONFIG_SET_FOOTER,       NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 10 */
-		{ CONFIG_ENCODE_README,    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 11 */
-		{ CONFIG_ENCODE_HEADER,    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 12 */
-		{ CONFIG_AUTO_LAYOUT,      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 13 */
+		{ "dir-listing.exclude",	  NULL, T_CONFIG_LOCAL, T_CONFIG_SCOPE_CONNECTION },   /* 0 */
+		{ "dir-listing.activate",         NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 1 */
+		{ "dir-listing.hide-dotfiles",    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 2 */
+		{ "dir-listing.external-css",     NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 3 */
+		{ "dir-listing.encoding",         NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 4 */
+		{ "dir-listing.show-readme",      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 5 */
+		{ "dir-listing.hide-readme-file", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 6 */
+		{ "dir-listing.show-header",      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 7 */
+		{ "dir-listing.hide-header-file", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 8 */
+		{ "server.dir-listing",           NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 9 */
 
 		{ NULL,                          NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
 	};
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
-		data_unset *du_excludes;
+		array *ca;
 
 		s = calloc(1, sizeof(plugin_config));
 		s->excludes = excludes_buffer_init();
@@ -252,12 +275,7 @@ SETDEFAULTS_FUNC(mod_dirlisting_set_defaults) {
 		s->hide_readme_file = 0;
 		s->show_header = 0;
 		s->hide_header_file = 0;
-		s->encode_readme = 1;
-		s->encode_header = 1;
-		s->auto_layout = 1;
-
 		s->encoding = buffer_init();
-		s->set_footer = buffer_init();
 
 		cv[0].destination = s->excludes;
 		cv[1].destination = &(s->dir_listing);
@@ -269,78 +287,33 @@ SETDEFAULTS_FUNC(mod_dirlisting_set_defaults) {
 		cv[7].destination = &(s->show_header);
 		cv[8].destination = &(s->hide_header_file);
 		cv[9].destination = &(s->dir_listing); /* old name */
-		cv[10].destination = s->set_footer;
-		cv[11].destination = &(s->encode_readme);
-		cv[12].destination = &(s->encode_header);
-		cv[13].destination = &(s->auto_layout);
 
 		p->config_storage[i] = s;
+		ca = ((data_config *)srv->config_context->data[i])->value;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ca, cv)) {
 			return HANDLER_ERROR;
 		}
 
-		if (NULL != (du_excludes = array_get_element(config->value, CONFIG_EXCLUDE))) {
-			array *excludes_list;
-			size_t j;
-
-			if (du_excludes->type != TYPE_ARRAY) {
-				log_error_write(srv, __FILE__, __LINE__, "sss",
-					"unexpected type for key: ", CONFIG_EXCLUDE, "array of strings");
-				return HANDLER_ERROR;
-			}
-
-			excludes_list = ((data_array*)du_excludes)->value;
-
-#ifndef HAVE_PCRE_H
-			if (excludes_list->used > 0) {
-				log_error_write(srv, __FILE__, __LINE__, "sss",
-					"pcre support is missing for: ", CONFIG_EXCLUDE, ", please install libpcre and the headers");
-				return HANDLER_ERROR;
-			}
-#else
-			for (j = 0; j < excludes_list->used; j++) {
-				data_unset *du_exclude = excludes_list->data[j];
-
-				if (du_exclude->type != TYPE_STRING) {
-					log_error_write(srv, __FILE__, __LINE__, "sssbs",
-						"unexpected type for key: ", CONFIG_EXCLUDE, "[",
-						du_exclude->key, "](string)");
-					return HANDLER_ERROR;
-				}
-
-				if (0 != excludes_buffer_append(s->excludes, ((data_string*)(du_exclude))->value)) {
-					log_error_write(srv, __FILE__, __LINE__, "sb",
-						"pcre-compile failed for", ((data_string*)(du_exclude))->value);
-					return HANDLER_ERROR;
-				}
-			}
-#endif
-		}
+		parse_config_entry(srv, s, ca, "dir-listing.exclude");
 	}
 
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_dirlisting_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(dir_listing);
-	PATCH(external_css);
-	PATCH(hide_dot_files);
-	PATCH(encoding);
-	PATCH(show_readme);
-	PATCH(hide_readme_file);
-	PATCH(show_header);
-	PATCH(hide_header_file);
-	PATCH(excludes);
-	PATCH(set_footer);
-	PATCH(encode_readme);
-	PATCH(encode_header);
-	PATCH(auto_layout);
+	PATCH_OPTION(dir_listing);
+	PATCH_OPTION(external_css);
+	PATCH_OPTION(hide_dot_files);
+	PATCH_OPTION(encoding);
+	PATCH_OPTION(show_readme);
+	PATCH_OPTION(hide_readme_file);
+	PATCH_OPTION(show_header);
+	PATCH_OPTION(hide_header_file);
+	PATCH_OPTION(excludes);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -354,40 +327,31 @@ static int mod_dirlisting_patch_connection(server *srv, connection *con, plugin_
 		for (j = 0; j < dc->value->used; j++) {
 			data_unset *du = dc->value->data[j];
 
-			if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_ACTIVATE)) ||
-			    buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_DIR_LISTING))) {
-				PATCH(dir_listing);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_HIDE_DOTFILES))) {
-				PATCH(hide_dot_files);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_EXTERNAL_CSS))) {
-				PATCH(external_css);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_ENCODING))) {
-				PATCH(encoding);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_SHOW_README))) {
-				PATCH(show_readme);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_HIDE_README_FILE))) {
-				PATCH(hide_readme_file);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_SHOW_HEADER))) {
-				PATCH(show_header);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_HIDE_HEADER_FILE))) {
-				PATCH(hide_header_file);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_SET_FOOTER))) {
-				PATCH(set_footer);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_EXCLUDE))) {
-				PATCH(excludes);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_ENCODE_README))) {
-				PATCH(encode_readme);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_ENCODE_HEADER))) {
-				PATCH(encode_header);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN(CONFIG_AUTO_LAYOUT))) {
-				PATCH(auto_layout);
+			if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.activate")) ||
+			    buffer_is_equal_string(du->key, CONST_STR_LEN("server.dir-listing"))) {
+				PATCH_OPTION(dir_listing);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.hide-dotfiles"))) {
+				PATCH_OPTION(hide_dot_files);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.external-css"))) {
+				PATCH_OPTION(external_css);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.encoding"))) {
+				PATCH_OPTION(encoding);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.show-readme"))) {
+				PATCH_OPTION(show_readme);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.hide-readme-file"))) {
+				PATCH_OPTION(hide_readme_file);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.show-header"))) {
+				PATCH_OPTION(show_header);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.hide-header-file"))) {
+				PATCH_OPTION(hide_header_file);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("dir-listing.excludes"))) {
+				PATCH_OPTION(excludes);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 typedef struct {
 	size_t  namelen;
@@ -463,8 +427,7 @@ static int http_list_directory_sizefmt(char *buf, off_t size) {
 		u++;
 	}
 
-	li_itostr(out, size);
-	out += strlen(out);
+	out   += ltostr(out, size);
 	out[0] = '.';
 	out[1] = remain + '0';
 	out[2] = *u;
@@ -476,83 +439,81 @@ static int http_list_directory_sizefmt(char *buf, off_t size) {
 static void http_list_directory_header(server *srv, connection *con, plugin_data *p, buffer *out) {
 	UNUSED(srv);
 
-	if (p->conf.auto_layout) {
-		buffer_append_string_len(out, CONST_STR_LEN(
-			"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n"
-			"<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">\n"
-			"<head>\n"
-			"<title>Index of "
-		));
-		buffer_append_string_encoded(out, CONST_BUF_LEN(con->uri.path), ENCODING_MINIMAL_XML);
-		buffer_append_string_len(out, CONST_STR_LEN("</title>\n"));
+	BUFFER_APPEND_STRING_CONST(out,
+		"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n"
+		"<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">\n"
+		"<head>\n"
+		"<title>Index of "
+	);
+	buffer_append_string_encoded(out, CONST_BUF_LEN(con->uri.path), ENCODING_MINIMAL_XML);
+	BUFFER_APPEND_STRING_CONST(out, "</title>\n");
 
-		if (!buffer_string_is_empty(p->conf.external_css)) {
-			buffer_append_string_len(out, CONST_STR_LEN("<link rel=\"stylesheet\" type=\"text/css\" href=\""));
-			buffer_append_string_buffer(out, p->conf.external_css);
-			buffer_append_string_len(out, CONST_STR_LEN("\" />\n"));
-		} else {
-			buffer_append_string_len(out, CONST_STR_LEN(
-				"<style type=\"text/css\">\n"
-				"a, a:active {text-decoration: none; color: blue;}\n"
-				"a:visited {color: #48468F;}\n"
-				"a:hover, a:focus {text-decoration: underline; color: red;}\n"
-				"body {background-color: #F5F5F5;}\n"
-				"h2 {margin-bottom: 12px;}\n"
-				"table {margin-left: 12px;}\n"
-				"th, td {"
-				" font: 90% monospace;"
-				" text-align: left;"
-				"}\n"
-				"th {"
-				" font-weight: bold;"
-				" padding-right: 14px;"
-				" padding-bottom: 3px;"
-				"}\n"
-				"td {padding-right: 14px;}\n"
-				"td.s, th.s {text-align: right;}\n"
-				"div.list {"
-				" background-color: white;"
-				" border-top: 1px solid #646464;"
-				" border-bottom: 1px solid #646464;"
-				" padding-top: 10px;"
-				" padding-bottom: 14px;"
-				"}\n"
-				"div.foot {"
-				" font: 90% monospace;"
-				" color: #787878;"
-				" padding-top: 4px;"
-				"}\n"
-				"</style>\n"
-			));
-		}
-
-		buffer_append_string_len(out, CONST_STR_LEN("</head>\n<body>\n"));
+	if (p->conf.external_css->used > 1) {
+		BUFFER_APPEND_STRING_CONST(out, "<link rel=\"stylesheet\" type=\"text/css\" href=\"");
+		buffer_append_string_buffer(out, p->conf.external_css);
+		BUFFER_APPEND_STRING_CONST(out, "\" />\n");
+	} else {
+		BUFFER_APPEND_STRING_CONST(out,
+			"<style type=\"text/css\">\n"
+			"a, a:active {text-decoration: none; color: blue;}\n"
+			"a:visited {color: #48468F;}\n"
+			"a:hover, a:focus {text-decoration: underline; color: red;}\n"
+			"body {background-color: #F5F5F5;}\n"
+			"h2 {margin-bottom: 12px;}\n"
+			"table {margin-left: 12px;}\n"
+			"th, td {"
+			" font-family: \"Courier New\", Courier, monospace;"
+			" font-size: 10pt;"
+			" text-align: left;"
+			"}\n"
+			"th {"
+			" font-weight: bold;"
+			" padding-right: 14px;"
+			" padding-bottom: 3px;"
+			"}\n"
+		);
+		BUFFER_APPEND_STRING_CONST(out,
+			"td {padding-right: 14px;}\n"
+			"td.s, th.s {text-align: right;}\n"
+			"div.list {"
+			" background-color: white;"
+			" border-top: 1px solid #646464;"
+			" border-bottom: 1px solid #646464;"
+			" padding-top: 10px;"
+			" padding-bottom: 14px;"
+			"}\n"
+			"div.foot {"
+			" font-family: \"Courier New\", Courier, monospace;"
+			" font-size: 10pt;"
+			" color: #787878;"
+			" padding-top: 4px;"
+			"}\n"
+			"</style>\n"
+		);
 	}
 
+	BUFFER_APPEND_STRING_CONST(out, "</head>\n<body>\n");
+
 	/* HEADER.txt */
 	if (p->conf.show_header) {
 		stream s;
 		/* if we have a HEADER file, display it in <pre class="header"></pre> */
 
-		buffer_copy_buffer(p->tmp_buf, con->physical.path);
-		buffer_append_slash(p->tmp_buf);
-		buffer_append_string_len(p->tmp_buf, CONST_STR_LEN("HEADER.txt"));
+		buffer_copy_string_buffer(p->tmp_buf, con->physical.path);
+		PATHNAME_APPEND_SLASH(p->tmp_buf);
+		BUFFER_APPEND_STRING_CONST(p->tmp_buf, "HEADER.txt");
 
 		if (-1 != stream_open(&s, p->tmp_buf)) {
-			if (p->conf.encode_header) {
-				buffer_append_string_len(out, CONST_STR_LEN("<pre class=\"header\">"));
-				buffer_append_string_encoded(out, s.start, s.size, ENCODING_MINIMAL_XML);
-				buffer_append_string_len(out, CONST_STR_LEN("</pre>"));
-			} else {
-				buffer_append_string_len(out, s.start, s.size);
-			}
+			BUFFER_APPEND_STRING_CONST(out, "<pre class=\"header\">");
+			buffer_append_string_encoded(out, s.start, s.size, ENCODING_MINIMAL_XML);
+			BUFFER_APPEND_STRING_CONST(out, "</pre>");
 		}
 		stream_close(&s);
 	}
 
-	buffer_append_string_len(out, CONST_STR_LEN("<h2>Index of "));
+	BUFFER_APPEND_STRING_CONST(out, "<h2>Index of ");
 	buffer_append_string_encoded(out, CONST_BUF_LEN(con->uri.path), ENCODING_MINIMAL_XML);
-	buffer_append_string_len(out, CONST_STR_LEN(
+	BUFFER_APPEND_STRING_CONST(out,
 		"</h2>\n"
 		"<div class=\"list\">\n"
 		"<table summary=\"Directory Listing\" cellpadding=\"0\" cellspacing=\"0\">\n"
@@ -571,57 +532,49 @@ static void http_list_directory_header(server *srv, connection *con, plugin_data
 			"<td class=\"s\">- &nbsp;</td>"
 			"<td class=\"t\">Directory</td>"
 		"</tr>\n"
-	));
+	);
 }
 
 static void http_list_directory_footer(server *srv, connection *con, plugin_data *p, buffer *out) {
 	UNUSED(srv);
 
-	buffer_append_string_len(out, CONST_STR_LEN(
+	BUFFER_APPEND_STRING_CONST(out,
 		"</tbody>\n"
 		"</table>\n"
 		"</div>\n"
-	));
+	);
 
 	if (p->conf.show_readme) {
 		stream s;
 		/* if we have a README file, display it in <pre class="readme"></pre> */
 
-		buffer_copy_buffer(p->tmp_buf,  con->physical.path);
-		buffer_append_slash(p->tmp_buf);
-		buffer_append_string_len(p->tmp_buf, CONST_STR_LEN("README.txt"));
+		buffer_copy_string_buffer(p->tmp_buf,  con->physical.path);
+		PATHNAME_APPEND_SLASH(p->tmp_buf);
+		BUFFER_APPEND_STRING_CONST(p->tmp_buf, "README.txt");
 
 		if (-1 != stream_open(&s, p->tmp_buf)) {
-			if (p->conf.encode_readme) {
-				buffer_append_string_len(out, CONST_STR_LEN("<pre class=\"readme\">"));
-				buffer_append_string_encoded(out, s.start, s.size, ENCODING_MINIMAL_XML);
-				buffer_append_string_len(out, CONST_STR_LEN("</pre>"));
-			} else {
-				buffer_append_string_len(out, s.start, s.size);
-			}
+			BUFFER_APPEND_STRING_CONST(out, "<pre class=\"readme\">");
+			buffer_append_string_encoded(out, s.start, s.size, ENCODING_MINIMAL_XML);
+			BUFFER_APPEND_STRING_CONST(out, "</pre>");
 		}
 		stream_close(&s);
 	}
 
-	if(p->conf.auto_layout) {
-		buffer_append_string_len(out, CONST_STR_LEN(
-			"<div class=\"foot\">"
-		));
+	BUFFER_APPEND_STRING_CONST(out,
+		"<div class=\"foot\">"
+	);
 
-		if (!buffer_string_is_empty(p->conf.set_footer)) {
-			buffer_append_string_buffer(out, p->conf.set_footer);
-		} else if (buffer_is_empty(con->conf.server_tag)) {
-			buffer_append_string_len(out, CONST_STR_LEN(PACKAGE_DESC));
-		} else {
-			buffer_append_string_buffer(out, con->conf.server_tag);
-		}
-
-		buffer_append_string_len(out, CONST_STR_LEN(
-			"</div>\n"
-			"</body>\n"
-			"</html>\n"
-		));
+	if (buffer_is_empty(con->conf.server_tag)) {
+		BUFFER_APPEND_STRING_CONST(out, PACKAGE_NAME "/" PACKAGE_VERSION);
+	} else {
+		buffer_append_string_buffer(out, con->conf.server_tag);
 	}
+
+	BUFFER_APPEND_STRING_CONST(out,
+		"</div>\n"
+		"</body>\n"
+		"</html>\n"
+	);
 }
 
 static int http_list_directory(server *srv, connection *con, plugin_data *p, buffer *dir) {
@@ -629,7 +582,6 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 	buffer *out;
 	struct dirent *dent;
 	struct stat st;
-	char *path, *path_file;
 	size_t i;
 	int hide_dotfiles = p->conf.hide_dot_files;
 	dirls_list_t dirs, files, *list;
@@ -639,7 +591,8 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 	size_t k;
 	const char *content_type;
 	long name_max;
-#if defined(HAVE_XATTR) || defined(HAVE_EXTATTR)
+
+#ifdef HAVE_XATTR
 	char attrval[128];
 	int attrlen;
 #endif
@@ -647,44 +600,45 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 	struct tm tm;
 #endif
 
-	if (buffer_string_is_empty(dir)) return -1;
-
-	i = buffer_string_length(dir);
+	/* empty pathname, never ... */
+	if (buffer_is_empty(dir)) return -1;
 
+	/* max-length for the opendir */
 #ifdef HAVE_PATHCONF
-	if (0 >= (name_max = pathconf(dir->ptr, _PC_NAME_MAX))) {
-		/* some broken fs (fuse) return 0 instead of -1 */
+	if (-1 == (name_max = pathconf(dir->ptr, _PC_NAME_MAX))) {
 #ifdef NAME_MAX
 		name_max = NAME_MAX;
 #else
-		name_max = 255; /* stupid default */
+		name_max = 256; /* stupid default */
 #endif
 	}
-#elif defined __WIN32
+#elif defined _WIN32
 	name_max = FILENAME_MAX;
 #else
 	name_max = NAME_MAX;
 #endif
 
-	path = malloc(buffer_string_length(dir) + name_max + 1);
-	force_assert(NULL != path);
-	strcpy(path, dir->ptr);
-	path_file = path + i;
+	buffer_copy_string_buffer(p->path, dir);
+	PATHNAME_APPEND_SLASH(p->path);
 
-	if (NULL == (dp = opendir(path))) {
+#ifdef _WIN32
+	/* append *.* to the path */
+	buffer_append_string(p->path, "*.*");
+#endif
+
+	if (NULL == (dp = opendir(p->path->ptr))) {
 		log_error_write(srv, __FILE__, __LINE__, "sbs",
 			"opendir failed:", dir, strerror(errno));
 
-		free(path);
 		return -1;
 	}
 
 	dirs.ent   = (dirls_entry_t**) malloc(sizeof(dirls_entry_t*) * DIRLIST_BLOB_SIZE);
-	force_assert(dirs.ent);
+	assert(dirs.ent);
 	dirs.size  = DIRLIST_BLOB_SIZE;
 	dirs.used  = 0;
 	files.ent  = (dirls_entry_t**) malloc(sizeof(dirls_entry_t*) * DIRLIST_BLOB_SIZE);
-	force_assert(files.ent);
+	assert(files.ent);
 	files.size = DIRLIST_BLOB_SIZE;
 	files.used = 0;
 
@@ -725,11 +679,7 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 					log_error_write(srv, __FILE__, __LINE__, "sd",
 						"execution error while matching:", n);
 
-					/* aborting would require a lot of manual cleanup here.
-					 * skip instead (to not leak names that break pcre matching)
-					 */
-					exclude_match = 1;
-					break;
+					return -1;
 				}
 			}
 			else {
@@ -750,9 +700,15 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 		 */
 		if (i > (size_t)name_max) continue;
 
-		memcpy(path_file, dent->d_name, i + 1);
-		if (stat(path, &st) != 0)
+		/* build the dirname */
+		buffer_copy_string_buffer(p->path, dir);
+		PATHNAME_APPEND_SLASH(p->path);
+		buffer_append_string(p->path, dent->d_name);
+
+		if (stat(p->path->ptr, &st) != 0) {
+			fprintf(stderr, "%s.%d: %s, %s\r\n", __FILE__, __LINE__, p->path->ptr, strerror(errno));
 			continue;
+		}
 
 		list = &files;
 		if (S_ISDIR(st.st_mode))
@@ -761,7 +717,7 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 		if (list->used == list->size) {
 			list->size += DIRLIST_BLOB_SIZE;
 			list->ent   = (dirls_entry_t**) realloc(list->ent, sizeof(dirls_entry_t*) * list->size);
-			force_assert(list->ent);
+			assert(list->ent);
 		}
 
 		tmp = (dirls_entry_t*) malloc(sizeof(dirls_entry_t) + 1 + i);
@@ -778,14 +734,14 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 
 	if (files.used) http_dirls_sort(files.ent, files.used);
 
-	out = buffer_init();
-	buffer_copy_string_len(out, CONST_STR_LEN("<?xml version=\"1.0\" encoding=\""));
-	if (buffer_string_is_empty(p->conf.encoding)) {
-		buffer_append_string_len(out, CONST_STR_LEN("iso-8859-1"));
+	out = chunkqueue_get_append_buffer(con->send);
+	BUFFER_COPY_STRING_CONST(out, "<?xml version=\"1.0\" encoding=\"");
+	if (buffer_is_empty(p->conf.encoding)) {
+		BUFFER_APPEND_STRING_CONST(out, "iso-8859-1");
 	} else {
 		buffer_append_string_buffer(out, p->conf.encoding);
 	}
-	buffer_append_string_len(out, CONST_STR_LEN("\"?>\n"));
+	BUFFER_APPEND_STRING_CONST(out, "\"?>\n");
 	http_list_directory_header(srv, con, p, out);
 
 	/* directories */
@@ -799,13 +755,13 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 		strftime(datebuf, sizeof(datebuf), "%Y-%b-%d %H:%M:%S", localtime(&(tmp->mtime)));
 #endif
 
-		buffer_append_string_len(out, CONST_STR_LEN("<tr><td class=\"n\"><a href=\""));
+		BUFFER_APPEND_STRING_CONST(out, "<tr><td class=\"n\"><a href=\"");
 		buffer_append_string_encoded(out, DIRLIST_ENT_NAME(tmp), tmp->namelen, ENCODING_REL_URI_PART);
-		buffer_append_string_len(out, CONST_STR_LEN("/\">"));
+		BUFFER_APPEND_STRING_CONST(out, "/\">");
 		buffer_append_string_encoded(out, DIRLIST_ENT_NAME(tmp), tmp->namelen, ENCODING_MINIMAL_XML);
-		buffer_append_string_len(out, CONST_STR_LEN("</a>/</td><td class=\"m\">"));
+		BUFFER_APPEND_STRING_CONST(out, "</a>/</td><td class=\"m\">");
 		buffer_append_string_len(out, datebuf, sizeof(datebuf) - 1);
-		buffer_append_string_len(out, CONST_STR_LEN("</td><td class=\"s\">- &nbsp;</td><td class=\"t\">Directory</td></tr>\n"));
+		BUFFER_APPEND_STRING_CONST(out, "</td><td class=\"s\">- &nbsp;</td><td class=\"t\">Directory</td></tr>\n");
 
 		free(tmp);
 	}
@@ -815,19 +771,16 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 		tmp = files.ent[i];
 
 		content_type = NULL;
-#if defined(HAVE_XATTR)
+
+#ifdef HAVE_XATTR
 		if (con->conf.use_xattr) {
-			memcpy(path_file, DIRLIST_ENT_NAME(tmp), tmp->namelen + 1);
+			/* build the dirname */
+			buffer_copy_string_buffer(p->path, dir);
+			PATHNAME_APPEND_SLASH(p->path);
+			buffer_append_string_len(p->path, DIRLIST_ENT_NAME(tmp), tmp->namelen);
+
 			attrlen = sizeof(attrval) - 1;
-			if (attr_get(path, "Content-Type", attrval, &attrlen, 0) == 0) {
-				attrval[attrlen] = '\0';
-				content_type = attrval;
-			}
-		}
-#elif defined(HAVE_EXTATTR)
-		if (con->conf.use_xattr) {
-			memcpy(path_file, DIRLIST_ENT_NAME(tmp), tmp->namelen + 1);
-			if(-1 != (attrlen = extattr_get_file(path, EXTATTR_NAMESPACE_USER, "Content-Type", attrval, sizeof(attrval)-1))) {
+			if (attr_get(p->path->ptr, "Content-Type", attrval, &attrlen, 0) == 0) {
 				attrval[attrlen] = '\0';
 				content_type = attrval;
 			}
@@ -840,10 +793,10 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 				data_string *ds = (data_string *)con->conf.mimetypes->data[k];
 				size_t ct_len;
 
-				if (buffer_is_empty(ds->key))
+				if (ds->key->used == 0)
 					continue;
 
-				ct_len = buffer_string_length(ds->key);
+				ct_len = ds->key->used - 1;
 				if (tmp->namelen < ct_len)
 					continue;
 
@@ -862,39 +815,38 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 #endif
 		http_list_directory_sizefmt(sizebuf, tmp->size);
 
-		buffer_append_string_len(out, CONST_STR_LEN("<tr><td class=\"n\"><a href=\""));
+		BUFFER_APPEND_STRING_CONST(out, "<tr><td class=\"n\"><a href=\"");
 		buffer_append_string_encoded(out, DIRLIST_ENT_NAME(tmp), tmp->namelen, ENCODING_REL_URI_PART);
-		buffer_append_string_len(out, CONST_STR_LEN("\">"));
+		BUFFER_APPEND_STRING_CONST(out, "\">");
 		buffer_append_string_encoded(out, DIRLIST_ENT_NAME(tmp), tmp->namelen, ENCODING_MINIMAL_XML);
-		buffer_append_string_len(out, CONST_STR_LEN("</a></td><td class=\"m\">"));
+		BUFFER_APPEND_STRING_CONST(out, "</a></td><td class=\"m\">");
 		buffer_append_string_len(out, datebuf, sizeof(datebuf) - 1);
-		buffer_append_string_len(out, CONST_STR_LEN("</td><td class=\"s\">"));
+		BUFFER_APPEND_STRING_CONST(out, "</td><td class=\"s\">");
 		buffer_append_string(out, sizebuf);
-		buffer_append_string_len(out, CONST_STR_LEN("</td><td class=\"t\">"));
+		BUFFER_APPEND_STRING_CONST(out, "</td><td class=\"t\">");
 		buffer_append_string(out, content_type);
-		buffer_append_string_len(out, CONST_STR_LEN("</td></tr>\n"));
+		BUFFER_APPEND_STRING_CONST(out, "</td></tr>\n");
 
 		free(tmp);
 	}
 
 	free(files.ent);
 	free(dirs.ent);
-	free(path);
 
 	http_list_directory_footer(srv, con, p, out);
 
 	/* Insert possible charset to Content-Type */
-	if (buffer_string_is_empty(p->conf.encoding)) {
-		response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
+	if (buffer_is_empty(p->conf.encoding)) {
+		response_header_insert(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
 	} else {
-		buffer_copy_string_len(p->content_charset, CONST_STR_LEN("text/html; charset="));
+		buffer_copy_string(p->content_charset, "text/html; charset=");
 		buffer_append_string_buffer(p->content_charset, p->conf.encoding);
-		response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(p->content_charset));
+		response_header_insert(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(p->content_charset));
 	}
 
-	con->file_finished = 1;
-	chunkqueue_append_buffer(con->write_queue, out);
-	buffer_free(out);
+	con->send->bytes_in += out->used - 1;
+
+	con->send->is_closed = 1;
 
 	return 0;
 }
@@ -904,40 +856,47 @@ static int http_list_directory(server *srv, connection *con, plugin_data *p, buf
 URIHANDLER_FUNC(mod_dirlisting_subrequest) {
 	plugin_data *p = p_d;
 	stat_cache_entry *sce = NULL;
+	buffer *mtime;
+	data_string *ds;
 
-	UNUSED(srv);
-
-	/* we only handle GET, POST and HEAD */
-	switch(con->request.http_method) {
-	case HTTP_METHOD_GET:
-	case HTTP_METHOD_POST:
-	case HTTP_METHOD_HEAD:
-		break;
-	default:
-		return HANDLER_GO_ON;
-	}
-
-	if (con->mode != DIRECT) return HANDLER_GO_ON;
-
-	if (buffer_is_empty(con->physical.path)) return HANDLER_GO_ON;
-	if (buffer_is_empty(con->uri.path)) return HANDLER_GO_ON;
-	if (con->uri.path->ptr[buffer_string_length(con->uri.path) - 1] != '/') return HANDLER_GO_ON;
+	if (con->uri.path->used < 2) return HANDLER_GO_ON;
+	if (con->uri.path->ptr[con->uri.path->used - 2] != '/') return HANDLER_GO_ON;
+	if (con->physical.path->used == 0) return HANDLER_GO_ON;
 
 	mod_dirlisting_patch_connection(srv, con, p);
 
 	if (!p->conf.dir_listing) return HANDLER_GO_ON;
 
+	if (HANDLER_ERROR == stat_cache_get_entry(srv, con, con->physical.path, &sce)) {
+		/* just a second ago the file was still there */
+		return HANDLER_GO_ON;
+	}
+
+	if (!S_ISDIR(sce->st.st_mode)) return HANDLER_GO_ON;
+
 	if (con->conf.log_request_handling) {
 		log_error_write(srv, __FILE__, __LINE__,  "s",  "-- handling the request as Dir-Listing");
 		log_error_write(srv, __FILE__, __LINE__,  "sb", "URI          :", con->uri.path);
 	}
 
-	if (HANDLER_ERROR == stat_cache_get_entry(srv, con, con->physical.path, &sce)) {
-		log_error_write(srv, __FILE__, __LINE__,  "SB", "stat_cache_get_entry failed: ", con->physical.path);
-		SEGFAULT();
+	/* perhaps this a cachable request
+	 * - we use the etag of the directory
+	 * */
+
+	etag_mutate(con->physical.etag, sce->etag);
+	response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
+
+	/* prepare header */
+	if (NULL == (ds = (data_string *)array_get_element(con->response.headers, CONST_STR_LEN("Last-Modified")))) {
+		mtime = strftime_cache_get(srv, sce->st.st_mtime);
+		response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
+	} else {
+		mtime = ds->value;
 	}
 
-	if (!S_ISDIR(sce->st.st_mode)) return HANDLER_GO_ON;
+	if (HANDLER_FINISHED == http_response_handle_cachable(srv, con, mtime)) {
+		return HANDLER_FINISHED;
+	}
 
 	if (http_list_directory(srv, con, p, con->physical.path)) {
 		/* dirlisting failed */
@@ -952,13 +911,12 @@ URIHANDLER_FUNC(mod_dirlisting_subrequest) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_dirlisting_plugin_init(plugin *p);
-int mod_dirlisting_plugin_init(plugin *p) {
+LI_EXPORT int mod_dirlisting_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("dirlisting");
 
 	p->init        = mod_dirlisting_init;
-	p->handle_subrequest_start  = mod_dirlisting_subrequest;
+	p->handle_start_backend  = mod_dirlisting_subrequest;
 	p->set_defaults  = mod_dirlisting_set_defaults;
 	p->cleanup     = mod_dirlisting_free;
 
