@@ -1,3 +1,8 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <time.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -6,11 +11,6 @@
 #include "plugin.h"
 #include "stat_cache.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <time.h>
-
 /**
  * this is a expire module for a lighttpd
  *
@@ -39,6 +39,8 @@ typedef struct {
 INIT_FUNC(mod_expire_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->expire_tstmp = buffer_init();
@@ -62,9 +64,9 @@ FREE_FUNC(mod_expire_free) {
 		size_t i;
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
-			if (!s) continue;
 
 			array_free(s->expire_url);
+
 			free(s);
 		}
 		free(p->config_storage);
@@ -75,10 +77,10 @@ FREE_FUNC(mod_expire_free) {
 	return HANDLER_GO_ON;
 }
 
-static int mod_expire_get_offset(server *srv, plugin_data *p, buffer *expire, time_t *offset) {
+static int mod_expire_get_offset(server *srv, plugin_data *p, buffer *expire, int *offset) {
 	char *ts;
 	int type = -1;
-	time_t retts = 0;
+	int retts = 0;
 
 	UNUSED(p);
 
@@ -221,7 +223,7 @@ SETDEFAULTS_FUNC(mod_expire_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
@@ -253,13 +255,11 @@ SETDEFAULTS_FUNC(mod_expire_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_expire_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(expire_url);
+	PATCH_OPTION(expire_url);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -274,14 +274,13 @@ static int mod_expire_patch_connection(server *srv, connection *con, plugin_data
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("expire.url"))) {
-				PATCH(expire_url);
+				PATCH_OPTION(expire_url);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 URIHANDLER_FUNC(mod_expire_path_handler) {
 	plugin_data *p = p_d;
@@ -302,41 +301,35 @@ URIHANDLER_FUNC(mod_expire_path_handler) {
 		if (ds->key->used == 0) continue;
 
 		if (0 == strncmp(con->uri.path->ptr, ds->key->ptr, ct_len)) {
-			time_t ts, expires;
+			int ts;
+			time_t t;
 			size_t len;
 			stat_cache_entry *sce = NULL;
 
-			/* if stat fails => sce == NULL, ignore return value */
-			(void) stat_cache_get_entry(srv, con, con->physical.path, &sce);
+			stat_cache_get_entry(srv, con, con->physical.path, &sce);
 
 			switch(mod_expire_get_offset(srv, p, ds->value, &ts)) {
 			case 0:
 				/* access */
-				expires = (ts + srv->cur_ts);
+				t = (ts + srv->cur_ts);
 				break;
 			case 1:
 				/* modification */
 
-				/* can't set modification based expire header if
-				 * mtime is not available
-				 */
-				if (NULL == sce) return HANDLER_GO_ON;
-
-				expires = (ts + sce->st.st_mtime);
+				t = (ts + sce->st.st_mtime);
 				break;
 			default:
 				/* -1 is handled at parse-time */
-				return HANDLER_ERROR;
+				break;
 			}
 
-			/* expires should be at least srv->cur_ts */
-			if (expires < srv->cur_ts) expires = srv->cur_ts;
 
 			if (0 == (len = strftime(p->expire_tstmp->ptr, p->expire_tstmp->size - 1,
-					   "%a, %d %b %Y %H:%M:%S GMT", gmtime(&(expires))))) {
+					   "%a, %d %b %Y %H:%M:%S GMT", gmtime(&(t))))) {
 				/* could not set expire header, out of mem */
 
 				return HANDLER_GO_ON;
+
 			}
 
 			p->expire_tstmp->used = len + 1;
@@ -345,10 +338,10 @@ URIHANDLER_FUNC(mod_expire_path_handler) {
 			response_header_overwrite(srv, con, CONST_STR_LEN("Expires"), CONST_BUF_LEN(p->expire_tstmp));
 
 			/* HTTP/1.1 */
-			buffer_copy_string_len(p->expire_tstmp, CONST_STR_LEN("max-age="));
-			buffer_append_long(p->expire_tstmp, expires - srv->cur_ts); /* as expires >= srv->cur_ts the difference is >= 0 */
+			buffer_copy_string(p->expire_tstmp, "max-age=");
+			buffer_append_long(p->expire_tstmp, ts);
 
-			response_header_append(srv, con, CONST_STR_LEN("Cache-Control"), CONST_BUF_LEN(p->expire_tstmp));
+			response_header_overwrite(srv, con, CONST_STR_LEN("Cache-Control"), CONST_BUF_LEN(p->expire_tstmp));
 
 			return HANDLER_GO_ON;
 		}
@@ -360,13 +353,12 @@ URIHANDLER_FUNC(mod_expire_path_handler) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_expire_plugin_init(plugin *p);
-int mod_expire_plugin_init(plugin *p) {
+LI_EXPORT int mod_expire_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("expire");
 
 	p->init        = mod_expire_init;
-	p->handle_subrequest_start = mod_expire_path_handler;
+	p->handle_response_header = mod_expire_path_handler;
 	p->set_defaults  = mod_expire_set_defaults;
 	p->cleanup     = mod_expire_free;
 
