@@ -1,15 +1,18 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
 #include "response.h"
-#include "http_chunk.h"
 #include "stat_cache.h"
 
 #include "plugin.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 /* plugin config for all request/connections */
 
@@ -32,6 +35,8 @@ typedef struct {
 INIT_FUNC(mod_flv_streaming_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->query_str = buffer_init();
@@ -84,7 +89,7 @@ SETDEFAULTS_FUNC(mod_flv_streaming_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
@@ -104,13 +109,11 @@ SETDEFAULTS_FUNC(mod_flv_streaming_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_flv_streaming_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(extensions);
+	PATCH_OPTION(extensions);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -125,14 +128,13 @@ static int mod_flv_streaming_patch_connection(server *srv, connection *con, plug
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("flv-streaming.extensions"))) {
-				PATCH(extensions);
+				PATCH_OPTION(extensions);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 static int split_get_params(array *get_params, buffer *qrystr) {
 	size_t is_key = 1;
@@ -189,8 +191,6 @@ URIHANDLER_FUNC(mod_flv_streaming_path_handler) {
 
 	UNUSED(srv);
 
-	if (con->mode != DIRECT) return HANDLER_GO_ON;
-
 	if (buffer_is_empty(con->physical.path)) return HANDLER_GO_ON;
 
 	mod_flv_streaming_patch_connection(srv, con, p);
@@ -217,7 +217,7 @@ URIHANDLER_FUNC(mod_flv_streaming_path_handler) {
 			buffer_copy_string_buffer(p->query_str, con->uri.query);
 			split_get_params(p->get_params, p->query_str);
 
-			if (NULL == (get_param = (data_string *)array_get_element(p->get_params, "start"))) {
+			if (NULL == (get_param = (data_string *)array_get_element(p->get_params, CONST_STR_LEN("start")))) {
 				return HANDLER_GO_ON;
 			}
 
@@ -242,14 +242,14 @@ URIHANDLER_FUNC(mod_flv_streaming_path_handler) {
 			}
 
 			/* we are safe now, let's build a flv header */
-			b = chunkqueue_get_append_buffer(con->write_queue);
-			buffer_copy_string_len(b, CONST_STR_LEN("FLV\x1\x1\0\0\0\x9\0\0\0\x9"));
+			b = chunkqueue_get_append_buffer(con->send);
+			BUFFER_COPY_STRING_CONST(b, "FLV\x1\x1\0\0\0\x9\0\0\0\x9");
 
-			http_chunk_append_file(srv, con, con->physical.path, start, sce->st.st_size - start);
+			chunkqueue_append_file(con->send, con->physical.path, start, sce->st.st_size - start);
 
 			response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("video/x-flv"));
 
-			con->file_finished = 1;
+			con->send->is_closed = 1;
 
 			return HANDLER_FINISHED;
 		}
@@ -261,8 +261,7 @@ URIHANDLER_FUNC(mod_flv_streaming_path_handler) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_flv_streaming_plugin_init(plugin *p);
-int mod_flv_streaming_plugin_init(plugin *p) {
+LI_EXPORT int mod_flv_streaming_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("flv_streaming");
 
