@@ -1,3 +1,13 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <assert.h>
+#include <setjmp.h>
+
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -10,12 +20,6 @@
 #include "status_counter.h"
 #include "etag.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <assert.h>
-#include <setjmp.h>
-
 #ifdef HAVE_LUA_H
 #include <lua.h>
 #include <lauxlib.h>
@@ -103,7 +107,7 @@ SETDEFAULTS_FUNC(mod_magnet_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
@@ -158,81 +162,6 @@ static int mod_magnet_patch_connection(server *srv, connection *con, plugin_data
 }
 #undef PATCH
 
-/* See http://lua-users.org/wiki/GeneralizedPairsAndIpairs for implementation details. */
-
-/* Override the default pairs() function to allow us to use a __pairs metakey */
-static int magnet_pairs(lua_State *L) {
-	luaL_checkany(L, 1);
-
-	if (luaL_getmetafield(L, 1, "__pairs")) {
-		lua_insert(L, 1);
-		lua_call(L, lua_gettop(L) - 1, LUA_MULTRET);
-		return lua_gettop(L);
-	} else {
-		lua_pushvalue(L, lua_upvalueindex(1));
-		lua_insert(L, 1);
-		lua_call(L, lua_gettop(L) - 1, LUA_MULTRET);
-		return lua_gettop(L);
-	}
-}
-
-/* Define a function that will iterate over an array* (in upval 1) using current position (upval 2) */
-static int magnet_array_next(lua_State *L) {
-	data_unset *du;
-	data_string *ds;
-	data_integer *di;
-
-	size_t pos = lua_tointeger(L, lua_upvalueindex(1));
-	array *a = lua_touserdata(L, lua_upvalueindex(2));
-
-	lua_settop(L, 0);
-
-	if (pos >= a->used) return 0;
-	if (NULL != (du = a->data[pos])) {
-		if (du->key->used) {
-			lua_pushlstring(L, du->key->ptr, du->key->used - 1);
-		}
-		else {
-			lua_pushlstring(L, "", 0);
-		}
-		switch (du->type) {
-			case TYPE_STRING:
-				ds = (data_string *)du;
-				if (ds->value && ds->value->used) {
-					lua_pushlstring(L, ds->value->ptr, ds->value->used - 1);
-				} else {
-					lua_pushnil(L);
-				}
-				break;
-			case TYPE_COUNT:
-			case TYPE_INTEGER:
-				di = (data_integer *)du;
-				lua_pushinteger(L, di->value);
-				break;
-			default:
-				lua_pushnil(L);
-				break;
-		}
-
-		/* Update our positional upval to reflect our new current position */
-		pos++;
-		lua_pushinteger(L, pos);
-		lua_replace(L, lua_upvalueindex(1));
-
-		/* Returning 2 items on the stack (key, value) */
-		return 2;
-	}
-	return 0;
-}
-
-/* Create the closure necessary to iterate over the array *a with the above function */
-static int magnet_array_pairs(lua_State *L, array *a) {
-	lua_pushinteger(L, 0); /* Push our current pos (the start) into upval 1 */
-	lua_pushlightuserdata(L, a); /* Push our array *a into upval 2 */
-	lua_pushcclosure(L, magnet_array_next, 2); /* Push our new closure with 2 upvals */
-	return 1;
-}
-
 static int magnet_print(lua_State *L) {
 	const char *s = luaL_checkstring(L, 1);
 	server *srv;
@@ -358,17 +287,23 @@ static int magnet_atpanic(lua_State *L) {
 }
 
 static int magnet_reqhdr_get(lua_State *L) {
+	server *srv;
 	connection *con;
 	data_string *ds;
 
 	const char *key = luaL_checkstring(L, 2);
 
+	lua_pushstring(L, "lighty.srv");
+	lua_gettable(L, LUA_REGISTRYINDEX);
+	srv = lua_touserdata(L, -1);
+	lua_pop(L, 1);
+
 	lua_pushstring(L, "lighty.con");
 	lua_gettable(L, LUA_REGISTRYINDEX);
 	con = lua_touserdata(L, -1);
 	lua_pop(L, 1);
 
-	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, key))) {
+	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, key, strlen(key)))) {
 		if (ds->value->used) {
 			lua_pushlstring(L, ds->value->ptr, ds->value->used - 1);
 		} else {
@@ -380,17 +315,6 @@ static int magnet_reqhdr_get(lua_State *L) {
 	return 1;
 }
 
-static int magnet_reqhdr_pairs(lua_State *L) {
-	connection *con;
-
-	lua_pushstring(L, "lighty.con");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	con = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	return magnet_array_pairs(L, con->request.headers);
-}
-
 static int magnet_status_get(lua_State *L) {
 	data_integer *di;
 	server *srv;
@@ -403,7 +327,7 @@ static int magnet_status_get(lua_State *L) {
 	srv = lua_touserdata(L, -1);
 	lua_pop(L, 1);
 
-	di = status_counter_get_counter(srv, key, key_len);
+	di = status_counter_get_counter(key, key_len);
 
 	lua_pushnumber(L, (double)di->value);
 
@@ -422,22 +346,11 @@ static int magnet_status_set(lua_State *L) {
 	srv = lua_touserdata(L, -1);
 	lua_pop(L, 1);
 
-	status_counter_set(srv, key, key_len, counter);
+	status_counter_set(key, key_len, counter);
 
 	return 0;
 }
 
-static int magnet_status_pairs(lua_State *L) {
-	server *srv;
-
-	lua_pushstring(L, "lighty.srv");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	srv = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	return magnet_array_pairs(L, srv->status);
-}
-
 typedef struct {
 	const char *name;
 	enum {
@@ -446,7 +359,6 @@ typedef struct {
 		MAGNET_ENV_PHYICAL_PATH,
 		MAGNET_ENV_PHYICAL_REL_PATH,
 		MAGNET_ENV_PHYICAL_DOC_ROOT,
-		MAGNET_ENV_PHYICAL_BASEDIR,
 
 		MAGNET_ENV_URI_PATH,
 		MAGNET_ENV_URI_PATH_RAW,
@@ -457,36 +369,32 @@ typedef struct {
 		MAGNET_ENV_REQUEST_METHOD,
 		MAGNET_ENV_REQUEST_URI,
 		MAGNET_ENV_REQUEST_ORIG_URI,
-		MAGNET_ENV_REQUEST_PATH_INFO,
-		MAGNET_ENV_REQUEST_REMOTE_IP,
 		MAGNET_ENV_REQUEST_PROTOCOL
-	} type;
+       	} type;
 } magnet_env_t;
 
-static const magnet_env_t magnet_env[] = {
-	{ "physical.path", MAGNET_ENV_PHYICAL_PATH },
-	{ "physical.rel-path", MAGNET_ENV_PHYICAL_REL_PATH },
-	{ "physical.doc-root", MAGNET_ENV_PHYICAL_DOC_ROOT },
-	{ "physical.basedir", MAGNET_ENV_PHYICAL_BASEDIR },
-
-	{ "uri.path", MAGNET_ENV_URI_PATH },
-	{ "uri.path-raw", MAGNET_ENV_URI_PATH_RAW },
-	{ "uri.scheme", MAGNET_ENV_URI_SCHEME },
-	{ "uri.authority", MAGNET_ENV_URI_AUTHORITY },
-	{ "uri.query", MAGNET_ENV_URI_QUERY },
-
-	{ "request.method", MAGNET_ENV_REQUEST_METHOD },
-	{ "request.uri", MAGNET_ENV_REQUEST_URI },
-	{ "request.orig-uri", MAGNET_ENV_REQUEST_ORIG_URI },
-	{ "request.path-info", MAGNET_ENV_REQUEST_PATH_INFO },
-	{ "request.remote-ip", MAGNET_ENV_REQUEST_REMOTE_IP },
-	{ "request.protocol", MAGNET_ENV_REQUEST_PROTOCOL },
-
-	{ NULL, MAGNET_ENV_UNSET }
-};
-
-static buffer *magnet_env_get_buffer_by_id(server *srv, connection *con, int id) {
+static buffer *magnet_env_get_buffer(server *srv, connection *con, const char *key) {
 	buffer *dest = NULL;
+	size_t i;
+
+	const magnet_env_t env[] = {
+		{ "physical.path", MAGNET_ENV_PHYICAL_PATH },
+		{ "physical.rel-path", MAGNET_ENV_PHYICAL_REL_PATH },
+		{ "physical.doc-root", MAGNET_ENV_PHYICAL_DOC_ROOT },
+
+		{ "uri.path", MAGNET_ENV_URI_PATH },
+		{ "uri.path-raw", MAGNET_ENV_URI_PATH_RAW },
+		{ "uri.scheme", MAGNET_ENV_URI_SCHEME },
+		{ "uri.authority", MAGNET_ENV_URI_AUTHORITY },
+		{ "uri.query", MAGNET_ENV_URI_QUERY },
+
+		{ "request.method", MAGNET_ENV_REQUEST_METHOD },
+		{ "request.uri", MAGNET_ENV_REQUEST_URI },
+		{ "request.orig-uri", MAGNET_ENV_REQUEST_ORIG_URI },
+		{ "request.protocol", MAGNET_ENV_REQUEST_PROTOCOL },
+
+		{ NULL, MAGNET_ENV_UNSET }
+	};
 
 	UNUSED(srv);
 
@@ -495,11 +403,14 @@ static buffer *magnet_env_get_buffer_by_id(server *srv, connection *con, int id)
 	 *
 	 */
 
-	switch (id) {
+	for (i = 0; env[i].name; i++) {
+		if (0 == strcmp(key, env[i].name)) break;
+	}
+
+	switch (env[i].type) {
 	case MAGNET_ENV_PHYICAL_PATH: dest = con->physical.path; break;
 	case MAGNET_ENV_PHYICAL_REL_PATH: dest = con->physical.rel_path; break;
 	case MAGNET_ENV_PHYICAL_DOC_ROOT: dest = con->physical.doc_root; break;
-	case MAGNET_ENV_PHYICAL_BASEDIR: dest = con->physical.basedir; break;
 
 	case MAGNET_ENV_URI_PATH: dest = con->uri.path; break;
 	case MAGNET_ENV_URI_PATH_RAW: dest = con->uri.path_raw; break;
@@ -507,18 +418,10 @@ static buffer *magnet_env_get_buffer_by_id(server *srv, connection *con, int id)
 	case MAGNET_ENV_URI_AUTHORITY: dest = con->uri.authority; break;
 	case MAGNET_ENV_URI_QUERY: dest = con->uri.query; break;
 
-	case MAGNET_ENV_REQUEST_METHOD:
-		buffer_copy_string(srv->tmp_buf, get_http_method_name(con->request.http_method));
-		dest = srv->tmp_buf;
-		break;
+	case MAGNET_ENV_REQUEST_METHOD:   break;
 	case MAGNET_ENV_REQUEST_URI:      dest = con->request.uri; break;
 	case MAGNET_ENV_REQUEST_ORIG_URI: dest = con->request.orig_uri; break;
-	case MAGNET_ENV_REQUEST_PATH_INFO: dest = con->request.pathinfo; break;
-	case MAGNET_ENV_REQUEST_REMOTE_IP: dest = con->dst_addr_buf; break;
-	case MAGNET_ENV_REQUEST_PROTOCOL:
-		buffer_copy_string(srv->tmp_buf, get_http_version_name(con->request.http_version));
-		dest = srv->tmp_buf;
-		break;
+	case MAGNET_ENV_REQUEST_PROTOCOL: break;
 
 	case MAGNET_ENV_UNSET: break;
 	}
@@ -526,16 +429,6 @@ static buffer *magnet_env_get_buffer_by_id(server *srv, connection *con, int id)
 	return dest;
 }
 
-static buffer *magnet_env_get_buffer(server *srv, connection *con, const char *key) {
-	size_t i;
-
-	for (i = 0; magnet_env[i].name; i++) {
-		if (0 == strcmp(key, magnet_env[i].name)) break;
-	}
-
-	return magnet_env_get_buffer_by_id(srv, con, magnet_env[i].type);
-}
-
 static int magnet_env_get(lua_State *L) {
 	server *srv;
 	connection *con;
@@ -593,97 +486,6 @@ static int magnet_env_set(lua_State *L) {
 	return 0;
 }
 
-static int magnet_env_next(lua_State *L) {
-	server *srv;
-	connection *con;
-	int pos = lua_tointeger(L, lua_upvalueindex(1));
-
-	buffer *dest;
-
-	lua_pushstring(L, "lighty.srv");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	srv = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	lua_pushstring(L, "lighty.con");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	con = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	lua_settop(L, 0);
-
-	if (NULL == magnet_env[pos].name) return 0; /* end of list */
-
-	lua_pushstring(L, magnet_env[pos].name);
-
-	dest = magnet_env_get_buffer_by_id(srv, con, magnet_env[pos].type);
-	if (dest && dest->used) {
-		lua_pushlstring(L, dest->ptr, dest->used - 1);
-	} else {
-		lua_pushnil(L);
-	}
-
-	/* Update our positional upval to reflect our new current position */
-	pos++;
-	lua_pushinteger(L, pos);
-	lua_replace(L, lua_upvalueindex(1));
-
-	/* Returning 2 items on the stack (key, value) */
-	return 2;
-}
-
-static int magnet_env_pairs(lua_State *L) {
-	lua_pushinteger(L, 0); /* Push our current pos (the start) into upval 1 */
-	lua_pushcclosure(L, magnet_env_next, 1); /* Push our new closure with 1 upvals */
-	return 1;
-}
-
-static int magnet_cgi_get(lua_State *L) {
-	connection *con;
-	data_string *ds;
-
-	const char *key = luaL_checkstring(L, 2);
-
-	lua_pushstring(L, "lighty.con");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	con = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	if (NULL != (ds = (data_string *)array_get_element(con->environment, key)) && ds->value->used)
-		lua_pushlstring(L, CONST_BUF_LEN(ds->value));
-	else
-		lua_pushnil(L);
-
-	return 1;
-}
-
-static int magnet_cgi_set(lua_State *L) {
-	connection *con;
-
-	const char *key = luaL_checkstring(L, 2);
-	const char *val = luaL_checkstring(L, 3);
-
-	lua_pushstring(L, "lighty.con");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	con = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	array_set_key_value(con->environment, key, strlen(key), val, strlen(val));
-
-	return 0;
-}
-
-static int magnet_cgi_pairs(lua_State *L) {
-	connection *con;
-
-	lua_pushstring(L, "lighty.con");
-	lua_gettable(L, LUA_REGISTRYINDEX);
-	con = lua_touserdata(L, -1);
-	lua_pop(L, 1);
-
-	return magnet_array_pairs(L, con->environment);
-}
-
 
 static int magnet_copy_response_header(server *srv, connection *con, plugin_data *p, lua_State *L) {
 	UNUSED(p);
@@ -696,7 +498,7 @@ static int magnet_copy_response_header(server *srv, connection *con, plugin_data
 	/* lighty.header */
 
 	lua_getfield(L, -1, "lighty"); /* lighty.* from the env  */
-	force_assert(lua_istable(L, -1));
+	assert(lua_istable(L, -1));
 
 	lua_getfield(L, -1, "header"); /* lighty.header */
 	if (lua_istable(L, -1)) {
@@ -740,11 +542,11 @@ static int magnet_attach_content(server *srv, connection *con, plugin_data *p, l
 	 * get the environment of the function
 	 */
 
-	force_assert(lua_isfunction(L, -1));
+	assert(lua_isfunction(L, -1));
 	lua_getfenv(L, -1); /* -1 is the function */
 
 	lua_getfield(L, -1, "lighty"); /* lighty.* from the env  */
-	force_assert(lua_istable(L, -1));
+	assert(lua_istable(L, -1));
 
 	lua_getfield(L, -1, "content"); /* lighty.content */
 	if (lua_istable(L, -1)) {
@@ -759,24 +561,19 @@ static int magnet_attach_content(server *srv, connection *con, plugin_data *p, l
 				size_t s_len = 0;
 				const char *s = lua_tolstring(L, -1, &s_len);
 
-				chunkqueue_append_mem(con->write_queue, s, s_len + 1);
+				chunkqueue_append_mem(con->send, s, s_len + 1);
 			} else if (lua_istable(L, -1)) {
 				lua_getfield(L, -1, "filename");
 				lua_getfield(L, -2, "length");
 				lua_getfield(L, -3, "offset");
 
 				if (lua_isstring(L, -3)) { /* filename has to be a string */
-					buffer *fn;
+					buffer *fn = buffer_init();
 					stat_cache_entry *sce;
-					const char *fn_str;
-					handler_t res;
 
-					fn_str = lua_tostring(L, -3);
-					fn = buffer_init_string(fn_str);
+					buffer_copy_string(fn, lua_tostring(L, -3));
 
-					res = stat_cache_get_entry(srv, con, fn, &sce);
-
-					if (HANDLER_GO_ON == res) {
+					if (HANDLER_GO_ON == stat_cache_get_entry(srv, con, fn, &sce)) {
 						off_t off = 0;
 						off_t len = 0;
 
@@ -791,16 +588,14 @@ static int magnet_attach_content(server *srv, connection *con, plugin_data *p, l
 						}
 
 						if (off < 0) {
-							buffer_free(fn);
-							return luaL_error(L, "offset for '%s' is negative", fn_str);
+							return luaL_error(L, "offset for '%s' is negative", fn->ptr);
 						}
 
 						if (len < off) {
-							buffer_free(fn);
-							return luaL_error(L, "offset > length for '%s'", fn_str);
+							return luaL_error(L, "offset > length for '%s'", fn->ptr);
 						}
 
-						chunkqueue_append_file(con->write_queue, fn, off, len - off);
+						chunkqueue_append_file(con->send, fn, off, len - off);
 					}
 
 					buffer_free(fn);
@@ -835,36 +630,9 @@ static int magnet_attach_content(server *srv, connection *con, plugin_data *p, l
 	return 0;
 }
 
-static int traceback (lua_State *L) {
-	if (!lua_isstring(L, 1))  /* 'message' not a string? */
-		return 1;  /* keep it intact */
-	lua_getfield(L, LUA_GLOBALSINDEX, "debug");
-	if (!lua_istable(L, -1)) {
-		lua_pop(L, 1);
-		return 1;
-	}
-	lua_getfield(L, -1, "traceback");
-	if (!lua_isfunction(L, -1)) {
-		lua_pop(L, 2);
-		return 1;
-	}
-	lua_pushvalue(L, 1);  /* pass error message */
-	lua_pushinteger(L, 2);  /* skip this function and traceback */
-	lua_call(L, 2, 1);  /* call debug.traceback */
-	return 1;
-}
-
-static int push_traceback(lua_State *L, int narg) {
-	int base = lua_gettop(L) - narg;  /* function index */
-	lua_pushcfunction(L, traceback);
-	lua_insert(L, base);
-	return base;
-}
-
 static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, buffer *name) {
 	lua_State *L;
 	int lua_return_value = -1;
-	int errfunc;
 	/* get the script-context */
 
 
@@ -880,10 +648,9 @@ static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, bu
 
 		lua_pop(L, 1);
 
-		force_assert(lua_gettop(L) == 0); /* only the function should be on the stack */
+		assert(lua_gettop(L) == 0); /* only the function should be on the stack */
 
 		con->http_status = 500;
-		con->mode = DIRECT;
 
 		return HANDLER_FINISHED;
 	}
@@ -927,8 +694,6 @@ static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, bu
 	lua_newtable(L); /* the meta-table for the request-table     (sp += 1) */
 	lua_pushcfunction(L, magnet_reqhdr_get);                  /* (sp += 1) */
 	lua_setfield(L, -2, "__index");                           /* (sp -= 1) */
-	lua_pushcfunction(L, magnet_reqhdr_pairs);                /* (sp += 1) */
-	lua_setfield(L, -2, "__pairs");                           /* (sp -= 1) */
 	lua_setmetatable(L, -2); /* tie the metatable to request     (sp -= 1) */
 	lua_setfield(L, -2, "request"); /* content = {}              (sp -= 1) */
 
@@ -938,30 +703,15 @@ static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, bu
 	lua_setfield(L, -2, "__index");                           /* (sp -= 1) */
 	lua_pushcfunction(L, magnet_env_set);                     /* (sp += 1) */
 	lua_setfield(L, -2, "__newindex");                        /* (sp -= 1) */
-	lua_pushcfunction(L, magnet_env_pairs);                   /* (sp += 1) */
-	lua_setfield(L, -2, "__pairs");                           /* (sp -= 1) */
 	lua_setmetatable(L, -2); /* tie the metatable to request     (sp -= 1) */
 	lua_setfield(L, -2, "env"); /* content = {}                  (sp -= 1) */
 
-	lua_newtable(L); /*  {}                                      (sp += 1) */
-	lua_newtable(L); /* the meta-table for the request-table     (sp += 1) */
-	lua_pushcfunction(L, magnet_cgi_get);                     /* (sp += 1) */
-	lua_setfield(L, -2, "__index");                           /* (sp -= 1) */
-	lua_pushcfunction(L, magnet_cgi_set);                     /* (sp += 1) */
-	lua_setfield(L, -2, "__newindex");                        /* (sp -= 1) */
-	lua_pushcfunction(L, magnet_cgi_pairs);                   /* (sp += 1) */
-	lua_setfield(L, -2, "__pairs");                           /* (sp -= 1) */
-	lua_setmetatable(L, -2); /* tie the metatable to req_env     (sp -= 1) */
-	lua_setfield(L, -2, "req_env"); /* content = {}              (sp -= 1) */
-
 	lua_newtable(L); /*  {}                                      (sp += 1) */
 	lua_newtable(L); /* the meta-table for the request-table     (sp += 1) */
 	lua_pushcfunction(L, magnet_status_get);                  /* (sp += 1) */
 	lua_setfield(L, -2, "__index");                           /* (sp -= 1) */
 	lua_pushcfunction(L, magnet_status_set);                  /* (sp += 1) */
 	lua_setfield(L, -2, "__newindex");                        /* (sp -= 1) */
-	lua_pushcfunction(L, magnet_status_pairs);                /* (sp += 1) */
-	lua_setfield(L, -2, "__pairs");                           /* (sp -= 1) */
 	lua_setmetatable(L, -2); /* tie the metatable to request     (sp -= 1) */
 	lua_setfield(L, -2, "status"); /* content = {}               (sp -= 1) */
 
@@ -980,11 +730,6 @@ static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, bu
 
 	lua_setfield(L, -2, "lighty"); /* lighty.*                   (sp -= 1) */
 
-	/* override the default pairs() function to our __pairs capable version */
-	lua_getglobal(L, "pairs"); /* push original pairs()          (sp += 1) */
-	lua_pushcclosure(L, magnet_pairs, 1);
-	lua_setfield(L, -2, "pairs");                             /* (sp -= 1) */
-
 	lua_newtable(L); /* the meta-table for the new env           (sp += 1) */
 	lua_pushvalue(L, LUA_GLOBALSINDEX);                       /* (sp += 1) */
 	lua_setfield(L, -2, "__index"); /* { __index = _G }          (sp -= 1) */
@@ -993,26 +738,22 @@ static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, bu
 
 	lua_setfenv(L, -2); /* on the stack should be a modified env (sp -= 1) */
 
-	errfunc = push_traceback(L, 0);
-	if (lua_pcall(L, 0, 1, errfunc)) {
-		lua_remove(L, errfunc);
+	if (lua_pcall(L, 0, 1, 0)) {
 		log_error_write(srv, __FILE__, __LINE__,
 			"ss",
 			"lua_pcall():",
 			lua_tostring(L, -1));
 		lua_pop(L, 1); /* remove the error-msg and the function copy from the stack */
 
-		force_assert(lua_gettop(L) == 1); /* only the function should be on the stack */
+		assert(lua_gettop(L) == 1); /* only the function should be on the stack */
 
 		con->http_status = 500;
-		con->mode = DIRECT;
 
 		return HANDLER_FINISHED;
 	}
-	lua_remove(L, errfunc);
 
 	/* we should have the function-copy and the return value on the stack */
-	force_assert(lua_gettop(L) == 2);
+	assert(lua_gettop(L) == 2);
 
 	if (lua_isnumber(L, -1)) {
 		/* if the ret-value is a number, take it */
@@ -1024,30 +765,26 @@ static handler_t magnet_attract(server *srv, connection *con, plugin_data *p, bu
 
 	if (lua_return_value > 99) {
 		con->http_status = lua_return_value;
-		con->file_finished = 1;
+		con->send->is_closed = 1;
 
 		/* try { ...*/
 		if (0 == setjmp(exceptionjmp)) {
 			magnet_attach_content(srv, con, p, L);
-			if (!chunkqueue_is_empty(con->write_queue)) {
-				con->mode = p->id;
-			}
 		} else {
 			/* } catch () { */
 			con->http_status = 500;
-			con->mode = DIRECT;
 		}
 
-		force_assert(lua_gettop(L) == 1); /* only the function should be on the stack */
+		assert(lua_gettop(L) == 1); /* only the function should be on the stack */
 
 		/* we are finished */
 		return HANDLER_FINISHED;
 	} else if (MAGNET_RESTART_REQUEST == lua_return_value) {
-		force_assert(lua_gettop(L) == 1); /* only the function should be on the stack */
+		assert(lua_gettop(L) == 1); /* only the function should be on the stack */
 
 		return HANDLER_COMEBACK;
 	} else {
-		force_assert(lua_gettop(L) == 1); /* only the function should be on the stack */
+		assert(lua_gettop(L) == 1); /* only the function should be on the stack */
 
 		return HANDLER_GO_ON;
 	}
@@ -1095,7 +832,6 @@ URIHANDLER_FUNC(mod_magnet_physical) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_magnet_plugin_init(plugin *p);
 int mod_magnet_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("magnet");
@@ -1112,7 +848,6 @@ int mod_magnet_plugin_init(plugin *p) {
 }
 
 #else
-int mod_magnet_plugin_init(plugin *p);
 int mod_magnet_plugin_init(plugin *p) {
 	UNUSED(p);
 	return -1;
