@@ -1,3 +1,7 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -5,9 +9,9 @@
 #include "plugin.h"
 #include "response.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 typedef struct {
 	pcre_keyvalue_buffer *redirect;
@@ -29,6 +33,8 @@ typedef struct {
 INIT_FUNC(mod_redirect_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->match_buf = buffer_init();
@@ -47,8 +53,6 @@ FREE_FUNC(mod_redirect_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
-
 			pcre_keyvalue_buffer_free(s->redirect);
 
 			free(s);
@@ -67,6 +71,7 @@ FREE_FUNC(mod_redirect_free) {
 
 SETDEFAULTS_FUNC(mod_redirect_set_defaults) {
 	plugin_data *p = p_d;
+	data_unset *du;
 	size_t i = 0;
 
 	config_values_t cv[] = {
@@ -78,29 +83,28 @@ SETDEFAULTS_FUNC(mod_redirect_set_defaults) {
 	if (!p) return HANDLER_ERROR;
 
 	/* 0 */
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 		size_t j;
-		data_unset *du;
-		data_array *da;
+		array *ca;
+		data_array *da = NULL;
 
 		s = calloc(1, sizeof(plugin_config));
 		s->redirect   = pcre_keyvalue_buffer_init();
-		s->redirect_code = 301;
 
 		cv[0].destination = s->redirect;
 		cv[1].destination = &(s->redirect_code);
 
 		p->config_storage[i] = s;
+		ca = ((data_config *)srv->config_context->data[i])->value;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ca, cv)) {
 			return HANDLER_ERROR;
 		}
 
-		if (NULL == (du = array_get_element(config->value, "url.redirect"))) {
+		if (NULL == (du = array_get_element(ca, CONST_STR_LEN("url.redirect")))) {
 			/* no url.redirect defined */
 			continue;
 		}
@@ -124,13 +128,12 @@ SETDEFAULTS_FUNC(mod_redirect_set_defaults) {
 				return HANDLER_ERROR;
 			}
 
-			if (0 != pcre_keyvalue_buffer_append(srv, s->redirect,
+			if (0 != pcre_keyvalue_buffer_append(s->redirect,
 							     ((data_string *)(da->value->data[j]))->key->ptr,
 							     ((data_string *)(da->value->data[j]))->value->ptr)) {
 
 				log_error_write(srv, __FILE__, __LINE__, "sb",
 						"pcre-compile failed for", da->value->data[j]->key);
-				return HANDLER_ERROR;
 			}
 		}
 	}
@@ -142,9 +145,8 @@ static int mod_redirect_patch_connection(server *srv, connection *con, plugin_da
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	p->conf.redirect = s->redirect;
-	p->conf.redirect_code = s->redirect_code;
-	p->conf.context = NULL;
+	PATCH_OPTION(redirect);
+	PATCH_OPTION(redirect_code);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -158,11 +160,11 @@ static int mod_redirect_patch_connection(server *srv, connection *con, plugin_da
 		for (j = 0; j < dc->value->used; j++) {
 			data_unset *du = dc->value->data[j];
 
-			if (0 == strcmp(du->key->ptr, "url.redirect")) {
-				p->conf.redirect = s->redirect;
+			if (buffer_is_equal_string(du->key, CONST_STR_LEN("url.redirect"))) {
+				PATCH_OPTION(redirect);
 				p->conf.context = dc;
-			} else if (0 == strcmp(du->key->ptr, "url.redirect-code")) {
-				p->conf.redirect_code = s->redirect_code;
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("url.redirect-code"))) {
+				PATCH_OPTION(redirect_code);
 			}
 		}
 	}
@@ -173,7 +175,7 @@ static int mod_redirect_patch_connection(server *srv, connection *con, plugin_da
 static handler_t mod_redirect_uri_handler(server *srv, connection *con, void *p_data) {
 #ifdef HAVE_PCRE_H
 	plugin_data *p = p_data;
-	size_t i;
+	int i;
 
 	/*
 	 * REWRITE URL
@@ -184,84 +186,20 @@ static handler_t mod_redirect_uri_handler(server *srv, connection *con, void *p_
 
 	mod_redirect_patch_connection(srv, con, p);
 
-	buffer_copy_buffer(p->match_buf, con->request.uri);
+	buffer_copy_string_buffer(p->match_buf, con->request.uri);
+	i = config_exec_pcre_keyvalue_buffer(con, p->conf.redirect, p->conf.context, p->match_buf, p->location);
 
-	for (i = 0; i < p->conf.redirect->used; i++) {
-		pcre *match;
-		pcre_extra *extra;
-		const char *pattern;
-		size_t pattern_len;
-		int n;
-		pcre_keyvalue *kv = p->conf.redirect->kv[i];
-# define N 10
-		int ovec[N * 3];
+	if (i >= 0) {
+		response_header_insert(srv, con, CONST_STR_LEN("Location"), CONST_BUF_LEN(p->location));
 
-		match       = kv->key;
-		extra       = kv->key_extra;
-		pattern     = kv->value->ptr;
-		pattern_len = buffer_string_length(kv->value);
+		con->http_status = p->conf.redirect_code > 99 && p->conf.redirect_code < 1000 ? p->conf.redirect_code : 301;
+		con->send->is_closed = 1;
 
-		if ((n = pcre_exec(match, extra, CONST_BUF_LEN(p->match_buf), 0, 0, ovec, 3 * N)) < 0) {
-			if (n != PCRE_ERROR_NOMATCH) {
-				log_error_write(srv, __FILE__, __LINE__, "sd",
-						"execution error while matching: ", n);
-				return HANDLER_ERROR;
-			}
-		} else {
-			const char **list;
-			size_t start;
-			size_t k;
-
-			/* it matched */
-			pcre_get_substring_list(p->match_buf->ptr, ovec, n, &list);
-
-			/* search for $[0-9] */
-
-			buffer_reset(p->location);
-
-			start = 0;
-			for (k = 0; k + 1 < pattern_len; k++) {
-				if (pattern[k] == '$' || pattern[k] == '%') {
-					/* got one */
-
-					size_t num = pattern[k + 1] - '0';
-
-					buffer_append_string_len(p->location, pattern + start, k - start);
-
-					if (!isdigit((unsigned char)pattern[k + 1])) {
-						/* enable escape: "%%" => "%", "%a" => "%a", "$$" => "$" */
-						buffer_append_string_len(p->location, pattern+k, pattern[k] == pattern[k+1] ? 1 : 2);
-					} else if (pattern[k] == '$') {
-						/* n is always > 0 */
-						if (num < (size_t)n) {
-							buffer_append_string(p->location, list[num]);
-						}
-					} else if (p->conf.context == NULL) {
-						/* we have no context, we are global */
-						log_error_write(srv, __FILE__, __LINE__, "sb",
-								"used a rewrite containing a %[0-9]+ in the global scope, ignored:",
-								kv->value);
-					} else {
-						config_append_cond_match_buffer(con, p->conf.context, p->location, num);
-					}
-
-					k++;
-					start = k + 1;
-				}
-			}
-
-			buffer_append_string_len(p->location, pattern + start, pattern_len - start);
-
-			pcre_free(list);
-
-			response_header_insert(srv, con, CONST_STR_LEN("Location"), CONST_BUF_LEN(p->location));
-
-			con->http_status = p->conf.redirect_code > 99 && p->conf.redirect_code < 1000 ? p->conf.redirect_code : 301;
-			con->mode = DIRECT;
-			con->file_finished = 1;
-
-			return HANDLER_FINISHED;
-		}
+		return HANDLER_FINISHED;
+	}
+	else if (i != PCRE_ERROR_NOMATCH) {
+		log_error_write(srv, __FILE__, __LINE__, "s",
+				"execution error while matching", i);
 	}
 #undef N
 
@@ -275,8 +213,7 @@ static handler_t mod_redirect_uri_handler(server *srv, connection *con, void *p_
 }
 
 
-int mod_redirect_plugin_init(plugin *p);
-int mod_redirect_plugin_init(plugin *p) {
+LI_EXPORT int mod_redirect_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("redirect");
 
