@@ -1,14 +1,22 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
 
 #include "plugin.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
-#include "md5.h"
+#ifdef USE_OPENSSL
+# include <openssl/md5.h>
+#else
+# include "md5.h"
+#endif
 
 #define HASHLEN 16
 typedef unsigned char HASH[HASHLEN];
@@ -29,7 +37,7 @@ typedef struct {
 	buffer *secret;
 	buffer *uri_prefix;
 
-	unsigned int timeout;
+	unsigned short timeout;
 } plugin_config;
 
 typedef struct {
@@ -46,6 +54,8 @@ typedef struct {
 INIT_FUNC(mod_secdownload_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->md5 = buffer_init();
@@ -91,13 +101,13 @@ SETDEFAULTS_FUNC(mod_secdownload_set_defaults) {
 		{ "secdownload.secret",            NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },       /* 0 */
 		{ "secdownload.document-root",     NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },       /* 1 */
 		{ "secdownload.uri-prefix",        NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },       /* 2 */
-		{ "secdownload.timeout",           NULL, T_CONFIG_INT, T_CONFIG_SCOPE_CONNECTION },        /* 3 */
+		{ "secdownload.timeout",           NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },        /* 3 */
 		{ NULL,                            NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
 	};
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
@@ -130,7 +140,7 @@ SETDEFAULTS_FUNC(mod_secdownload_set_defaults) {
  * @return if the supplied string is a valid MD5 string 1 is returned otherwise 0
  */
 
-static int is_hex_len(const char *str, size_t len) {
+int is_hex_len(const char *str, size_t len) {
 	size_t i;
 
 	if (NULL == str) return 0;
@@ -148,16 +158,14 @@ static int is_hex_len(const char *str, size_t len) {
 	return i == len;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_secdownload_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(secret);
-	PATCH(doc_root);
-	PATCH(uri_prefix);
-	PATCH(timeout);
+	PATCH_OPTION(secret);
+	PATCH_OPTION(doc_root);
+	PATCH_OPTION(uri_prefix);
+	PATCH_OPTION(timeout);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -172,32 +180,28 @@ static int mod_secdownload_patch_connection(server *srv, connection *con, plugin
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("secdownload.secret"))) {
-				PATCH(secret);
+				PATCH_OPTION(secret);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("secdownload.document-root"))) {
-				PATCH(doc_root);
+				PATCH_OPTION(doc_root);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("secdownload.uri-prefix"))) {
-				PATCH(uri_prefix);
+				PATCH_OPTION(uri_prefix);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("secdownload.timeout"))) {
-				PATCH(timeout);
+				PATCH_OPTION(timeout);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
-
 
 URIHANDLER_FUNC(mod_secdownload_uri_handler) {
 	plugin_data *p = p_d;
-	li_MD5_CTX Md5Ctx;
+	MD5_CTX Md5Ctx;
 	HASH HA1;
 	const char *rel_uri, *ts_str, *md5_str;
 	time_t ts = 0;
 	size_t i;
 
-	if (con->mode != DIRECT) return HANDLER_GO_ON;
-
 	if (con->uri.path->used == 0) return HANDLER_GO_ON;
 
 	mod_secdownload_patch_connection(srv, con, p);
@@ -237,10 +241,9 @@ URIHANDLER_FUNC(mod_secdownload_uri_handler) {
 	}
 
 	/* timed-out */
-	if ( (srv->cur_ts > ts && (unsigned int) (srv->cur_ts - ts) > p->conf.timeout) ||
-	     (srv->cur_ts < ts && (unsigned int) (ts - srv->cur_ts) > p->conf.timeout) ) {
-		/* "Gone" as the url will never be valid again instead of "408 - Timeout" where the request may be repeated */
-		con->http_status = 410;
+	if (srv->cur_ts - ts > p->conf.timeout ||
+	    srv->cur_ts - ts < -p->conf.timeout) {
+		con->http_status = 408;
 
 		return HANDLER_FINISHED;
 	}
@@ -255,15 +258,14 @@ URIHANDLER_FUNC(mod_secdownload_uri_handler) {
 	buffer_copy_string_buffer(p->md5, p->conf.secret);
 	buffer_append_string(p->md5, rel_uri);
 	buffer_append_string_len(p->md5, ts_str, 8);
-	force_assert(p->md5->used > 0);
 
-	li_MD5_Init(&Md5Ctx);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)p->md5->ptr, p->md5->used - 1);
-	li_MD5_Final(HA1, &Md5Ctx);
+	MD5_Init(&Md5Ctx);
+	MD5_Update(&Md5Ctx, (unsigned char *)p->md5->ptr, p->md5->used - 1);
+	MD5_Final(HA1, &Md5Ctx);
 
 	buffer_copy_string_hex(p->md5, (char *)HA1, 16);
 
-	if (0 != strncasecmp(md5_str, p->md5->ptr, 32)) {
+	if (0 != strncmp(md5_str, p->md5->ptr, 32)) {
 		con->http_status = 403;
 
 		log_error_write(srv, __FILE__, __LINE__, "sss",
@@ -277,7 +279,6 @@ URIHANDLER_FUNC(mod_secdownload_uri_handler) {
 	 */
 
 	buffer_copy_string_buffer(con->physical.doc_root, p->conf.doc_root);
-	buffer_copy_string_buffer(con->physical.basedir, p->conf.doc_root);
 	buffer_copy_string(con->physical.rel_path, rel_uri);
 	buffer_copy_string_buffer(con->physical.path, con->physical.doc_root);
 	buffer_append_string_buffer(con->physical.path, con->physical.rel_path);
@@ -287,8 +288,7 @@ URIHANDLER_FUNC(mod_secdownload_uri_handler) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_secdownload_plugin_init(plugin *p);
-int mod_secdownload_plugin_init(plugin *p) {
+LI_EXPORT int mod_secdownload_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("secdownload");
 
