@@ -1,3 +1,6 @@
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -6,9 +9,6 @@
 
 #include "response.h"
 
-#include <stdlib.h>
-#include <string.h>
-
 /* plugin config for all request/connections */
 
 typedef struct {
@@ -30,7 +30,7 @@ typedef struct {
 	plugin_config conf;
 } plugin_data;
 
-static handler_ctx * handler_ctx_init(void) {
+static handler_ctx * handler_ctx_init() {
 	handler_ctx * hctx;
 
 	hctx = calloc(1, sizeof(*hctx));
@@ -49,6 +49,8 @@ static void handler_ctx_free(handler_ctx *hctx) {
 INIT_FUNC(mod_setenv_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	return p;
@@ -67,8 +69,6 @@ FREE_FUNC(mod_setenv_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
-
 			array_free(s->request_header);
 			array_free(s->response_header);
 			array_free(s->environment);
@@ -98,10 +98,9 @@ SETDEFAULTS_FUNC(mod_setenv_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
@@ -115,7 +114,7 @@ SETDEFAULTS_FUNC(mod_setenv_set_defaults) {
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 	}
@@ -123,15 +122,13 @@ SETDEFAULTS_FUNC(mod_setenv_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_setenv_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(request_header);
-	PATCH(response_header);
-	PATCH(environment);
+	PATCH_OPTION(request_header);
+	PATCH_OPTION(response_header);
+	PATCH_OPTION(environment);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -146,18 +143,17 @@ static int mod_setenv_patch_connection(server *srv, connection *con, plugin_data
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("setenv.add-request-header"))) {
-				PATCH(request_header);
+				PATCH_OPTION(request_header);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("setenv.add-response-header"))) {
-				PATCH(response_header);
+				PATCH_OPTION(response_header);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("setenv.add-environment"))) {
-				PATCH(environment);
+				PATCH_OPTION(environment);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 URIHANDLER_FUNC(mod_setenv_uri_handler) {
 	plugin_data *p = p_d;
@@ -188,8 +184,8 @@ URIHANDLER_FUNC(mod_setenv_uri_handler) {
 			ds_dst = data_string_init();
 		}
 
-		buffer_copy_buffer(ds_dst->key, ds->key);
-		buffer_copy_buffer(ds_dst->value, ds->value);
+		buffer_copy_string_buffer(ds_dst->key, ds->key);
+		buffer_copy_string_buffer(ds_dst->value, ds->value);
 
 		array_insert_unique(con->request.headers, (data_unset *)ds_dst);
 	}
@@ -202,8 +198,8 @@ URIHANDLER_FUNC(mod_setenv_uri_handler) {
 			ds_dst = data_string_init();
 		}
 
-		buffer_copy_buffer(ds_dst->key, ds->key);
-		buffer_copy_buffer(ds_dst->value, ds->value);
+		buffer_copy_string_buffer(ds_dst->key, ds->key);
+		buffer_copy_string_buffer(ds_dst->value, ds->value);
 
 		array_insert_unique(con->environment, (data_unset *)ds_dst);
 	}
@@ -218,7 +214,7 @@ URIHANDLER_FUNC(mod_setenv_uri_handler) {
 	return HANDLER_GO_ON;
 }
 
-CONNECTION_FUNC(mod_setenv_reset) {
+REQUESTDONE_FUNC(mod_setenv_reset) {
 	plugin_data *p = p_d;
 
 	UNUSED(srv);
@@ -233,8 +229,7 @@ CONNECTION_FUNC(mod_setenv_reset) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_setenv_plugin_init(plugin *p);
-int mod_setenv_plugin_init(plugin *p) {
+LI_EXPORT int mod_setenv_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("setenv");
 
