@@ -1,3 +1,8 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+#include <errno.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -5,11 +10,11 @@
 
 #include "plugin.h"
 
-#include <assert.h>
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <errno.h>
+#include "sys-files.h"
+
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 typedef struct {
 	buffer *server_root;
@@ -35,6 +40,8 @@ typedef struct {
 INIT_FUNC(mod_simple_vhost_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->doc_root = buffer_init();
@@ -89,10 +96,9 @@ SETDEFAULTS_FUNC(mod_simple_vhost_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
@@ -115,7 +121,7 @@ SETDEFAULTS_FUNC(mod_simple_vhost_set_defaults) {
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 	}
@@ -125,32 +131,37 @@ SETDEFAULTS_FUNC(mod_simple_vhost_set_defaults) {
 
 static int build_doc_root(server *srv, connection *con, plugin_data *p, buffer *out, buffer *host) {
 	stat_cache_entry *sce = NULL;
-	force_assert(!buffer_string_is_empty(p->conf.server_root));
 
-	buffer_string_prepare_copy(out, 127);
-	buffer_copy_buffer(out, p->conf.server_root);
+	buffer_prepare_copy(out, 128);
 
-	if (!buffer_string_is_empty(host)) {
-		/* a hostname has to start with a alpha-numerical character
-		 * and must not contain a slash "/"
-		 */
-		char *dp;
+	if (p->conf.server_root->used) {
+		buffer_copy_string_buffer(out, p->conf.server_root);
 
-		buffer_append_slash(out);
+		if (host->used) {
+			/* a hostname has to start with a alpha-numerical character
+			 * and must not contain a slash "/"
+			 */
+			char *dp;
 
-		if (NULL == (dp = strchr(host->ptr, ':'))) {
-			buffer_append_string_buffer(out, host);
-		} else {
-			buffer_append_string_len(out, host->ptr, dp - host->ptr);
+			PATHNAME_APPEND_SLASH(out);
+
+			if (NULL == (dp = strchr(host->ptr, ':'))) {
+				buffer_append_string_buffer(out, host);
+			} else {
+				buffer_append_string_len(out, host->ptr, dp - host->ptr);
+			}
 		}
-	}
-	buffer_append_slash(out);
+		PATHNAME_APPEND_SLASH(out);
 
-	if (buffer_string_length(p->conf.document_root) > 1 && p->conf.document_root->ptr[0] == '/') {
-		buffer_append_string_len(out, p->conf.document_root->ptr + 1, buffer_string_length(p->conf.document_root) - 1);
+		if (p->conf.document_root->used > 2 && p->conf.document_root->ptr[0] == '/') {
+			buffer_append_string_len(out, p->conf.document_root->ptr + 1, p->conf.document_root->used - 2);
+		} else {
+			buffer_append_string_buffer(out, p->conf.document_root);
+			PATHNAME_APPEND_SLASH(out);
+		}
 	} else {
-		buffer_append_string_buffer(out, p->conf.document_root);
-		buffer_append_slash(out);
+		buffer_copy_string_buffer(out, con->conf.document_root);
+		PATHNAME_APPEND_SLASH(out);
 	}
 
 	if (HANDLER_ERROR == stat_cache_get_entry(srv, con, out, &sce)) {
@@ -166,22 +177,19 @@ static int build_doc_root(server *srv, connection *con, plugin_data *p, buffer *
 	return 0;
 }
 
-
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_simple_vhost_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(server_root);
-	PATCH(default_host);
-	PATCH(document_root);
+	PATCH_OPTION(server_root);
+	PATCH_OPTION(default_host);
+	PATCH_OPTION(document_root);
 
-	PATCH(docroot_cache_key);
-	PATCH(docroot_cache_value);
-	PATCH(docroot_cache_servername);
+	PATCH_OPTION(docroot_cache_key);
+	PATCH_OPTION(docroot_cache_value);
+	PATCH_OPTION(docroot_cache_servername);
 
-	PATCH(debug);
+	PATCH_OPTION(debug);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -196,23 +204,22 @@ static int mod_simple_vhost_patch_connection(server *srv, connection *con, plugi
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("simple-vhost.server-root"))) {
-				PATCH(server_root);
-				PATCH(docroot_cache_key);
-				PATCH(docroot_cache_value);
-				PATCH(docroot_cache_servername);
+				PATCH_OPTION(server_root);
+				PATCH_OPTION(docroot_cache_key);
+				PATCH_OPTION(docroot_cache_value);
+				PATCH_OPTION(docroot_cache_servername);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("simple-vhost.default-host"))) {
-				PATCH(default_host);
+				PATCH_OPTION(default_host);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("simple-vhost.document-root"))) {
-				PATCH(document_root);
+				PATCH_OPTION(document_root);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("simple-vhost.debug"))) {
-				PATCH(debug);
+				PATCH_OPTION(debug);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 static handler_t mod_simple_vhost_docroot(server *srv, connection *con, void *p_data) {
 	plugin_data *p = p_data;
@@ -225,49 +232,41 @@ static handler_t mod_simple_vhost_docroot(server *srv, connection *con, void *p_
 
 	mod_simple_vhost_patch_connection(srv, con, p);
 
-	/* build_doc_root() requires a server_root; skip module if simple-vhost.server-root is not set
-	 * or set to an empty string (especially don't cache any results!)
-	 */
-	if (buffer_string_is_empty(p->conf.server_root)) return HANDLER_GO_ON;
-
-	if (!buffer_string_is_empty(p->conf.docroot_cache_key) &&
-	    !buffer_string_is_empty(con->uri.authority) &&
+	if (p->conf.docroot_cache_key->used &&
+	    con->uri.authority->used &&
 	    buffer_is_equal(p->conf.docroot_cache_key, con->uri.authority)) {
 		/* cache hit */
-		buffer_copy_buffer(con->server_name,       p->conf.docroot_cache_servername);
-		buffer_copy_buffer(con->physical.doc_root, p->conf.docroot_cache_value);
+		buffer_copy_string_buffer(con->physical.doc_root, p->conf.docroot_cache_value);
+		buffer_copy_string_buffer(con->server_name,       p->conf.docroot_cache_servername);
 	} else {
 		/* build document-root */
-		if (buffer_string_is_empty(con->uri.authority) ||
+		if ((con->uri.authority->used == 0) ||
 		    build_doc_root(srv, con, p, p->doc_root, con->uri.authority)) {
 			/* not found, fallback the default-host */
-			if (0 == build_doc_root(srv, con, p,
+			if (build_doc_root(srv, con, p,
 					   p->doc_root,
 					   p->conf.default_host)) {
-				/* default host worked */
-				buffer_copy_buffer(con->server_name, p->conf.default_host);
-				buffer_copy_buffer(con->physical.doc_root, p->doc_root);
-				/* do not cache default host */
+				return HANDLER_GO_ON;
+			} else {
+				buffer_copy_string_buffer(con->server_name, p->conf.default_host);
 			}
-			return HANDLER_GO_ON;
+		} else {
+			buffer_copy_string_buffer(con->server_name, con->uri.authority);
 		}
 
-		/* found host */
-		buffer_copy_buffer(con->server_name, con->uri.authority);
-		buffer_copy_buffer(con->physical.doc_root, p->doc_root);
-
 		/* copy to cache */
-		buffer_copy_buffer(p->conf.docroot_cache_key,        con->uri.authority);
-		buffer_copy_buffer(p->conf.docroot_cache_value,      p->doc_root);
-		buffer_copy_buffer(p->conf.docroot_cache_servername, con->server_name);
+		buffer_copy_string_buffer(p->conf.docroot_cache_key,        con->uri.authority);
+		buffer_copy_string_buffer(p->conf.docroot_cache_value,      p->doc_root);
+		buffer_copy_string_buffer(p->conf.docroot_cache_servername, con->server_name);
+
+		buffer_copy_string_buffer(con->physical.doc_root, p->doc_root);
 	}
 
 	return HANDLER_GO_ON;
 }
 
 
-int mod_simple_vhost_plugin_init(plugin *p);
-int mod_simple_vhost_plugin_init(plugin *p) {
+LI_EXPORT int mod_simple_vhost_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("simple_vhost");
 
