@@ -1,11 +1,12 @@
-#include "buffer.h"
-#include "log.h"
+#include <ctype.h>
+#include <string.h>
+
 #include "mod_ssi.h"
 #include "mod_ssi_expr.h"
 #include "mod_ssi_exprparser.h"
 
-#include <ctype.h>
-#include <string.h>
+#include "buffer.h"
+#include "log.h"
 
 typedef struct {
 	const char *input;
@@ -19,7 +20,7 @@ typedef struct {
 	int in_cond;
 } ssi_tokenizer_t;
 
-ssi_val_t *ssi_val_init(void) {
+ssi_val_t *ssi_val_init() {
 	ssi_val_t *s;
 
 	s = calloc(1, sizeof(*s));
@@ -59,7 +60,7 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 			t->offset++;
 			t->line_pos++;
 
-			buffer_copy_string_len(token, CONST_STR_LEN("(=)"));
+			buffer_copy_string(token, "(=)");
 
 			break;
 		case '>':
@@ -69,14 +70,14 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 				tid = TK_GE;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(>=)"));
+				buffer_copy_string(token, "(>=)");
 			} else {
 				t->offset += 1;
 				t->line_pos += 1;
 
 				tid = TK_GT;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(>)"));
+				buffer_copy_string(token, "(>)");
 			}
 
 			break;
@@ -87,14 +88,14 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 				tid = TK_LE;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(<=)"));
+				buffer_copy_string(token, "(<=)");
 			} else {
 				t->offset += 1;
 				t->line_pos += 1;
 
 				tid = TK_LT;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(<)"));
+				buffer_copy_string(token, "(<)");
 			}
 
 			break;
@@ -106,14 +107,14 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 				tid = TK_NE;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(!=)"));
+				buffer_copy_string(token, "(!=)");
 			} else {
 				t->offset += 1;
 				t->line_pos += 1;
 
 				tid = TK_NOT;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(!)"));
+				buffer_copy_string(token, "(!)");
 			}
 
 			break;
@@ -124,7 +125,7 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 				tid = TK_AND;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(&&)"));
+				buffer_copy_string(token, "(&&)");
 			} else {
 				log_error_write(srv, __FILE__, __LINE__, "sds",
 						"pos:", t->line_pos,
@@ -140,7 +141,7 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 				tid = TK_OR;
 
-				buffer_copy_string_len(token, CONST_STR_LEN("(||)"));
+				buffer_copy_string(token, "(||)");
 			} else {
 				log_error_write(srv, __FILE__, __LINE__, "sds",
 						"pos:", t->line_pos,
@@ -183,7 +184,7 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 			tid = TK_LPARAN;
 
-			buffer_copy_string_len(token, CONST_STR_LEN("("));
+			buffer_copy_string(token, "(");
 			break;
 		case ')':
 			t->offset++;
@@ -191,7 +192,7 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 			tid = TK_RPARAN;
 
-			buffer_copy_string_len(token, CONST_STR_LEN(")"));
+			buffer_copy_string(token, ")");
 			break;
 		case '$':
 			if (t->input[t->offset + 1] == '{') {
@@ -214,12 +215,12 @@ static int ssi_expr_tokenizer(server *srv, connection *con, plugin_data *p,
 
 			tid = TK_VALUE;
 
-			if (NULL != (ds = (data_string *)array_get_element(p->ssi_cgi_env, token->ptr))) {
+			if (NULL != (ds = (data_string *)array_get_element(p->ssi_cgi_env, CONST_BUF_LEN(token)))) {
 				buffer_copy_string_buffer(token, ds->value);
-			} else if (NULL != (ds = (data_string *)array_get_element(p->ssi_vars, token->ptr))) {
+			} else if (NULL != (ds = (data_string *)array_get_element(p->ssi_vars, CONST_BUF_LEN(token)))) {
 				buffer_copy_string_buffer(token, ds->value);
 			} else {
-				buffer_copy_string_len(token, CONST_STR_LEN(""));
+				buffer_copy_string(token, "");
 			}
 
 			t->offset += i;
