@@ -1,3 +1,8 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -6,14 +11,12 @@
 
 #include "stat_cache.h"
 #include "etag.h"
-#include "http_chunk.h"
 #include "response.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
+#include "sys-files.h"
+#include "sys-strings.h"
 
+#include "http_req_range.h"
 /**
  * this is a staticfile for a lighttpd plugin
  *
@@ -25,8 +28,6 @@
 
 typedef struct {
 	array *exclude_ext;
-	unsigned short etags_used;
-	unsigned short disable_pathinfo;
 } plugin_config;
 
 typedef struct {
@@ -34,6 +35,8 @@ typedef struct {
 
 	buffer *range_buf;
 
+	http_req_range *ranges;
+
 	plugin_config **config_storage;
 
 	plugin_config conf;
@@ -43,14 +46,18 @@ typedef struct {
 INIT_FUNC(mod_staticfile_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->range_buf = buffer_init();
 
+	p->ranges = http_request_range_init();
+
 	return p;
 }
 
-/* detroy the plugin data */
+/* destroy the plugin data */
 FREE_FUNC(mod_staticfile_free) {
 	plugin_data *p = p_d;
 
@@ -71,6 +78,8 @@ FREE_FUNC(mod_staticfile_free) {
 	}
 	buffer_free(p->range_buf);
 
+	http_request_range_free(p->ranges);
+
 	free(p);
 
 	return HANDLER_GO_ON;
@@ -84,26 +93,20 @@ SETDEFAULTS_FUNC(mod_staticfile_set_defaults) {
 
 	config_values_t cv[] = {
 		{ "static-file.exclude-extensions", NULL, T_CONFIG_ARRAY, T_CONFIG_SCOPE_CONNECTION },       /* 0 */
-		{ "static-file.etags",    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 1 */
-		{ "static-file.disable-pathinfo", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 2 */
 		{ NULL,                         NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
 	};
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
 		s->exclude_ext    = array_init();
-		s->etags_used     = 1;
-		s->disable_pathinfo = 0;
 
 		cv[0].destination = s->exclude_ext;
-		cv[1].destination = &(s->etags_used);
-		cv[2].destination = &(s->disable_pathinfo);
 
 		p->config_storage[i] = s;
 
@@ -115,15 +118,11 @@ SETDEFAULTS_FUNC(mod_staticfile_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_staticfile_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(exclude_ext);
-	PATCH(etags_used);
-	PATCH(disable_pathinfo);
+	PATCH_OPTION(exclude_ext);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -138,213 +137,175 @@ static int mod_staticfile_patch_connection(server *srv, connection *con, plugin_
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("static-file.exclude-extensions"))) {
-				PATCH(exclude_ext);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("static-file.etags"))) {
-				PATCH(etags_used);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("static-file.disable-pathinfo"))) {
-				PATCH(disable_pathinfo);
+				PATCH_OPTION(exclude_ext);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 static int http_response_parse_range(server *srv, connection *con, plugin_data *p) {
 	int multipart = 0;
-	int error;
-	off_t start, end;
-	const char *s, *minus;
 	char *boundary = "fkj49sn38dcn3";
 	data_string *ds;
 	stat_cache_entry *sce = NULL;
 	buffer *content_type = NULL;
+	buffer *range = NULL;
+	http_req_range *ranges, *r;
+
+	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, CONST_STR_LEN("Range")))) {
+		range = ds->value;
+	} else {
+		/* we don't have a Range header */
+
+		return -1;
+	}
 
 	if (HANDLER_ERROR == stat_cache_get_entry(srv, con, con->physical.path, &sce)) {
 		SEGFAULT();
 	}
 
-	start = 0;
-	end = sce->st.st_size - 1;
-
 	con->response.content_length = 0;
 
-	if (NULL != (ds = (data_string *)array_get_element(con->response.headers, "Content-Type"))) {
+	if (NULL != (ds = (data_string *)array_get_element(con->response.headers, CONST_STR_LEN("Content-Type")))) {
 		content_type = ds->value;
 	}
 
-	for (s = con->request.http_range, error = 0;
-	     !error && *s && NULL != (minus = strchr(s, '-')); ) {
-		char *err;
-		off_t la, le;
+	/* start the range-header parser
+	 * bytes=<num>  */
 
-		if (s == minus) {
-			/* -<stop> */
+	ranges = p->ranges;
+	http_request_range_reset(ranges);
+	switch (http_request_range_parse(range, ranges)) {
+	case PARSE_ERROR:
+		return -1; /* no range valid Range Header */
+	case PARSE_SUCCESS:
+		break;
+	default:
+		TRACE("%s", "foobar");
+		return -1;
+	}
 
-			le = strtoll(s, &err, 10);
+	if (ranges->next) {
+		multipart = 1;
+	}
 
-			if (le == 0) {
-				/* RFC 2616 - 14.35.1 */
+	/* patch the '-1' */
+	for (r = ranges; r; r = r->next) {
+		if (r->start == -1) {
+			/* -<end>
+			 *
+			 * the last <end> bytes  */
+			r->start = sce->st.st_size - r->end;
+			r->end = sce->st.st_size - 1;
+		}
+		if (r->end == -1) {
+			/* <start>-
+			 * all but the first <start> bytes */
 
-				con->http_status = 416;
-				error = 1;
-			} else if (*err == '\0') {
-				/* end */
-				s = err;
-
-				end = sce->st.st_size - 1;
-				start = sce->st.st_size + le;
-			} else if (*err == ',') {
-				multipart = 1;
-				s = err + 1;
-
-				end = sce->st.st_size - 1;
-				start = sce->st.st_size + le;
-			} else {
-				error = 1;
-			}
-
-		} else if (*(minus+1) == '\0' || *(minus+1) == ',') {
-			/* <start>- */
-
-			la = strtoll(s, &err, 10);
-
-			if (err == minus) {
-				/* ok */
-
-				if (*(err + 1) == '\0') {
-					s = err + 1;
-
-					end = sce->st.st_size - 1;
-					start = la;
-
-				} else if (*(err + 1) == ',') {
-					multipart = 1;
-					s = err + 2;
-
-					end = sce->st.st_size - 1;
-					start = la;
-				} else {
-					error = 1;
-				}
-			} else {
-				/* error */
-				error = 1;
-			}
-		} else {
-			/* <start>-<stop> */
-
-			la = strtoll(s, &err, 10);
-
-			if (err == minus) {
-				le = strtoll(minus+1, &err, 10);
-
-				/* RFC 2616 - 14.35.1 */
-				if (la > le) {
-					error = 1;
-				}
-
-				if (*err == '\0') {
-					/* ok, end*/
-					s = err;
-
-					end = le;
-					start = la;
-				} else if (*err == ',') {
-					multipart = 1;
-					s = err + 1;
-
-					end = le;
-					start = la;
-				} else {
-					/* error */
-
-					error = 1;
-				}
-			} else {
-				/* error */
-
-				error = 1;
-			}
+			r->end = sce->st.st_size - 1;
 		}
 
-		if (!error) {
-			if (start < 0) start = 0;
-
-			/* RFC 2616 - 14.35.1 */
-			if (end > sce->st.st_size - 1) end = sce->st.st_size - 1;
-
-			if (start > sce->st.st_size - 1) {
-				error = 1;
-
-				con->http_status = 416;
-			}
+		if (r->end > sce->st.st_size - 1) {
+			/* RFC 2616 - 14.35.1
+			 *
+			 * if last-byte-pos not present or > size-of-file
+			 * take the size-of-file
+			 *
+			 *  */
+			r->end = sce->st.st_size - 1;
 		}
 
-		if (!error) {
-			if (multipart) {
-				/* write boundary-header */
-				buffer *b;
+		if (r->start > sce->st.st_size - 1) {
+			/* RFC 2616 - 14.35.1
+			 *
+			 * if first-byte-pos > file-size, 416
+			 */
 
-				b = chunkqueue_get_append_buffer(con->write_queue);
+			con->http_status = 416;
+			return -1;
+		}
 
-				buffer_copy_string_len(b, CONST_STR_LEN("\r\n--"));
-				buffer_append_string(b, boundary);
+		if (r->start > r->end) {
+			/* RFC 2616 - 14.35.1
+			 *
+			 * if last-byte-pos is present, it has to be >= first-byte-pos
+			 *
+			 * invalid ranges have to be handle as no Range specified
+			 *  */
 
-				/* write Content-Range */
-				buffer_append_string_len(b, CONST_STR_LEN("\r\nContent-Range: bytes "));
-				buffer_append_off_t(b, start);
-				buffer_append_string_len(b, CONST_STR_LEN("-"));
-				buffer_append_off_t(b, end);
-				buffer_append_string_len(b, CONST_STR_LEN("/"));
-				buffer_append_off_t(b, sce->st.st_size);
-
-				buffer_append_string_len(b, CONST_STR_LEN("\r\nContent-Type: "));
-				buffer_append_string_buffer(b, content_type);
-
-				/* write END-OF-HEADER */
-				buffer_append_string_len(b, CONST_STR_LEN("\r\n\r\n"));
-
-				con->response.content_length += b->used - 1;
-
-			}
-
-			chunkqueue_append_file(con->write_queue, con->physical.path, start, end - start + 1);
-			con->response.content_length += end - start + 1;
+			return -1;
 		}
 	}
 
-	/* something went wrong */
-	if (error) return -1;
+	if (r) {
+		/* we ran into an range violation */
+		return -1;
+	}
 
 	if (multipart) {
-		/* add boundary end */
 		buffer *b;
+		for (r = ranges; r; r = r->next) {
+			/* write boundary-header */
 
-		b = chunkqueue_get_append_buffer(con->write_queue);
+			b = chunkqueue_get_append_buffer(con->send);
+
+			buffer_copy_string(b, "\r\n--");
+			buffer_append_string(b, boundary);
+
+			/* write Content-Range */
+			buffer_append_string(b, "\r\nContent-Range: bytes ");
+			buffer_append_off_t(b, r->start);
+			buffer_append_string(b, "-");
+			buffer_append_off_t(b, r->end);
+			buffer_append_string(b, "/");
+			buffer_append_off_t(b, sce->st.st_size);
+
+			buffer_append_string(b, "\r\nContent-Type: ");
+			buffer_append_string_buffer(b, content_type);
+
+			/* write END-OF-HEADER */
+			buffer_append_string(b, "\r\n\r\n");
+
+			con->response.content_length += b->used - 1;
+			con->send->bytes_in += b->used - 1;
+
+			chunkqueue_append_file(con->send, con->physical.path, r->start, r->end - r->start + 1);
+			con->response.content_length += r->end - r->start + 1;
+			con->send->bytes_in += r->end - r->start + 1;
+		}
+
+		/* add boundary end */
+		b = chunkqueue_get_append_buffer(con->send);
 
 		buffer_copy_string_len(b, "\r\n--", 4);
 		buffer_append_string(b, boundary);
 		buffer_append_string_len(b, "--\r\n", 4);
 
 		con->response.content_length += b->used - 1;
+		con->send->bytes_in += b->used - 1;
 
 		/* set header-fields */
 
-		buffer_copy_string_len(p->range_buf, CONST_STR_LEN("multipart/byteranges; boundary="));
+		buffer_copy_string(p->range_buf, "multipart/byteranges; boundary=");
 		buffer_append_string(p->range_buf, boundary);
 
 		/* overwrite content-type */
 		response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(p->range_buf));
-	} else {
-		/* add Content-Range-header */
 
-		buffer_copy_string_len(p->range_buf, CONST_STR_LEN("bytes "));
-		buffer_append_off_t(p->range_buf, start);
-		buffer_append_string_len(p->range_buf, CONST_STR_LEN("-"));
-		buffer_append_off_t(p->range_buf, end);
-		buffer_append_string_len(p->range_buf, CONST_STR_LEN("/"));
+	} else {
+		r = ranges;
+
+		chunkqueue_append_file(con->send, con->physical.path, r->start, r->end - r->start + 1);
+		con->response.content_length += r->end - r->start + 1;
+		con->send->bytes_in += r->end - r->start + 1;
+
+		buffer_copy_string(p->range_buf, "bytes ");
+		buffer_append_off_t(p->range_buf, r->start);
+		buffer_append_string(p->range_buf, "-");
+		buffer_append_off_t(p->range_buf, r->end);
+		buffer_append_string(p->range_buf, "/");
 		buffer_append_off_t(p->range_buf, sce->st.st_size);
 
 		response_header_insert(srv, con, CONST_STR_LEN("Content-Range"), CONST_BUF_LEN(p->range_buf));
@@ -357,10 +318,10 @@ static int http_response_parse_range(server *srv, connection *con, plugin_data *
 URIHANDLER_FUNC(mod_staticfile_subrequest) {
 	plugin_data *p = p_d;
 	size_t k;
+	int s_len;
 	stat_cache_entry *sce = NULL;
-	buffer *mtime = NULL;
+	buffer *mtime;
 	data_string *ds;
-	int allow_caching = 1;
 
 	/* someone else has done a decision for us */
 	if (con->http_status != 0) return HANDLER_GO_ON;
@@ -382,12 +343,7 @@ URIHANDLER_FUNC(mod_staticfile_subrequest) {
 
 	mod_staticfile_patch_connection(srv, con, p);
 
-	if (p->conf.disable_pathinfo && 0 != con->request.pathinfo->used) {
-		if (con->conf.log_request_handling) {
-			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- NOT handling file as static file, pathinfo forbidden");
-		}
-		return HANDLER_GO_ON;
-	}
+	s_len = con->uri.path->used - 1;
 
 	/* ignore certain extensions */
 	for (k = 0; k < p->conf.exclude_ext->used; k++) {
@@ -396,10 +352,9 @@ URIHANDLER_FUNC(mod_staticfile_subrequest) {
 		if (ds->value->used == 0) continue;
 
 		if (buffer_is_equal_right_len(con->physical.path, ds->value, ds->value->used - 1)) {
-			if (con->conf.log_request_handling) {
-				log_error_write(srv, __FILE__, __LINE__,  "s",  "-- NOT handling file as static file, extension forbidden");
-			}
-			return HANDLER_GO_ON;
+			con->http_status = 403;
+
+			return HANDLER_FINISHED;
 		}
 	}
 
@@ -418,7 +373,7 @@ URIHANDLER_FUNC(mod_staticfile_subrequest) {
 		return HANDLER_FINISHED;
 	}
 
-	/* we only handline regular files */
+	/* we only handle regular files */
 #ifdef HAVE_LSTAT
 	if ((sce->is_symlink == 1) && !con->conf.follow_symlink) {
 		con->http_status = 403;
@@ -444,85 +399,53 @@ URIHANDLER_FUNC(mod_staticfile_subrequest) {
 		return HANDLER_FINISHED;
 	}
 
-	/* mod_compress might set several data directly, don't overwrite them */
+	/* mod_compress might set several parameters directly; don't overwrite them */
 
 	/* set response content-type, if not set already */
 
-	if (NULL == array_get_element(con->response.headers, "Content-Type")) {
+	if (NULL == array_get_element(con->response.headers, CONST_STR_LEN("Content-Type"))) {
 		if (buffer_is_empty(sce->content_type)) {
-			/* we are setting application/octet-stream, but also announce that
-			 * this header field might change in the seconds few requests 
-			 *
-			 * This should fix the aggressive caching of FF and the script download
-			 * seen by the first installations
-			 */
 			response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("application/octet-stream"));
-
-			allow_caching = 0;
 		} else {
 			response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(sce->content_type));
 		}
 	}
 
-	if (con->conf.range_requests) {
-		response_header_overwrite(srv, con, CONST_STR_LEN("Accept-Ranges"), CONST_STR_LEN("bytes"));
+	if (NULL == array_get_element(con->response.headers, CONST_STR_LEN("ETag"))) {
+		/* generate e-tag */
+		etag_mutate(con->physical.etag, sce->etag);
+
+		response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
+	}
+	response_header_overwrite(srv, con, CONST_STR_LEN("Accept-Ranges"), CONST_STR_LEN("bytes"));
+
+	/* prepare header */
+	if (NULL == (ds = (data_string *)array_get_element(con->response.headers, CONST_STR_LEN("Last-Modified")))) {
+		mtime = strftime_cache_get(srv, sce->st.st_mtime);
+		response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
+	} else {
+		mtime = ds->value;
 	}
 
-	if (allow_caching) {
-		if (p->conf.etags_used && con->etag_flags != 0 && !buffer_is_empty(sce->etag)) {
-			if (NULL == array_get_element(con->response.headers, "ETag")) {
-				/* generate e-tag */
-				etag_mutate(con->physical.etag, sce->etag);
-
-				response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
-			}
-		}
-
-		/* prepare header */
-		if (NULL == (ds = (data_string *)array_get_element(con->response.headers, "Last-Modified"))) {
-			mtime = strftime_cache_get(srv, sce->st.st_mtime);
-			response_header_overwrite(srv, con, CONST_STR_LEN("Last-Modified"), CONST_BUF_LEN(mtime));
-		} else {
-			mtime = ds->value;
-		}
-
-		if (HANDLER_FINISHED == http_response_handle_cachable(srv, con, mtime)) {
-			return HANDLER_FINISHED;
-		}
-	}
-
-	if (con->request.http_range && con->conf.range_requests) {
+	if (HANDLER_FINISHED == http_response_handle_cachable(srv, con, mtime)) {
+		return HANDLER_FINISHED;
+	} else if (con->conf.range_requests &&
+	           NULL != array_get_element(con->request.headers, CONST_STR_LEN("Range"))) {
 		int do_range_request = 1;
 		/* check if we have a conditional GET */
 
-		if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "If-Range"))) {
+		if (NULL != (ds = (data_string *)array_get_element(con->request.headers, CONST_STR_LEN("If-Range")))) {
 			/* if the value is the same as our ETag, we do a Range-request,
 			 * otherwise a full 200 */
 
-			if (ds->value->ptr[0] == '"') {
-				/**
-				 * client wants a ETag
-				 */
-				if (!con->physical.etag) {
-					do_range_request = 0;
-				} else if (!buffer_is_equal(ds->value, con->physical.etag)) {
-					do_range_request = 0;
-				}
-			} else if (!mtime) {
-				/**
-				 * we don't have a Last-Modified and can match the If-Range: 
-				 *
-				 * sending all
-				 */
-				do_range_request = 0;
-			} else if (!buffer_is_equal(ds->value, mtime)) {
+			if (!buffer_is_equal(ds->value, con->physical.etag)) {
 				do_range_request = 0;
 			}
 		}
 
 		if (do_range_request) {
 			/* content prepared, I'm done */
-			con->file_finished = 1;
+			con->send->is_closed = 1;
 
 			if (0 == http_response_parse_range(srv, con, p)) {
 				con->http_status = 206;
@@ -536,23 +459,75 @@ URIHANDLER_FUNC(mod_staticfile_subrequest) {
 	/* we add it here for all requests
 	 * the HEAD request will drop it afterwards again
 	 */
-	http_chunk_append_file(srv, con, con->physical.path, 0, sce->st.st_size);
+	chunkqueue_append_file(con->send, con->physical.path, 0, sce->st.st_size);
 
-	con->http_status = 200;
-	con->file_finished = 1;
+	con->send->is_closed = 1;
+	con->send->bytes_in = sce->st.st_size;
 
 	return HANDLER_FINISHED;
 }
 
+/**
+ * mark all the content as read
+ */
+CONNECTION_FUNC(mod_staticfile_dev_null) {
+	chunk *c;
+	chunkqueue *in = con->recv;
+
+	UNUSED(srv);
+	UNUSED(p_d);
+
+	if (con->mode != DIRECT) return HANDLER_GO_ON;
+
+	/* there is nothing that we have to send out anymore */
+	if (in->bytes_in == in->bytes_out &&
+	    in->is_closed) return HANDLER_GO_ON;
+
+	for (c = in->first; in->bytes_out < in->bytes_in; c = c->next) {
+		off_t weWant = in->bytes_in - in->bytes_out;
+		off_t weHave = 0;
+
+		/* we announce toWrite octects
+		 * now take all the request_content chunk that we need to fill this request
+		 */
+
+		switch (c->type) {
+		case FILE_CHUNK:
+			weHave = c->file.length - c->offset;
+
+			if (weHave > weWant) weHave = weWant;
+
+			c->offset += weHave;
+			in->bytes_out += weHave;
+
+			break;
+		case MEM_CHUNK:
+			/* append to the buffer */
+			weHave = c->mem->used - 1 - c->offset;
+
+			if (weHave > weWant) weHave = weWant;
+
+			c->offset += weHave;
+			in->bytes_out += weHave;
+
+			break;
+		default:
+			break;
+		}
+	}
+
+	return HANDLER_GO_ON;
+
+}
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_staticfile_plugin_init(plugin *p);
-int mod_staticfile_plugin_init(plugin *p) {
+LI_EXPORT int mod_staticfile_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("staticfile");
 
 	p->init        = mod_staticfile_init;
-	p->handle_subrequest_start = mod_staticfile_subrequest;
+	p->handle_start_backend = mod_staticfile_subrequest;
+	p->handle_send_request_content = mod_staticfile_dev_null;
 	p->set_defaults  = mod_staticfile_set_defaults;
 	p->cleanup     = mod_staticfile_free;
 
