@@ -1,24 +1,28 @@
-#include "server.h"
-#include "connections.h"
-#include "response.h"
-#include "connections.h"
-#include "log.h"
-
-#include "plugin.h"
-
-#include "inet_ntop_cache.h"
-
+/*
+ * make sure _GNU_SOURCE is defined.
+ */
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
 #include <sys/types.h>
 
 #include <fcntl.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 #include <errno.h>
 #include <time.h>
 #include <stdio.h>
 
-#include "version.h"
+#include "server.h"
+#include "connections.h"
+#include "response.h"
+#include "connections.h"
+#include "log.h"
+#include "status_counter.h"
+
+#include "plugin.h"
+
+#include "inet_ntop_cache.h"
 
 typedef struct {
 	buffer *config_url;
@@ -57,6 +61,8 @@ INIT_FUNC(mod_status_init) {
 	plugin_data *p;
 	size_t i;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->traffic_out = p->requests = 0;
@@ -115,10 +121,9 @@ SETDEFAULTS_FUNC(mod_status_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
 		s = calloc(1, sizeof(plugin_config));
@@ -134,7 +139,7 @@ SETDEFAULTS_FUNC(mod_status_set_defaults) {
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 	}
@@ -145,24 +150,24 @@ SETDEFAULTS_FUNC(mod_status_set_defaults) {
 
 
 static int mod_status_row_append(buffer *b, const char *key, const char *value) {
-	buffer_append_string_len(b, CONST_STR_LEN("   <tr>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("    <td><b>"));
+	BUFFER_APPEND_STRING_CONST(b, "   <tr>\n");
+	BUFFER_APPEND_STRING_CONST(b, "    <td><b>");
 	buffer_append_string(b, key);
-	buffer_append_string_len(b, CONST_STR_LEN("</b></td>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("    <td>"));
+	BUFFER_APPEND_STRING_CONST(b, "</b></td>\n");
+	BUFFER_APPEND_STRING_CONST(b, "    <td>");
 	buffer_append_string(b, value);
-	buffer_append_string_len(b, CONST_STR_LEN("</td>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("   </tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</td>\n");
+	BUFFER_APPEND_STRING_CONST(b, "   </tr>\n");
 
 	return 0;
 }
 
 static int mod_status_header_append(buffer *b, const char *key) {
-	buffer_append_string_len(b, CONST_STR_LEN("   <tr>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("    <th colspan=\"2\">"));
+	BUFFER_APPEND_STRING_CONST(b, "   <tr>\n");
+	BUFFER_APPEND_STRING_CONST(b, "    <th colspan=\"2\">");
 	buffer_append_string(b, key);
-	buffer_append_string_len(b, CONST_STR_LEN("</th>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("   </tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</th>\n");
+	BUFFER_APPEND_STRING_CONST(b, "   </tr>\n");
 
 	return 0;
 }
@@ -171,13 +176,13 @@ static int mod_status_header_append_sort(buffer *b, void *p_d, const char* key)
 	plugin_data *p = p_d;
 
 	if (p->conf.sort) {
-		buffer_append_string_len(b, CONST_STR_LEN("<th class=\"status\"><a href=\"#\" class=\"sortheader\" onclick=\"resort(this);return false;\">"));
+		BUFFER_APPEND_STRING_CONST(b, "<th class=\"status\"><a href=\"#\" class=\"sortheader\" onclick=\"resort(this);return false;\">");
 		buffer_append_string(b, key);
-		buffer_append_string_len(b, CONST_STR_LEN("<span class=\"sortarrow\">:</span></a></th>\n"));
+		BUFFER_APPEND_STRING_CONST(b, "<span class=\"sortarrow\">:</span></a></th>\n");
 	} else {
-		buffer_append_string_len(b, CONST_STR_LEN("<th class=\"status\">"));
+		BUFFER_APPEND_STRING_CONST(b, "<th class=\"status\">");
 		buffer_append_string(b, key);
-		buffer_append_string_len(b, CONST_STR_LEN("</th>\n"));
+		BUFFER_APPEND_STRING_CONST(b, "</th>\n");
 	}
 
 	return 0;
@@ -200,40 +205,43 @@ static int mod_status_get_multiplier(double *avg, char *multiplier, int size) {
 
 static handler_t mod_status_handle_server_status_html(server *srv, connection *con, void *p_d) {
 	plugin_data *p = p_d;
-	buffer *b = buffer_init();
+	buffer *b;
 	size_t j;
 	double avg;
 	char multiplier = '\0';
-	char buf[32];
+	char buf[128];
 	time_t ts;
 
 	int days, hours, mins, seconds;
 
-	buffer_copy_string_len(b, CONST_STR_LEN(
+	b = chunkqueue_get_append_buffer(con->send);
+
+	BUFFER_COPY_STRING_CONST(b,
 				 "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
 				 "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
 				 "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
 				 "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
 				 " <head>\n"
-				 "  <title>Status</title>\n"
+				 "  <title>Status</title>\n");
 
+	BUFFER_APPEND_STRING_CONST(b,
 				   "  <style type=\"text/css\">\n"
 				   "    table.status { border: black solid thin; }\n"
-				   "    td { white-space: nowrap; }\n"
 				   "    td.int { background-color: #f0f0f0; text-align: right }\n"
 				   "    td.string { background-color: #f0f0f0; text-align: left }\n"
 				   "    th.status { background-color: black; color: white; font-weight: bold; }\n"
 				   "    a.sortheader { background-color: black; color: white; font-weight: bold; text-decoration: none; display: block; }\n"
 				   "    span.sortarrow { color: white; text-decoration: none; }\n"
-				   "  </style>\n"));
+				   "  </style>\n");
 
 	if (p->conf.sort) {
-		buffer_append_string_len(b, CONST_STR_LEN(
+		BUFFER_APPEND_STRING_CONST(b,
 					   "<script type=\"text/javascript\">\n"
 					   "// <!--\n"
 					   "var sort_column;\n"
-					   "var prev_span = null;\n"
+					   "var prev_span = null;\n");
 
+		BUFFER_APPEND_STRING_CONST(b,
 					   "function get_inner_text(el) {\n"
 					   " if((typeof el == 'string')||(typeof el == 'undefined'))\n"
 					   "  return el;\n"
@@ -249,8 +257,9 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 					   "  }\n"
 					   " }\n"
 					   " return str;\n"
-					   "}\n"
+					   "}\n");
 
+		BUFFER_APPEND_STRING_CONST(b,
 					   "function sortfn(a,b) {\n"
 					   " var at = get_inner_text(a.cells[sort_column]);\n"
 					   " var bt = get_inner_text(b.cells[sort_column]);\n"
@@ -263,8 +272,9 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 					   "  else if (aa<bb) return -1;\n"
 					   "  else return 1;\n"
 					   " }\n"
-					   "}\n"
+					   "}\n");
 
+		BUFFER_APPEND_STRING_CONST(b,
 					   "function resort(lnk) {\n"
 					   " var span = lnk.childNodes[1];\n"
 					   " var table = lnk.parentNode.parentNode.parentNode.parentNode;\n"
@@ -272,8 +282,9 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 					   " for (j=1;j<table.rows.length;j++)\n"
 					   "  rows[j-1] = table.rows[j];\n"
 					   " sort_column = lnk.parentNode.cellIndex;\n"
-					   " rows.sort(sortfn);\n"
+					   " rows.sort(sortfn);\n");
 
+		BUFFER_APPEND_STRING_CONST(b,
 					   " if (prev_span != null) prev_span.innerHTML = '';\n"
 					   " if (span.getAttribute('sortdir')=='down') {\n"
 					   "  span.innerHTML = '&uarr;';\n"
@@ -288,25 +299,25 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 					   " prev_span = span;\n"
 					   "}\n"
 					   "// -->\n"
-					   "</script>\n"));
+					   "</script>\n");
 	}
 
-	buffer_append_string_len(b, CONST_STR_LEN(
+	BUFFER_APPEND_STRING_CONST(b,
 				 " </head>\n"
-				 " <body>\n"));
+				 " <body>\n");
 
 
 
 	/* connection listing */
-	buffer_append_string_len(b, CONST_STR_LEN("<h1>Server-Status (" PACKAGE_NAME " " PACKAGE_VERSION ")</h1>"));
+	BUFFER_APPEND_STRING_CONST(b, "<h1>Server-Status</h1>");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<table summary=\"status\" class=\"status\">"));
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Hostname</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<table class=\"status\" id=\"status\" summary=\"Server Status\">");
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Hostname</td><td class=\"string\"><span id=\"host_addr\">");
 	buffer_append_string_buffer(b, con->uri.authority);
-	buffer_append_string_len(b, CONST_STR_LEN(" ("));
+	BUFFER_APPEND_STRING_CONST(b, "</span> (<span id=\"host_name\">");
 	buffer_append_string_buffer(b, con->server_name);
-	buffer_append_string_len(b, CONST_STR_LEN(")</td></tr>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Uptime</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>)</td></tr>\n");
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Uptime</td><td class=\"string\" id=\"uptime\">");
 
 	ts = srv->cur_ts - srv->startup_ts;
 
@@ -322,99 +333,99 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 	seconds = ts;
 
 	if (days) {
-		buffer_append_int(b, days);
-		buffer_append_string_len(b, CONST_STR_LEN(" days "));
+		buffer_append_long(b, days);
+		BUFFER_APPEND_STRING_CONST(b, " days ");
 	}
 
 	if (hours) {
-		buffer_append_int(b, hours);
-		buffer_append_string_len(b, CONST_STR_LEN(" hours "));
+		buffer_append_long(b, hours);
+		BUFFER_APPEND_STRING_CONST(b, " hours ");
 	}
 
 	if (mins) {
-		buffer_append_int(b, mins);
-		buffer_append_string_len(b, CONST_STR_LEN(" min "));
+		buffer_append_long(b, mins);
+		BUFFER_APPEND_STRING_CONST(b, " min ");
 	}
 
-	buffer_append_int(b, seconds);
-	buffer_append_string_len(b, CONST_STR_LEN(" s"));
+	buffer_append_long(b, seconds);
+	BUFFER_APPEND_STRING_CONST(b, " s");
 
-	buffer_append_string_len(b, CONST_STR_LEN("</td></tr>\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Started at</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n");
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Started at</td><td class=\"string\">");
 
 	ts = srv->startup_ts;
 
-	strftime(buf, sizeof(buf) - 1, "%Y-%m-%d %H:%M:%S", localtime(&ts));
+	strftime(buf, sizeof(buf) - 1, "<span id=\"start_date\">%Y-%m-%d</span> <span id=\"start_time\">%H:%M:%S</span>", localtime(&ts));
 	buffer_append_string(b, buf);
-	buffer_append_string_len(b, CONST_STR_LEN("</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n");
 
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><th colspan=\"2\">absolute (since start)</th></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><th colspan=\"2\">absolute (since start)</th></tr>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Requests</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Requests</td><td class=\"string\" ><span id=\"requests\">");
 	avg = p->abs_requests;
 
 	mod_status_get_multiplier(&avg, &multiplier, 1000);
 
-	buffer_append_int(b, avg);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "</span> <span id=\"requests_mult\">");
 	if (multiplier)	buffer_append_string_len(b, &multiplier, 1);
-	buffer_append_string_len(b, CONST_STR_LEN("req</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>req</td></tr>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Traffic</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Traffic</td><td class=\"string\"><span id=\"traffic\">");
 	avg = p->abs_traffic_out;
 
 	mod_status_get_multiplier(&avg, &multiplier, 1024);
 
 	sprintf(buf, "%.2f", avg);
 	buffer_append_string(b, buf);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	BUFFER_APPEND_STRING_CONST(b, "</span>  <span id=\"traffic_mult\">");
 	if (multiplier)	buffer_append_string_len(b, &multiplier, 1);
-	buffer_append_string_len(b, CONST_STR_LEN("byte</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>byte</td></tr>\n");
 
 
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><th colspan=\"2\">average (since start)</th></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><th colspan=\"2\">average (since start)</th></tr>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Requests</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Requests</td><td class=\"string\"><span id=\"requests_avg\">");
 	avg = p->abs_requests / (srv->cur_ts - srv->startup_ts);
 
 	mod_status_get_multiplier(&avg, &multiplier, 1000);
 
-	buffer_append_int(b, avg);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "</span> <span id=\"requests_avg_mult\">");
 	if (multiplier)	buffer_append_string_len(b, &multiplier, 1);
-	buffer_append_string_len(b, CONST_STR_LEN("req/s</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>req/s</td></tr>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Traffic</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Traffic</td><td class=\"string\"><span id=\"traffic_avg\">");
 	avg = p->abs_traffic_out / (srv->cur_ts - srv->startup_ts);
 
 	mod_status_get_multiplier(&avg, &multiplier, 1024);
 
 	sprintf(buf, "%.2f", avg);
 	buffer_append_string(b, buf);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	BUFFER_APPEND_STRING_CONST(b, "</span> <span id=\"traffic_avg_mult\">");
 	if (multiplier)	buffer_append_string_len(b, &multiplier, 1);
-	buffer_append_string_len(b, CONST_STR_LEN("byte/s</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>byte/s</td></tr>\n");
 
 
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><th colspan=\"2\">average (5s sliding average)</th></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><th colspan=\"2\">average (5s sliding average)</th></tr>\n");
 	for (j = 0, avg = 0; j < 5; j++) {
 		avg += p->mod_5s_requests[j];
 	}
 
 	avg /= 5;
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Requests</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Requests</td><td class=\"string\"><span id=\"requests_sliding_avg\">");
 
 	mod_status_get_multiplier(&avg, &multiplier, 1000);
 
-	buffer_append_int(b, avg);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "</span> <span id=\"requests_sliding_avg_mult\">");
 	if (multiplier)	buffer_append_string_len(b, &multiplier, 1);
 
-	buffer_append_string_len(b, CONST_STR_LEN("req/s</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>req/s</td></tr>\n");
 
 	for (j = 0, avg = 0; j < 5; j++) {
 		avg += p->mod_5s_traffic_out[j];
@@ -422,51 +433,44 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 
 	avg /= 5;
 
-	buffer_append_string_len(b, CONST_STR_LEN("<tr><td>Traffic</td><td class=\"string\">"));
+	BUFFER_APPEND_STRING_CONST(b, "<tr><td>Traffic</td><td class=\"string\"><span id=\"requests_sliding_traffic\">");
 
 	mod_status_get_multiplier(&avg, &multiplier, 1024);
 
 	sprintf(buf, "%.2f", avg);
 	buffer_append_string(b, buf);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	BUFFER_APPEND_STRING_CONST(b, "</span> <span id=\"requests_sliding_traffic_mult\">");
 	if (multiplier)	buffer_append_string_len(b, &multiplier, 1);
-	buffer_append_string_len(b, CONST_STR_LEN("byte/s</td></tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</span>byte/s</td></tr>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("</table>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</table>\n");
 
 
-	buffer_append_string_len(b, CONST_STR_LEN(
-		"<hr />\n<pre><b>legend</b>\n"
-		". = connect, C = close, E = hard error, k = keep-alive\n"
-		"r = read, R = read-POST, W = write, h = handle-request\n"
-		"q = request-start,  Q = request-end\n"
-		"s = response-start, S = response-end\n"));
+	BUFFER_APPEND_STRING_CONST(b, "<hr />\n<pre><b>legend</b>\n");
+	BUFFER_APPEND_STRING_CONST(b, ". = connect, C = close, E = hard error\n");
+	BUFFER_APPEND_STRING_CONST(b, "r = read, R = read-POST, W = write, h = handle-request\n");
+	BUFFER_APPEND_STRING_CONST(b, "q = request-start,  Q = request-end\n");
+	BUFFER_APPEND_STRING_CONST(b, "s = response-start, S = response-end\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<b>"));
-	buffer_append_int(b, srv->conns->used);
-	buffer_append_string_len(b, CONST_STR_LEN(" connections</b>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "<strong><span id=\"connections\">");
+	buffer_append_long(b, srv->conns->used);
+	BUFFER_APPEND_STRING_CONST(b, "</span> connections</strong>\n");
 
 	for (j = 0; j < srv->conns->used; j++) {
 		connection *c = srv->conns->ptr[j];
-		const char *state;
-
-		if (CON_STATE_READ == c->state && !buffer_string_is_empty(c->request.orig_uri)) {
-			state = "k";
-		} else {
-			state = connection_get_short_state(c->state);
-		}
+		const char *state = connection_get_short_state(c->state);
 
 		buffer_append_string_len(b, state, 1);
 
 		if (((j + 1) % 50) == 0) {
-			buffer_append_string_len(b, CONST_STR_LEN("\n"));
+			BUFFER_APPEND_STRING_CONST(b, "\n");
 		}
 	}
 
-	buffer_append_string_len(b, CONST_STR_LEN("\n</pre><hr />\n<h2>Connections</h2>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "\n</pre><hr />\n<h2>Connections</h2>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("<table summary=\"status\" class=\"status\">\n"));
-	buffer_append_string_len(b, CONST_STR_LEN("<tr>"));
+	BUFFER_APPEND_STRING_CONST(b, "<table class=\"status\" summary=\"Current connections\" id=\"clients\">\n");
+	BUFFER_APPEND_STRING_CONST(b, "<tr>");
 	mod_status_header_append_sort(b, p_d, "Client IP");
 	mod_status_header_append_sort(b, p_d, "Read");
 	mod_status_header_append_sort(b, p_d, "Written");
@@ -475,89 +479,73 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 	mod_status_header_append_sort(b, p_d, "Host");
 	mod_status_header_append_sort(b, p_d, "URI");
 	mod_status_header_append_sort(b, p_d, "File");
-	buffer_append_string_len(b, CONST_STR_LEN("</tr>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "</tr>\n");
 
 	for (j = 0; j < srv->conns->used; j++) {
 		connection *c = srv->conns->ptr[j];
 
-		buffer_append_string_len(b, CONST_STR_LEN("<tr><td class=\"string\">"));
+		BUFFER_APPEND_STRING_CONST(b, "<tr><td class=\"string ip\">");
 
 		buffer_append_string(b, inet_ntop_cache_get_ip(srv, &(c->dst_addr)));
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"int\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"int bytes_read\">");
 
-		if (c->request.content_length) {
-			buffer_append_int(b, c->request_content_queue->bytes_in);
-			buffer_append_string_len(b, CONST_STR_LEN("/"));
-			buffer_append_int(b, c->request.content_length);
+		if (c->request.content_length != -1) {
+			buffer_append_long(b, c->recv->bytes_in);
+			BUFFER_APPEND_STRING_CONST(b, "/");
+			buffer_append_long(b, c->request.content_length);
 		} else {
-			buffer_append_string_len(b, CONST_STR_LEN("0/0"));
+			BUFFER_APPEND_STRING_CONST(b, "0/0");
 		}
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"int\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"int bytes_written\">");
 
-		buffer_append_int(b, c->write_queue->bytes_out);
-		buffer_append_string_len(b, CONST_STR_LEN("/"));
-		buffer_append_int(b, c->write_queue->bytes_out + chunkqueue_length(c->write_queue));
+		buffer_append_off_t(b, chunkqueue_written(c->send_raw));
+		BUFFER_APPEND_STRING_CONST(b, "/");
+		buffer_append_off_t(b, chunkqueue_length(c->send_raw));
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"string\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"string state\">");
 
-		if (CON_STATE_READ == c->state && !buffer_string_is_empty(c->request.orig_uri)) {
-			buffer_append_string_len(b, CONST_STR_LEN("keep-alive"));
-		} else {
-			buffer_append_string(b, connection_get_state(c->state));
-		}
+		buffer_append_string(b, connection_get_state(c->state));
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"int\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"int time\">");
 
-		buffer_append_int(b, srv->cur_ts - c->request_start);
+		buffer_append_long(b, srv->cur_ts - c->request_start);
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"string\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"string host\">");
 
-		if (buffer_string_is_empty(c->server_name)) {
+		if (buffer_is_empty(c->server_name)) {
 			buffer_append_string_buffer(b, c->uri.authority);
 		}
 		else {
 			buffer_append_string_buffer(b, c->server_name);
 		}
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"string\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"string uri\">");
 
-		if (!buffer_string_is_empty(c->uri.path)) {
+		if (!buffer_is_empty(c->uri.path)) {
 			buffer_append_string_encoded(b, CONST_BUF_LEN(c->uri.path), ENCODING_HTML);
 		}
 
-		if (!buffer_string_is_empty(c->uri.query)) {
-			buffer_append_string_len(b, CONST_STR_LEN("?"));
-			buffer_append_string_encoded(b, CONST_BUF_LEN(c->uri.query), ENCODING_HTML);
-		}
-
-		if (!buffer_string_is_empty(c->request.orig_uri)) {
-			buffer_append_string_len(b, CONST_STR_LEN(" ("));
-			buffer_append_string_encoded(b, CONST_BUF_LEN(c->request.orig_uri), ENCODING_HTML);
-			buffer_append_string_len(b, CONST_STR_LEN(")"));
-		}
-		buffer_append_string_len(b, CONST_STR_LEN("</td><td class=\"string\">"));
+		BUFFER_APPEND_STRING_CONST(b, "</td><td class=\"string file\">");
 
 		buffer_append_string_buffer(b, c->physical.path);
 
-		buffer_append_string_len(b, CONST_STR_LEN("</td></tr>\n"));
+		BUFFER_APPEND_STRING_CONST(b, "</td></tr>\n");
 	}
 
 
-	buffer_append_string_len(b, CONST_STR_LEN(
-		      "</table>\n"));
+	BUFFER_APPEND_STRING_CONST(b,
+		      "</table>\n");
 
 
-	buffer_append_string_len(b, CONST_STR_LEN(
+	BUFFER_APPEND_STRING_CONST(b,
 		      " </body>\n"
 		      "</html>\n"
-		      ));
-
-	chunkqueue_append_buffer(con->write_queue, b);
-	buffer_free(b);
+		      );
 
 	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
+	con->send->bytes_in += b->used-1;
 
 	return 0;
 }
@@ -565,59 +553,55 @@ static handler_t mod_status_handle_server_status_html(server *srv, connection *c
 
 static handler_t mod_status_handle_server_status_text(server *srv, connection *con, void *p_d) {
 	plugin_data *p = p_d;
-	buffer *b = buffer_init();
+	buffer *b;
 	double avg;
 	time_t ts;
-	char buf[32];
-	unsigned int k;
-	unsigned int l;
+	size_t j;
+
+	b = chunkqueue_get_append_buffer(con->send);
 
 	/* output total number of requests */
-	buffer_append_string_len(b, CONST_STR_LEN("Total Accesses: "));
+	BUFFER_APPEND_STRING_CONST(b, "Total Accesses: ");
 	avg = p->abs_requests;
-	snprintf(buf, sizeof(buf) - 1, "%.0f", avg);
-	buffer_append_string(b, buf);
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
 	/* output total traffic out in kbytes */
-	buffer_append_string_len(b, CONST_STR_LEN("Total kBytes: "));
+	BUFFER_APPEND_STRING_CONST(b, "Total kBytes: ");
 	avg = p->abs_traffic_out / 1024;
-	snprintf(buf, sizeof(buf) - 1, "%.0f", avg);
-	buffer_append_string(b, buf);
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
 	/* output uptime */
-	buffer_append_string_len(b, CONST_STR_LEN("Uptime: "));
+	BUFFER_APPEND_STRING_CONST(b, "Uptime: ");
 	ts = srv->cur_ts - srv->startup_ts;
-	buffer_append_int(b, ts);
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
+	buffer_append_long(b, ts);
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
 	/* output busy servers */
-	buffer_append_string_len(b, CONST_STR_LEN("BusyServers: "));
-	buffer_append_int(b, srv->conns->used);
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
+	BUFFER_APPEND_STRING_CONST(b, "BusyServers: ");
+	buffer_append_long(b, srv->conns->used);
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN("IdleServers: "));
-	buffer_append_int(b, srv->conns->size - srv->conns->used);
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
+	/* output traffic */
+	BUFFER_APPEND_STRING_CONST(b, "Traffic: ");
+	avg = p->abs_traffic_out / (srv->cur_ts - srv->startup_ts);
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
-	/* output scoreboard */
-	buffer_append_string_len(b, CONST_STR_LEN("Scoreboard: "));
-	for (k = 0; k < srv->conns->used; k++) {
-		connection *c = srv->conns->ptr[k];
-		const char *state = connection_get_short_state(c->state);
-		buffer_append_string_len(b, state, 1);
+	/* output traffic 5s */
+	BUFFER_APPEND_STRING_CONST(b, "Traffic5s: ");
+	for (j = 0, avg = 0; j < 5; j++) {
+		avg += p->mod_5s_traffic_out[j];
 	}
-	for (l = 0; l < srv->conns->size - srv->conns->used; l++) {
-		buffer_append_string_len(b, CONST_STR_LEN("_"));
-	}
-	buffer_append_string_len(b, CONST_STR_LEN("\n"));
-
-	chunkqueue_append_buffer(con->write_queue, b);
-	buffer_free(b);
+	avg /= 5;
+	buffer_append_long(b, avg);
+	BUFFER_APPEND_STRING_CONST(b, "\n");
 
 	/* set text/plain output */
+
 	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/plain"));
+	con->send->bytes_in += b->used-1;
 
 	return 0;
 }
@@ -625,34 +609,34 @@ static handler_t mod_status_handle_server_status_text(server *srv, connection *c
 static handler_t mod_status_handle_server_statistics(server *srv, connection *con, void *p_d) {
 	buffer *b;
 	size_t i;
-	array *st = srv->status;
+	array *st = status_counter_get_array();
+
 	UNUSED(p_d);
 
 	if (0 == st->used) {
 		/* we have nothing to send */
 		con->http_status = 204;
-		con->file_finished = 1;
+		con->send->is_closed = 1;
 
 		return HANDLER_FINISHED;
 	}
 
-	b = buffer_init();
+	b = chunkqueue_get_append_buffer(con->send);
+
 	for (i = 0; i < st->used; i++) {
 		size_t ndx = st->sorted[i];
 
 		buffer_append_string_buffer(b, st->data[ndx]->key);
-		buffer_append_string_len(b, CONST_STR_LEN(": "));
-		buffer_append_int(b, ((data_integer *)(st->data[ndx]))->value);
-		buffer_append_string_len(b, CONST_STR_LEN("\n"));
+		buffer_append_string(b, ": ");
+		buffer_append_long(b, ((data_integer *)(st->data[ndx]))->value);
+		buffer_append_string(b, "\n");
 	}
 
-	chunkqueue_append_buffer(con->write_queue, b);
-	buffer_free(b);
-
 	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/plain"));
 
 	con->http_status = 200;
-	con->file_finished = 1;
+	con->send->bytes_in += b->used-1;
+	con->send->is_closed = 1;
 
 	return HANDLER_FINISHED;
 }
@@ -667,7 +651,7 @@ static handler_t mod_status_handle_server_status(server *srv, connection *con, v
 	}
 
 	con->http_status = 200;
-	con->file_finished = 1;
+	con->send->is_closed = 1;
 
 	return HANDLER_FINISHED;
 }
@@ -675,40 +659,39 @@ static handler_t mod_status_handle_server_status(server *srv, connection *con, v
 
 static handler_t mod_status_handle_server_config(server *srv, connection *con, void *p_d) {
 	plugin_data *p = p_d;
-	buffer *b = buffer_init();
-	buffer *m = p->module_list;
+	buffer *b, *m = p->module_list;
 	size_t i;
 
 	struct ev_map { fdevent_handler_t et; const char *name; } event_handlers[] =
 	{
-		/* - epoll is most reliable
+		/* - poll is most reliable
 		 * - select works everywhere
+		 * - linux-* are experimental
 		 */
-#ifdef USE_LINUX_EPOLL
-		{ FDEVENT_HANDLER_LINUX_SYSEPOLL, "linux-sysepoll" },
-#endif
 #ifdef USE_POLL
 		{ FDEVENT_HANDLER_POLL,           "poll" },
 #endif
 #ifdef USE_SELECT
 		{ FDEVENT_HANDLER_SELECT,         "select" },
 #endif
-#ifdef USE_LIBEV
-		{ FDEVENT_HANDLER_LIBEV,          "libev" },
+#ifdef USE_LINUX_EPOLL
+		{ FDEVENT_HANDLER_LINUX_SYSEPOLL, "linux-sysepoll" },
+#endif
+#ifdef USE_LINUX_SIGIO
+		{ FDEVENT_HANDLER_LINUX_RTSIG,    "linux-rtsig" },
 #endif
 #ifdef USE_SOLARIS_DEVPOLL
 		{ FDEVENT_HANDLER_SOLARIS_DEVPOLL,"solaris-devpoll" },
 #endif
-#ifdef USE_SOLARIS_PORT
-		{ FDEVENT_HANDLER_SOLARIS_PORT,   "solaris-eventports" },
-#endif
 #ifdef USE_FREEBSD_KQUEUE
 		{ FDEVENT_HANDLER_FREEBSD_KQUEUE, "freebsd-kqueue" },
 #endif
 		{ FDEVENT_HANDLER_UNSET,          NULL }
 	};
 
-	buffer_copy_string_len(b, CONST_STR_LEN(
+	b = chunkqueue_get_append_buffer(con->send);
+
+	BUFFER_COPY_STRING_CONST(b,
 			   "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
 			   "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
 			   "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
@@ -717,8 +700,8 @@ static handler_t mod_status_handle_server_config(server *srv, connection *con, v
 			   "  <title>Status</title>\n"
 			   " </head>\n"
 			   " <body>\n"
-			   "  <h1>" PACKAGE_DESC "</h1>\n"
-			   "  <table summary=\"status\" border=\"1\">\n"));
+			   "  <h1>" PACKAGE_NAME " " PACKAGE_VERSION "</h1>\n"
+			   "  <table summary=\"status\" border=\"1\">\n");
 
 	mod_status_header_append(b, "Server-Features");
 #ifdef HAVE_PCRE_H
@@ -743,43 +726,39 @@ static handler_t mod_status_handle_server_config(server *srv, connection *con, v
 		plugin *pl = ps[i];
 
 		if (i == 0) {
-			buffer_copy_buffer(m, pl->name);
+			buffer_copy_string_buffer(m, pl->name);
 		} else {
-			buffer_append_string_len(m, CONST_STR_LEN("<br />"));
+			BUFFER_APPEND_STRING_CONST(m, "<br />");
 			buffer_append_string_buffer(m, pl->name);
 		}
 	}
 
 	mod_status_row_append(b, "Loaded Modules", m->ptr);
 
-	buffer_append_string_len(b, CONST_STR_LEN("  </table>\n"));
+	BUFFER_APPEND_STRING_CONST(b, "  </table>\n");
 
-	buffer_append_string_len(b, CONST_STR_LEN(
+	BUFFER_APPEND_STRING_CONST(b,
 		      " </body>\n"
 		      "</html>\n"
-		      ));
-
-	chunkqueue_append_buffer(con->write_queue, b);
-	buffer_free(b);
+		      );
 
 	response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
 
 	con->http_status = 200;
-	con->file_finished = 1;
+	con->send->bytes_in += b->used-1;
+	con->send->is_closed = 1;
 
 	return HANDLER_FINISHED;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_status_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(status_url);
-	PATCH(config_url);
-	PATCH(sort);
-	PATCH(statistics_url);
+	PATCH_OPTION(status_url);
+	PATCH_OPTION(config_url);
+	PATCH_OPTION(sort);
+	PATCH_OPTION(statistics_url);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -794,13 +773,13 @@ static int mod_status_patch_connection(server *srv, connection *con, plugin_data
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("status.status-url"))) {
-				PATCH(status_url);
+				PATCH_OPTION(status_url);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("status.config-url"))) {
-				PATCH(config_url);
+				PATCH_OPTION(config_url);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("status.enable-sort"))) {
-				PATCH(sort);
+				PATCH_OPTION(sort);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("status.statistics-url"))) {
-				PATCH(statistics_url);
+				PATCH_OPTION(statistics_url);
 			}
 		}
 	}
@@ -811,17 +790,15 @@ static int mod_status_patch_connection(server *srv, connection *con, plugin_data
 static handler_t mod_status_handler(server *srv, connection *con, void *p_d) {
 	plugin_data *p = p_d;
 
-	if (con->mode != DIRECT) return HANDLER_GO_ON;
-
 	mod_status_patch_connection(srv, con, p);
 
-	if (!buffer_string_is_empty(p->conf.status_url) &&
+	if (!buffer_is_empty(p->conf.status_url) &&
 	    buffer_is_equal(p->conf.status_url, con->uri.path)) {
 		return mod_status_handle_server_status(srv, con, p_d);
-	} else if (!buffer_string_is_empty(p->conf.config_url) &&
+	} else if (!buffer_is_empty(p->conf.config_url) &&
 	    buffer_is_equal(p->conf.config_url, con->uri.path)) {
 		return mod_status_handle_server_config(srv, con, p_d);
-	} else if (!buffer_string_is_empty(p->conf.statistics_url) &&
+	} else if (!buffer_is_empty(p->conf.statistics_url) &&
 	    buffer_is_equal(p->conf.statistics_url, con->uri.path)) {
 		return mod_status_handle_server_statistics(srv, con, p_d);
 	}
@@ -872,8 +849,7 @@ REQUESTDONE_FUNC(mod_status_account) {
 	return HANDLER_GO_ON;
 }
 
-int mod_status_plugin_init(plugin *p);
-int mod_status_plugin_init(plugin *p) {
+LI_EXPORT int mod_status_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("status");
 
@@ -883,7 +859,7 @@ int mod_status_plugin_init(plugin *p) {
 
 	p->handle_uri_clean    = mod_status_handler;
 	p->handle_trigger      = mod_status_trigger;
-	p->handle_request_done = mod_status_account;
+	p->handle_response_done = mod_status_account;
 
 	p->data        = NULL;
 
