@@ -1,3 +1,7 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -6,23 +10,16 @@
 #include "response.h"
 #include "inet_ntop_cache.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <fcntl.h>
-#include <string.h>
-
-#if (defined(HAVE_GDBM_H) || defined(HAVE_MEMCACHE_H)) && defined(HAVE_PCRE_H)
-
 #if defined(HAVE_GDBM_H)
-# include <gdbm.h>
+#include <gdbm.h>
 #endif
 
 #if defined(HAVE_PCRE_H)
-# include <pcre.h>
+#include <pcre.h>
 #endif
 
 #if defined(HAVE_MEMCACHE_H)
-# include <memcache.h>
+#include <memcache.h>
 #endif
 
 /**
@@ -71,6 +68,8 @@ typedef struct {
 INIT_FUNC(mod_trigger_b4_dl_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->tmp_buf = buffer_init();
@@ -91,7 +90,7 @@ FREE_FUNC(mod_trigger_b4_dl_free) {
 		for (i = 0; i < srv->config_context->used; i++) {
 			plugin_config *s = p->config_storage[i];
 
-			if (NULL == s) continue;
+			if (!s) continue;
 
 			buffer_free(s->db_filename);
 			buffer_free(s->download_url);
@@ -145,10 +144,9 @@ SETDEFAULTS_FUNC(mod_trigger_b4_dl_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
-		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 #if defined(HAVE_PCRE_H)
 		const char *errptr;
@@ -174,21 +172,20 @@ SETDEFAULTS_FUNC(mod_trigger_b4_dl_set_defaults) {
 
 		p->config_storage[i] = s;
 
-		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
+		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
 			return HANDLER_ERROR;
 		}
 #if defined(HAVE_GDBM_H)
-		if (!buffer_string_is_empty(s->db_filename)) {
+		if (!buffer_is_empty(s->db_filename)) {
 			if (NULL == (s->db = gdbm_open(s->db_filename->ptr, 4096, GDBM_WRCREAT | GDBM_NOLOCK, S_IRUSR | S_IWUSR, 0))) {
 				log_error_write(srv, __FILE__, __LINE__, "s",
 						"gdbm-open failed");
 				return HANDLER_ERROR;
 			}
-			fd_close_on_exec(gdbm_fdesc(s->db));
 		}
 #endif
 #if defined(HAVE_PCRE_H)
-		if (!buffer_string_is_empty(s->download_url)) {
+		if (!buffer_is_empty(s->download_url)) {
 			if (NULL == (s->download_regex = pcre_compile(s->download_url->ptr,
 								      0, &errptr, &erroff, NULL))) {
 
@@ -199,7 +196,7 @@ SETDEFAULTS_FUNC(mod_trigger_b4_dl_set_defaults) {
 			}
 		}
 
-		if (!buffer_string_is_empty(s->trigger_url)) {
+		if (!buffer_is_empty(s->trigger_url)) {
 			if (NULL == (s->trigger_regex = pcre_compile(s->trigger_url->ptr,
 								     0, &errptr, &erroff, NULL))) {
 
@@ -246,25 +243,23 @@ SETDEFAULTS_FUNC(mod_trigger_b4_dl_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_trigger_b4_dl_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
 #if defined(HAVE_GDBM)
-	PATCH(db);
+	PATCH_OPTION(db);
 #endif
 #if defined(HAVE_PCRE_H)
-	PATCH(download_regex);
-	PATCH(trigger_regex);
+	PATCH_OPTION(download_regex);
+	PATCH_OPTION(trigger_regex);
 #endif
-	PATCH(trigger_timeout);
-	PATCH(deny_url);
-	PATCH(mc_namespace);
-	PATCH(debug);
+	PATCH_OPTION(trigger_timeout);
+	PATCH_OPTION(deny_url);
+	PATCH_OPTION(mc_namespace);
+	PATCH_OPTION(debug);
 #if defined(HAVE_MEMCACHE_H)
-	PATCH(mc);
+	PATCH_OPTION(mc);
 #endif
 
 	/* skip the first, the global context */
@@ -281,27 +276,27 @@ static int mod_trigger_b4_dl_patch_connection(server *srv, connection *con, plug
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.download-url"))) {
 #if defined(HAVE_PCRE_H)
-				PATCH(download_regex);
+				PATCH_OPTION(download_regex);
 #endif
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.trigger-url"))) {
 # if defined(HAVE_PCRE_H)
-				PATCH(trigger_regex);
+				PATCH_OPTION(trigger_regex);
 # endif
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.gdbm-filename"))) {
 #if defined(HAVE_GDBM_H)
-				PATCH(db);
+				PATCH_OPTION(db);
 #endif
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.trigger-timeout"))) {
-				PATCH(trigger_timeout);
+				PATCH_OPTION(trigger_timeout);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.debug"))) {
-				PATCH(debug);
+				PATCH_OPTION(debug);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.deny-url"))) {
-				PATCH(deny_url);
+				PATCH_OPTION(deny_url);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.memcache-namespace"))) {
-				PATCH(mc_namespace);
+				PATCH_OPTION(mc_namespace);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("trigger-before-download.memcache-hosts"))) {
 #if defined(HAVE_MEMCACHE_H)
-				PATCH(mc);
+				PATCH_OPTION(mc);
 #endif
 			}
 		}
@@ -309,7 +304,6 @@ static int mod_trigger_b4_dl_patch_connection(server *srv, connection *con, plug
 
 	return 0;
 }
-#undef PATCH
 
 URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	plugin_data *p = p_d;
@@ -321,9 +315,7 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 # define N 10
 	int ovec[N * 3];
 
-	if (con->mode != DIRECT) return HANDLER_GO_ON;
-
-	if (buffer_is_empty(con->uri.path)) return HANDLER_GO_ON;
+	if (con->uri.path->used == 0) return HANDLER_GO_ON;
 
 	mod_trigger_b4_dl_patch_connection(srv, con, p);
 
@@ -344,7 +336,7 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	if (!p->conf.mc) return HANDLER_GO_ON;
 # endif
 
-	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "X-Forwarded-For"))) {
+	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, CONST_STR_LEN("X-Forwarded-For")))) {
 		/* X-Forwarded-For contains the ip behind the proxy */
 
 		remote_ip = ds->value->ptr;
@@ -359,7 +351,7 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	}
 
 	/* check if URL is a trigger -> insert IP into DB */
-	if ((n = pcre_exec(p->conf.trigger_regex, NULL, CONST_BUF_LEN(con->uri.path), 0, 0, ovec, 3 * N)) < 0) {
+	if ((n = pcre_exec(p->conf.trigger_regex, NULL, con->uri.path->ptr, con->uri.path->used - 1, 0, 0, ovec, 3 * N)) < 0) {
 		if (n != PCRE_ERROR_NOMATCH) {
 			log_error_write(srv, __FILE__, __LINE__, "sd",
 					"execution error while matching:", n);
@@ -386,12 +378,11 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 # endif
 # if defined(HAVE_MEMCACHE_H)
 		if (p->conf.mc) {
-			size_t i, len;
-			buffer_copy_buffer(p->tmp_buf, p->conf.mc_namespace);
+			size_t i;
+			buffer_copy_string_buffer(p->tmp_buf, p->conf.mc_namespace);
 			buffer_append_string(p->tmp_buf, remote_ip);
 
-			len = buffer_string_length(p->tmp_buf);
-			for (i = 0; i < len; i++) {
+			for (i = 0; i < p->tmp_buf->used - 1; i++) {
 				if (p->tmp_buf->ptr[i] == ' ') p->tmp_buf->ptr[i] = '-';
 			}
 
@@ -411,7 +402,7 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 	}
 
 	/* check if URL is a download -> check IP in DB, update timestamp */
-	if ((n = pcre_exec(p->conf.download_regex, NULL, CONST_BUF_LEN(con->uri.path), 0, 0, ovec, 3 * N)) < 0) {
+	if ((n = pcre_exec(p->conf.download_regex, NULL, con->uri.path->ptr, con->uri.path->used - 1, 0, 0, ovec, 3 * N)) < 0) {
 		if (n != PCRE_ERROR_NOMATCH) {
 			log_error_write(srv, __FILE__, __LINE__, "sd",
 					"execution error while matching: ", n);
@@ -433,13 +424,13 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 				/* not found, redirect */
 
 				response_header_insert(srv, con, CONST_STR_LEN("Location"), CONST_BUF_LEN(p->conf.deny_url));
+
 				con->http_status = 307;
-				con->file_finished = 1;
 
 				return HANDLER_FINISHED;
 			}
 
-			memcpy(&last_hit, val.dptr, sizeof(time_t));
+			last_hit = *(time_t *)(val.dptr);
 
 			free(val.dptr);
 
@@ -448,7 +439,6 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 
 				response_header_insert(srv, con, CONST_STR_LEN("Location"), CONST_BUF_LEN(p->conf.deny_url));
 				con->http_status = 307;
-				con->file_finished = 1;
 
 				if (p->conf.db) {
 					if (0 != gdbm_delete(p->conf.db, key)) {
@@ -473,13 +463,12 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 # if defined(HAVE_MEMCACHE_H)
 		if (p->conf.mc) {
 			void *r;
-			size_t i, len;
+			size_t i;
 
-			buffer_copy_buffer(p->tmp_buf, p->conf.mc_namespace);
+			buffer_copy_string_buffer(p->tmp_buf, p->conf.mc_namespace);
 			buffer_append_string(p->tmp_buf, remote_ip);
 
-			len = buffer_string_length(p->tmp_buf);
-			for (i = 0; i < len; i++) {
+			for (i = 0; i < p->tmp_buf->used - 1; i++) {
 				if (p->tmp_buf->ptr[i] == ' ') p->tmp_buf->ptr[i] = '-';
 			}
 
@@ -500,7 +489,6 @@ URIHANDLER_FUNC(mod_trigger_b4_dl_uri_handler) {
 				response_header_insert(srv, con, CONST_STR_LEN("Location"), CONST_BUF_LEN(p->conf.deny_url));
 
 				con->http_status = 307;
-				con->file_finished = 1;
 
 				return HANDLER_FINISHED;
 			}
@@ -558,7 +546,7 @@ TRIGGER_FUNC(mod_trigger_b4_dl_handle_trigger) {
 
 			val = gdbm_fetch(s->db, key);
 
-			memcpy(&last_hit, val.dptr, sizeof(time_t));
+			last_hit = *(time_t *)(val.dptr);
 
 			free(val.dptr);
 
@@ -579,8 +567,7 @@ TRIGGER_FUNC(mod_trigger_b4_dl_handle_trigger) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_trigger_b4_dl_plugin_init(plugin *p);
-int mod_trigger_b4_dl_plugin_init(plugin *p) {
+LI_EXPORT int mod_trigger_b4_dl_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("trigger_b4_dl");
 
@@ -596,15 +583,3 @@ int mod_trigger_b4_dl_plugin_init(plugin *p) {
 
 	return 0;
 }
-
-#else
-
-#pragma message("(either gdbm or libmemcache) and pcre are required, but were not found")
-
-int mod_trigger_b4_dl_plugin_init(plugin *p);
-int mod_trigger_b4_dl_plugin_init(plugin *p) {
-	UNUSED(p);
-	return -1;
-}
-
-#endif
