@@ -1,3 +1,8 @@
+#include <sys/types.h>
+
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
@@ -5,14 +10,10 @@
 #include "response.h"
 
 #include "plugin.h"
-
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
+#include "sys-files.h"
 
 #ifdef HAVE_PWD_H
-# include <pwd.h>
+#include <pwd.h>
 #endif
 
 /* plugin config for all request/connections */
@@ -21,8 +22,6 @@ typedef struct {
 	array *include_user;
 	buffer *path;
 	buffer *basepath;
-	unsigned short letterhomes;
-	unsigned short active;
 } plugin_config;
 
 typedef struct {
@@ -40,6 +39,8 @@ typedef struct {
 INIT_FUNC(mod_userdir_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	p->username = buffer_init();
@@ -89,14 +90,12 @@ SETDEFAULTS_FUNC(mod_userdir_set_defaults) {
 		{ "userdir.exclude-user",       NULL, T_CONFIG_ARRAY,  T_CONFIG_SCOPE_CONNECTION },       /* 1 */
 		{ "userdir.include-user",       NULL, T_CONFIG_ARRAY,  T_CONFIG_SCOPE_CONNECTION },       /* 2 */
 		{ "userdir.basepath",           NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },       /* 3 */
-		{ "userdir.letterhomes",        NULL, T_CONFIG_BOOLEAN,T_CONFIG_SCOPE_CONNECTION },       /* 4 */
-		{ "userdir.active",             NULL, T_CONFIG_BOOLEAN,T_CONFIG_SCOPE_CONNECTION },       /* 5 */
 		{ NULL,                         NULL, T_CONFIG_UNSET,  T_CONFIG_SCOPE_UNSET }
 	};
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
@@ -106,17 +105,11 @@ SETDEFAULTS_FUNC(mod_userdir_set_defaults) {
 		s->include_user = array_init();
 		s->path = buffer_init();
 		s->basepath = buffer_init();
-		s->letterhomes = 0;
-		/* enabled by default for backward compatibility; if userdir.path isn't set userdir is disabled too,
-		 * but you can't disable it by setting it to an empty string. */
-		s->active = 1;
 
 		cv[0].destination = s->path;
 		cv[1].destination = s->exclude_user;
 		cv[2].destination = s->include_user;
 		cv[3].destination = s->basepath;
-		cv[4].destination = &(s->letterhomes);
-		cv[5].destination = &(s->active);
 
 		p->config_storage[i] = s;
 
@@ -128,18 +121,14 @@ SETDEFAULTS_FUNC(mod_userdir_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_userdir_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(path);
-	PATCH(exclude_user);
-	PATCH(include_user);
-	PATCH(basepath);
-	PATCH(letterhomes);
-	PATCH(active);
+	PATCH_OPTION(path);
+	PATCH_OPTION(exclude_user);
+	PATCH_OPTION(include_user);
+	PATCH_OPTION(basepath);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -154,27 +143,23 @@ static int mod_userdir_patch_connection(server *srv, connection *con, plugin_dat
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("userdir.path"))) {
-				PATCH(path);
+				PATCH_OPTION(path);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("userdir.exclude-user"))) {
-				PATCH(exclude_user);
+				PATCH_OPTION(exclude_user);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("userdir.include-user"))) {
-				PATCH(include_user);
+				PATCH_OPTION(include_user);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("userdir.basepath"))) {
-				PATCH(basepath);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("userdir.letterhomes"))) {
-				PATCH(letterhomes);
-			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("userdir.active"))) {
-				PATCH(active);
+				PATCH_OPTION(basepath);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 URIHANDLER_FUNC(mod_userdir_docroot_handler) {
 	plugin_data *p = p_d;
+	int uri_len;
 	size_t k;
 	char *rel_url;
 #ifdef HAVE_PWD_H
@@ -185,10 +170,7 @@ URIHANDLER_FUNC(mod_userdir_docroot_handler) {
 
 	mod_userdir_patch_connection(srv, con, p);
 
-	/* enforce the userdir.path to be set in the config, ugly fix for #1587;
-	 * should be replaced with a clean .enabled option in 1.5
-	 */
-	if (!p->conf.active || p->conf.path->used == 0) return HANDLER_GO_ON;
+	uri_len = con->uri.path->used - 1;
 
 	/* /~user/foo.html -> /home/user/public_html/foo.html */
 
@@ -268,19 +250,12 @@ URIHANDLER_FUNC(mod_userdir_docroot_handler) {
 				return HANDLER_GO_ON;
 			}
 		}
-		if (con->conf.force_lowercase_filenames) {
-			buffer_to_lower(p->username);
-		}
 
 		buffer_copy_string_buffer(p->temp_path, p->conf.basepath);
-		BUFFER_APPEND_SLASH(p->temp_path);
-		if (p->conf.letterhomes) {
-			buffer_append_string_len(p->temp_path, p->username->ptr, 1);
-			BUFFER_APPEND_SLASH(p->temp_path);
-		}
+		PATHNAME_APPEND_SLASH(p->temp_path);
 		buffer_append_string_buffer(p->temp_path, p->username);
 	}
-	BUFFER_APPEND_SLASH(p->temp_path);
+	PATHNAME_APPEND_SLASH(p->temp_path);
 	buffer_append_string_buffer(p->temp_path, p->conf.path);
 
 	if (buffer_is_empty(p->conf.basepath)) {
@@ -293,26 +268,8 @@ URIHANDLER_FUNC(mod_userdir_docroot_handler) {
 		}
 	}
 
-	buffer_copy_string_buffer(con->physical.basedir, p->temp_path);
-
-	/* the physical rel_path is basically the same as uri.path;
-	 * but it is converted to lowercase in case of force_lowercase_filenames and some special handling
-	 * for trailing '.', ' ' and '/' on windows
-	 * we assume that no docroot/physical handler changed this
-	 * (docroot should only set the docroot/server name, phyiscal should only change the phyiscal.path;
-	 *  the exception mod_secure_download doesn't work with userdir anyway)
-	 */
-	BUFFER_APPEND_SLASH(p->temp_path);
-	/* if no second '/' is found, we assume that it was stripped from the uri.path for the special handling
-	 * on windows.
-	 * we do not care about the trailing slash here on windows, as we already ensured it is a directory
-	 *
-	 * TODO: what to do with trailing dots in usernames on windows? they may result in the same directory
-	 *       as a username without them.
-	 */
-	if (NULL != (rel_url = strchr(con->physical.rel_path->ptr + 2, '/'))) {
-		buffer_append_string(p->temp_path, rel_url + 1); /* skip the / */
-	}
+	PATHNAME_APPEND_SLASH(p->temp_path);
+	buffer_append_string(p->temp_path, rel_url + 1); /* skip the / */
 	buffer_copy_string_buffer(con->physical.path, p->temp_path);
 
 	buffer_reset(p->temp_path);
@@ -322,7 +279,6 @@ URIHANDLER_FUNC(mod_userdir_docroot_handler) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_userdir_plugin_init(plugin *p);
 int mod_userdir_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("userdir");
