@@ -1,21 +1,25 @@
+#include <ctype.h>
+#include <stdlib.h>
+#include <string.h>
+
 #include "base.h"
 #include "log.h"
 #include "buffer.h"
 
 #include "plugin.h"
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-
-#include "md5.h"
+#ifdef USE_OPENSSL
+# include <openssl/md5.h>
+#else
+# include "md5.h"
+#endif
 
 /* plugin config for all request/connections */
 
 typedef struct {
 	buffer *cookie_name;
 	buffer *cookie_domain;
-	unsigned int cookie_max_age;
+	unsigned short cookie_max_age;
 } plugin_config;
 
 typedef struct {
@@ -30,6 +34,8 @@ typedef struct {
 INIT_FUNC(mod_usertrack_init) {
 	plugin_data *p;
 
+	UNUSED(srv);
+
 	p = calloc(1, sizeof(*p));
 
 	return p;
@@ -69,7 +75,7 @@ SETDEFAULTS_FUNC(mod_usertrack_set_defaults) {
 
 	config_values_t cv[] = {
 		{ "usertrack.cookie-name",       NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },       /* 0 */
-		{ "usertrack.cookie-max-age",    NULL, T_CONFIG_INT, T_CONFIG_SCOPE_CONNECTION },          /* 1 */
+		{ "usertrack.cookie-max-age",    NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },        /* 1 */
 		{ "usertrack.cookie-domain",     NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },       /* 2 */
 
 		{ "usertrack.cookiename",        NULL, T_CONFIG_DEPRECATED, T_CONFIG_SCOPE_CONNECTION },
@@ -78,7 +84,7 @@ SETDEFAULTS_FUNC(mod_usertrack_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		plugin_config *s;
@@ -99,7 +105,7 @@ SETDEFAULTS_FUNC(mod_usertrack_set_defaults) {
 		}
 
 		if (buffer_is_empty(s->cookie_name)) {
-			buffer_copy_string_len(s->cookie_name, CONST_STR_LEN("TRACKID"));
+			buffer_copy_string(s->cookie_name, "TRACKID");
 		} else {
 			size_t j;
 			for (j = 0; j < s->cookie_name->used - 1; j++) {
@@ -132,15 +138,13 @@ SETDEFAULTS_FUNC(mod_usertrack_set_defaults) {
 	return HANDLER_GO_ON;
 }
 
-#define PATCH(x) \
-	p->conf.x = s->x;
 static int mod_usertrack_patch_connection(server *srv, connection *con, plugin_data *p) {
 	size_t i, j;
 	plugin_config *s = p->config_storage[0];
 
-	PATCH(cookie_name);
-	PATCH(cookie_domain);
-	PATCH(cookie_max_age);
+	PATCH_OPTION(cookie_name);
+	PATCH_OPTION(cookie_domain);
+	PATCH_OPTION(cookie_max_age);
 
 	/* skip the first, the global context */
 	for (i = 1; i < srv->config_context->used; i++) {
@@ -155,31 +159,30 @@ static int mod_usertrack_patch_connection(server *srv, connection *con, plugin_d
 			data_unset *du = dc->value->data[j];
 
 			if (buffer_is_equal_string(du->key, CONST_STR_LEN("usertrack.cookie-name"))) {
-				PATCH(cookie_name);
+				PATCH_OPTION(cookie_name);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("usertrack.cookie-max-age"))) {
-				PATCH(cookie_max_age);
+				PATCH_OPTION(cookie_max_age);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("usertrack.cookie-domain"))) {
-				PATCH(cookie_domain);
+				PATCH_OPTION(cookie_domain);
 			}
 		}
 	}
 
 	return 0;
 }
-#undef PATCH
 
 URIHANDLER_FUNC(mod_usertrack_uri_handler) {
 	plugin_data *p = p_d;
 	data_string *ds;
 	unsigned char h[16];
-	li_MD5_CTX Md5Ctx;
+	MD5_CTX Md5Ctx;
 	char hh[32];
 
 	if (con->uri.path->used == 0) return HANDLER_GO_ON;
 
 	mod_usertrack_patch_connection(srv, con, p);
 
-	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, "Cookie"))) {
+	if (NULL != (ds = (data_string *)array_get_element(con->request.headers, CONST_STR_LEN("Cookie")))) {
 		char *g;
 		/* we have a cookie, does it contain a valid name ? */
 
@@ -210,38 +213,37 @@ URIHANDLER_FUNC(mod_usertrack_uri_handler) {
 	if (NULL == (ds = (data_string *)array_get_unused_element(con->response.headers, TYPE_STRING))) {
 		ds = data_response_init();
 	}
-	buffer_copy_string_len(ds->key, CONST_STR_LEN("Set-Cookie"));
+	buffer_copy_string(ds->key, "Set-Cookie");
 	buffer_copy_string_buffer(ds->value, p->conf.cookie_name);
-	buffer_append_string_len(ds->value, CONST_STR_LEN("="));
+	buffer_append_string(ds->value, "=");
 
 
 	/* taken from mod_auth.c */
 
 	/* generate shared-secret */
-	li_MD5_Init(&Md5Ctx);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)con->uri.path->ptr, con->uri.path->used - 1);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)"+", 1);
+	MD5_Init(&Md5Ctx);
+	MD5_Update(&Md5Ctx, (unsigned char *)con->uri.path->ptr, con->uri.path->used - 1);
+	MD5_Update(&Md5Ctx, (unsigned char *)"+", 1);
 
 	/* we assume sizeof(time_t) == 4 here, but if not it ain't a problem at all */
-	LI_ltostr(hh, srv->cur_ts);
-	li_MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
-	li_MD5_Update(&Md5Ctx, (unsigned char *)srv->entropy, sizeof(srv->entropy));
-	LI_ltostr(hh, rand());
-	li_MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
+	ltostr(hh, srv->cur_ts);
+	MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
+	ltostr(hh, rand());
+	MD5_Update(&Md5Ctx, (unsigned char *)hh, strlen(hh));
 
-	li_MD5_Final(h, &Md5Ctx);
+	MD5_Final(h, &Md5Ctx);
 
 	buffer_append_string_encoded(ds->value, (char *)h, 16, ENCODING_HEX);
-	buffer_append_string_len(ds->value, CONST_STR_LEN("; Path=/"));
-	buffer_append_string_len(ds->value, CONST_STR_LEN("; Version=1"));
+	buffer_append_string(ds->value, "; Path=/");
+	buffer_append_string(ds->value, "; Version=1");
 
 	if (!buffer_is_empty(p->conf.cookie_domain)) {
-		buffer_append_string_len(ds->value, CONST_STR_LEN("; Domain="));
+		buffer_append_string(ds->value, "; Domain=");
 		buffer_append_string_encoded(ds->value, CONST_BUF_LEN(p->conf.cookie_domain), ENCODING_REL_URI);
 	}
 
 	if (p->conf.cookie_max_age) {
-		buffer_append_string_len(ds->value, CONST_STR_LEN("; max-age="));
+		buffer_append_string(ds->value, "; max-age=");
 		buffer_append_long(ds->value, p->conf.cookie_max_age);
 	}
 
@@ -252,8 +254,7 @@ URIHANDLER_FUNC(mod_usertrack_uri_handler) {
 
 /* this function is called at dlopen() time and inits the callbacks */
 
-int mod_usertrack_plugin_init(plugin *p);
-int mod_usertrack_plugin_init(plugin *p) {
+LI_EXPORT int mod_usertrack_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
 	p->name        = buffer_init_string("usertrack");
 
