@@ -1,26 +1,25 @@
-#include "proc_open.h"
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <errno.h>
+#include "proc_open.h"
 
 #ifdef WIN32
-# include <io.h>
-# include <fcntl.h>
+#include <io.h>
+#include <fcntl.h>
 #else
-# include <sys/wait.h>
-# include <unistd.h>
+#include <sys/wait.h>
+#include <unistd.h>
 #endif
 
 
-#ifdef WIN32
+#ifdef _WIN32
 /* {{{ win32 stuff */
 # define SHELLENV "ComSpec"
 # define SECURITY_DC , SECURITY_ATTRIBUTES *security
 # define SECURITY_CC , security
 # define pipe(pair) (CreatePipe(&pair[0], &pair[1], security, 2048L) ? 0 : -1)
-static inline HANDLE dup_handle(HANDLE src, BOOL inherit, BOOL closeorig)
+static HANDLE dup_handle(HANDLE src, BOOL inherit, BOOL closeorig)
 {
 	HANDLE copy, self = GetCurrentProcess();
 
@@ -185,7 +184,7 @@ int proc_open(proc_handler_t *proc, const char *command) {
 		buffer_append_string(cmdline, shell);
 	} else {
 		buffer_append_string(cmdline, windir);
-		buffer_append_string_len(cmdline, CONST_STR_LEN("\\system32\\cmd.exe"));
+		buffer_append_string(cmdline, "\\system32\\cmd.exe");
 	}
 	buffer_append_string_len(cmdline, CONST_STR_LEN(" /c "));
 	buffer_append_string(cmdline, command);
@@ -256,7 +255,7 @@ int proc_open(proc_handler_t *proc, const char *command) {
 		 */
 		proc_close_parents(proc);
 
-		execl(shell, shell, "-c", command, (char *)NULL);
+		execl(shell, shell, "-c", command, NULL);
 		_exit(127);
 
 	} else if (child < 0) {
@@ -271,11 +270,11 @@ int proc_open(proc_handler_t *proc, const char *command) {
 	}
 }
 /* }}} */
-#endif /* WIN32 */
+#endif /* _WIN32 */
 
 /* {{{ proc_read_fd_to_buffer */
 static void proc_read_fd_to_buffer(int fd, buffer *b) {
-	ssize_t s;
+	int s; /* win32 has not ssize_t */
 
 	for (;;) {
 		buffer_prepare_append(b, 512);
@@ -288,40 +287,31 @@ static void proc_read_fd_to_buffer(int fd, buffer *b) {
 }
 /* }}} */
 /* {{{ proc_open_buffer */
-int proc_open_buffer(const char *command, buffer *in, buffer *out, buffer *err) {
-	proc_handler_t proc;
+int proc_open_buffer(proc_handler_t *proc, const char *command, buffer *in, buffer *out, buffer *err) {
 
-	if (proc_open(&proc, command) != 0) {
+	UNUSED(err);
+
+	if (proc_open(proc, command) != 0) {
 		return -1;
 	}
 
 	if (in) {
-		if (write(proc.in.fd, (void *)in->ptr, in->used) < 0) {
+		if (write(proc->in.fd, (void *)in->ptr, in->used) < 0) {
 			perror("error writing pipe");
 			return -1;
 		}
 	}
-	pipe_close(&proc.in);
+	pipe_close(&proc->in);
 
 	if (out) {
-		proc_read_fd_to_buffer(proc.out.fd, out);
+		proc_read_fd_to_buffer(proc->out.fd, out);
 	}
-	pipe_close(&proc.out);
+	pipe_close(&proc->out);
 
 	if (err) {
-		proc_read_fd_to_buffer(proc.err.fd, err);
-	} else {
-		buffer *tmp = buffer_init();
-		proc_read_fd_to_buffer(proc.err.fd, tmp);
-		if (tmp->used > 0 &&  write(2, (void*)tmp->ptr, tmp->used) < 0) {
-			perror("error writing pipe");
-			return -1;
-		}
-		buffer_free(tmp);
+		proc_read_fd_to_buffer(proc->err.fd, err);
 	}
-	pipe_close(&proc.err);
-
-	proc_close(&proc);
+	pipe_close(&proc->err);
 
 	return 0;
 }
@@ -376,7 +366,7 @@ int main() {
 		RESET();
 
 		fprintf(stdout, "test: echo 321 with read\n"); fflush(stdout);
-		if (proc_open_buffer("echo 321", NULL, out, err) != 0) {
+		if (proc_open_buffer(&proc, "echo 321", NULL, out, err) != 0) {
 			ERROR_OUT();
 		}
 		fprintf(stdout, "result: ->%s<-\n\n", out->ptr); fflush(stdout);
@@ -384,7 +374,7 @@ int main() {
 
 		fprintf(stdout, "test: echo 123 | " CMD_CAT "\n"); fflush(stdout);
 		buffer_copy_string_len(in, CONST_STR_LEN("123\n"));
-		if (proc_open_buffer(CMD_CAT, in, out, err) != 0) {
+		if (proc_open_buffer(&proc, CMD_CAT, in, out, err) != 0) {
 			ERROR_OUT();
 		}
 		fprintf(stdout, "result: ->%s<-\n\n", out->ptr); fflush(stdout);
