@@ -1,14 +1,3 @@
-#include "response.h"
-#include "keyvalue.h"
-#include "log.h"
-#include "stat_cache.h"
-#include "chunk.h"
-
-#include "configfile.h"
-#include "connections.h"
-
-#include "plugin.h"
-
 #include <sys/types.h>
 #include <sys/stat.h>
 
@@ -18,39 +7,58 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
-#include <unistd.h>
 #include <ctype.h>
 #include <assert.h>
 
 #include <stdio.h>
 
-#include "sys-socket.h"
-#include "version.h"
+#include "settings.h"
 
-int http_response_write_header(server *srv, connection *con) {
+#include "response.h"
+#include "keyvalue.h"
+#include "log.h"
+#include "stat_cache.h"
+#include "chunk.h"
+
+#include "connections.h"
+
+#include "plugin.h"
+
+#include "sys-socket.h"
+#include "sys-files.h"
+#include "sys-strings.h"
+
+int http_response_write_header(server *srv, connection *con, chunkqueue *raw) {
 	buffer *b;
 	size_t i;
 	int have_date = 0;
 	int have_server = 0;
+	int allow_keep_alive = 0;
 
-	b = chunkqueue_get_prepend_buffer(con->write_queue);
+	b = chunkqueue_get_prepend_buffer(raw);
 
 	if (con->request.http_version == HTTP_VERSION_1_1) {
-		buffer_copy_string_len(b, CONST_STR_LEN("HTTP/1.1 "));
+		BUFFER_COPY_STRING_CONST(b, "HTTP/1.1 ");
 	} else {
-		buffer_copy_string_len(b, CONST_STR_LEN("HTTP/1.0 "));
+		BUFFER_COPY_STRING_CONST(b, "HTTP/1.0 ");
 	}
 	buffer_append_long(b, con->http_status);
-	buffer_append_string_len(b, CONST_STR_LEN(" "));
+	BUFFER_APPEND_STRING_CONST(b, " ");
 	buffer_append_string(b, get_http_status_name(con->http_status));
 
-	/* disable keep-alive if requested */
-	if (con->request_count > con->conf.max_keep_alive_requests || 0 == con->conf.max_keep_alive_idle) {
-		con->keep_alive = 0;
-	} else {
-		con->keep_alive_idle = con->conf.max_keep_alive_idle;
+	if (con->response.transfer_encoding & HTTP_TRANSFER_ENCODING_CHUNKED) {
+		response_header_overwrite(srv, con, CONST_STR_LEN("Transfer-Encoding"), CONST_STR_LEN("chunked"));
+		allow_keep_alive = 1;
+	} else if (con->response.content_length >= 0) {
+		buffer_copy_off_t(srv->tmp_buf, con->response.content_length);
+		response_header_overwrite(srv, con, CONST_STR_LEN("Content-Length"), srv->tmp_buf->ptr, srv->tmp_buf->used - 1);
+
+		allow_keep_alive = 1;
 	}
 
+	/* keep-alive needs Content-Length or chunked encoding. */
+	if (!allow_keep_alive) con->keep_alive = 0;
+
 	if (con->request.http_version != HTTP_VERSION_1_1 || con->keep_alive == 0) {
 		if (con->keep_alive) {
 			response_header_overwrite(srv, con, CONST_STR_LEN("Connection"), CONST_STR_LEN("keep-alive"));
@@ -59,10 +67,6 @@ int http_response_write_header(server *srv, connection *con) {
 		}
 	}
 
-	if (con->response.transfer_encoding & HTTP_TRANSFER_ENCODING_CHUNKED) {
-		response_header_overwrite(srv, con, CONST_STR_LEN("Transfer-Encoding"), CONST_STR_LEN("chunked"));
-	}
-
 
 	/* add all headers */
 	for (i = 0; i < con->response.headers->used; i++) {
@@ -71,29 +75,25 @@ int http_response_write_header(server *srv, connection *con) {
 		ds = (data_string *)con->response.headers->data[i];
 
 		if (ds->value->used && ds->key->used &&
-		    0 != strncasecmp(ds->key->ptr, CONST_STR_LEN("X-LIGHTTPD-")) &&
-			0 != strncasecmp(ds->key->ptr, CONST_STR_LEN("X-Sendfile"))) {
-			if (0 == strcasecmp(ds->key->ptr, "Date")) have_date = 1;
-			if (0 == strcasecmp(ds->key->ptr, "Server")) have_server = 1;
-			if (0 == strcasecmp(ds->key->ptr, "Content-Encoding") && 304 == con->http_status) continue;
+		    0 != strncmp(ds->key->ptr, "X-LIGHTTPD-", sizeof("X-LIGHTTPD-") - 1) &&
+		    0 != strcasecmp(ds->key->ptr, "X-Sendfile")) {
+			if (buffer_is_equal_string(ds->key, CONST_STR_LEN("Date"))) have_date = 1;
+			if (buffer_is_equal_string(ds->key, CONST_STR_LEN("Server"))) have_server = 1;
 
-			buffer_append_string_len(b, CONST_STR_LEN("\r\n"));
+			BUFFER_APPEND_STRING_CONST(b, "\r\n");
 			buffer_append_string_buffer(b, ds->key);
-			buffer_append_string_len(b, CONST_STR_LEN(": "));
-#if 0
-			/** 
-			 * the value might contain newlines, encode them with at least one white-space
-			 */
-			buffer_append_string_encoded(b, CONST_BUF_LEN(ds->value), ENCODING_HTTP_HEADER);
-#else
+			BUFFER_APPEND_STRING_CONST(b, ": ");
 			buffer_append_string_buffer(b, ds->value);
+#if 0
+			log_error_write(srv, __FILE__, __LINE__, "bb",
+					ds->key, ds->value);
 #endif
 		}
 	}
 
 	if (!have_date) {
 		/* HTTP/1.1 requires a Date: header */
-		buffer_append_string_len(b, CONST_STR_LEN("\r\nDate: "));
+		BUFFER_APPEND_STRING_CONST(b, "\r\nDate: ");
 
 		/* cache the generated timestamp */
 		if (srv->cur_ts != srv->last_generated_date_ts) {
@@ -112,17 +112,18 @@ int http_response_write_header(server *srv, connection *con) {
 
 	if (!have_server) {
 		if (buffer_is_empty(con->conf.server_tag)) {
-			buffer_append_string_len(b, CONST_STR_LEN("\r\nServer: " PACKAGE_DESC));
-		} else if (con->conf.server_tag->used > 1) {
-			buffer_append_string_len(b, CONST_STR_LEN("\r\nServer: "));
-			buffer_append_string_encoded(b, CONST_BUF_LEN(con->conf.server_tag), ENCODING_HTTP_HEADER);
+			BUFFER_APPEND_STRING_CONST(b, "\r\nServer: " PACKAGE_NAME "/" PACKAGE_VERSION);
+		} else {
+			BUFFER_APPEND_STRING_CONST(b, "\r\nServer: ");
+			buffer_append_string_buffer(b, con->conf.server_tag);
 		}
 	}
 
-	buffer_append_string_len(b, CONST_STR_LEN("\r\n\r\n"));
+	BUFFER_APPEND_STRING_CONST(b, "\r\n\r\n");
 
 
 	con->bytes_header = b->used - 1;
+	raw->bytes_in += b->used - 1;
 
 	if (con->conf.log_response_header) {
 		log_error_write(srv, __FILE__, __LINE__, "sSb", "Response-Header:", "\n", b);
@@ -131,97 +132,15 @@ int http_response_write_header(server *srv, connection *con) {
 	return 0;
 }
 
-#ifdef USE_OPENSSL
-static void https_add_ssl_entries(connection *con) {
-	X509 *xs;
-	X509_NAME *xn;
-	X509_NAME_ENTRY *xe;
-	int i, nentries;
-
-	if (
-		SSL_get_verify_result(con->ssl) != X509_V_OK
-		|| !(xs = SSL_get_peer_certificate(con->ssl))
-	) {
-		return;
-	}
-
-	xn = X509_get_subject_name(xs);
-	for (i = 0, nentries = X509_NAME_entry_count(xn); i < nentries; ++i) {
-		int xobjnid;
-		const char * xobjsn;
-		data_string *envds;
-
-		if (!(xe = X509_NAME_get_entry(xn, i))) {
-			continue;
-		}
-		xobjnid = OBJ_obj2nid((ASN1_OBJECT*)X509_NAME_ENTRY_get_object(xe));
-		xobjsn = OBJ_nid2sn(xobjnid);
-		if (!xobjsn) {
-			continue;
-		}
-
-		if (NULL == (envds = (data_string *)array_get_unused_element(con->environment, TYPE_STRING))) {
-			envds = data_string_init();
-		}
-		buffer_copy_string_len(envds->key, CONST_STR_LEN("SSL_CLIENT_S_DN_"));
-		buffer_append_string(envds->key, xobjsn);
-		buffer_copy_string_len(
-			envds->value,
-			(const char *)xe->value->data, xe->value->length
-		);
-		/* pick one of the exported values as "REMOTE_USER", for example
-		 * ssl.verifyclient.username   = "SSL_CLIENT_S_DN_UID" or "SSL_CLIENT_S_DN_emailAddress"
-		 */
-		if (buffer_is_equal(con->conf.ssl_verifyclient_username, envds->key)) {
-			data_string *ds;
-			if (NULL == (ds = (data_string *)array_get_element(con->environment, "REMOTE_USER"))) {
-				if (NULL == (ds = (data_string *)array_get_unused_element(con->environment, TYPE_STRING))) {
-					ds = data_string_init();
-				}
-				buffer_copy_string(ds->key, "REMOTE_USER");
-				array_insert_unique(con->environment, (data_unset *)ds);
-			}
-			buffer_copy_string_buffer(ds->value, envds->value);
-		}
-		array_insert_unique(con->environment, (data_unset *)envds);
-	}
-	if (con->conf.ssl_verifyclient_export_cert) {
-		BIO *bio;
-		if (NULL != (bio = BIO_new(BIO_s_mem()))) {
-			data_string *envds;
-			int n;
-
-			PEM_write_bio_X509(bio, xs);
-			n = BIO_pending(bio);
-
-			if (NULL == (envds = (data_string *)array_get_unused_element(con->environment, TYPE_STRING))) {
-				envds = data_string_init();
-			}
-
-			buffer_copy_string_len(envds->key, CONST_STR_LEN("SSL_CLIENT_CERT"));
-			buffer_prepare_copy(envds->value, n+1);
-			BIO_read(bio, envds->value->ptr, n);
-			BIO_free(bio);
-			envds->value->ptr[n] = '\0';
-			envds->value->used = n+1;
-			array_insert_unique(con->environment, (data_unset *)envds);
-		}
-	}
-	X509_free(xs);
-}
-#endif
 
 
-handler_t http_response_prepare(server *srv, connection *con) {
+handler_t handle_get_backend(server *srv, connection *con) {
 	handler_t r;
 
-	/* looks like someone has already done a decision */
+	/* looks like someone has already made a decision */
 	if (con->mode == DIRECT &&
 	    (con->http_status != 0 && con->http_status != 200)) {
 		/* remove a packets in the queue */
-		if (con->file_finished == 0) {
-			chunkqueue_reset(con->write_queue);
-		}
 
 		return HANDLER_FINISHED;
 	}
@@ -230,7 +149,7 @@ handler_t http_response_prepare(server *srv, connection *con) {
 	if (con->mode == DIRECT && con->physical.path->used == 0) {
 		char *qstr;
 
-		/* we only come here when we have the parse the full request again
+		/* we only come here when we have to parse the full request again
 		 *
 		 * a HANDLER_COMEBACK from mod_rewrite and mod_fastcgi might be a
 		 * problem here as mod_setenv might get called multiple times
@@ -242,9 +161,6 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		 *
 		 *  */
 
-		config_cond_cache_reset(srv, con);
-		config_setup_connection(srv, con); /* Perhaps this could be removed at other places. */
-
 		if (con->conf.log_condition_handling) {
 			log_error_write(srv, __FILE__, __LINE__,  "s",  "run condition");
 		}
@@ -272,23 +188,15 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		 *
 		 */
 
-		/* initial scheme value. can be overwritten for example by mod_extforward later */
-		if (con->srv_socket->is_ssl) {
-			buffer_copy_string_len(con->uri.scheme, CONST_STR_LEN("https"));
-		} else {
-			buffer_copy_string_len(con->uri.scheme, CONST_STR_LEN("http"));
-		}
+		buffer_copy_string(con->uri.scheme, con->conf.is_ssl ? "https" : "http");
 		buffer_copy_string_buffer(con->uri.authority, con->request.http_host);
 		buffer_to_lower(con->uri.authority);
 
-		config_patch_connection(srv, con, COMP_HTTP_SCHEME);    /* Scheme:      */
 		config_patch_connection(srv, con, COMP_HTTP_HOST);      /* Host:        */
 		config_patch_connection(srv, con, COMP_HTTP_REMOTE_IP); /* Client-IP */
 		config_patch_connection(srv, con, COMP_HTTP_REFERER);   /* Referer:     */
 		config_patch_connection(srv, con, COMP_HTTP_USER_AGENT);/* User-Agent:  */
-		config_patch_connection(srv, con, COMP_HTTP_LANGUAGE);  /* Accept-Language:  */
 		config_patch_connection(srv, con, COMP_HTTP_COOKIE);    /* Cookie:  */
-		config_patch_connection(srv, con, COMP_HTTP_REQUEST_METHOD); /* REQUEST_METHOD */
 
 		/** their might be a fragment which has to be cut away */
 		if (NULL != (qstr = strchr(con->request.uri->ptr, '#'))) {
@@ -305,44 +213,23 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			buffer_copy_string_buffer(con->uri.path_raw, con->request.uri);
 		}
 
-		/* decode url to path
-		 *
-		 * - decode url-encodings  (e.g. %20 -> ' ')
-		 * - remove path-modifiers (e.g. /../)
-		 */
-
-		if (con->request.http_method == HTTP_METHOD_OPTIONS &&
-		    con->uri.path_raw->ptr[0] == '*' && con->uri.path_raw->ptr[1] == '\0') {
-			/* OPTIONS * ... */
-			buffer_copy_string_buffer(con->uri.path, con->uri.path_raw);
-		} else {
-			buffer_copy_string_buffer(srv->tmp_buf, con->uri.path_raw);
-			buffer_urldecode_path(srv->tmp_buf);
-			buffer_path_simplify(con->uri.path, srv->tmp_buf);
-		}
-
-		config_patch_connection(srv, con, COMP_HTTP_URL); /* HTTPurl */
-		config_patch_connection(srv, con, COMP_HTTP_QUERY_STRING); /* HTTPqs */
-
-#ifdef USE_OPENSSL
-		if (con->srv_socket->is_ssl && con->conf.ssl_verifyclient) {
-			https_add_ssl_entries(con);
-		}
-#endif
-
-		/* do we have to downgrade to 1.0 ? */
-		if (!con->conf.allow_http11) {
-			con->request.http_version = HTTP_VERSION_1_0;
-		}
-
 		if (con->conf.log_request_handling) {
 			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- splitting Request-URI");
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "Request-URI     : ", con->request.uri);
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-scheme      : ", con->uri.scheme);
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-authority   : ", con->uri.authority);
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-path (raw)  : ", con->uri.path_raw);
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-path (clean): ", con->uri.path);
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-query       : ", con->uri.query);
+			log_error_write(srv, __FILE__, __LINE__,  "sb", "Request-URI  : ", con->request.uri);
+			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-scheme   : ", con->uri.scheme);
+			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-authority: ", con->uri.authority);
+			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-path     : ", con->uri.path_raw);
+			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-query    : ", con->uri.query);
+		}
+
+		/* disable keep-alive if requested */
+
+		if (con->request_count > con->conf.max_keep_alive_requests) {
+			con->keep_alive = 0;
+		}
+
+		if (srv->sockets_disabled) {
+			con->keep_alive = 0;
 		}
 
 
@@ -367,6 +254,29 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			break;
 		}
 
+		/* build filename
+		 *
+		 * - decode url-encodings  (e.g. %20 -> ' ')
+		 * - remove path-modifiers (e.g. /../)
+		 */
+
+
+
+		if (con->request.http_method == HTTP_METHOD_OPTIONS &&
+		    con->uri.path_raw->ptr[0] == '*' && con->uri.path_raw->ptr[1] == '\0') {
+			/* OPTIONS * ... */
+			buffer_copy_string_buffer(con->uri.path, con->uri.path_raw);
+		} else {
+			buffer_copy_string_buffer(srv->tmp_buf, con->uri.path_raw);
+			buffer_urldecode_path(srv->tmp_buf);
+			buffer_path_simplify(con->uri.path, srv->tmp_buf);
+		}
+
+		if (con->conf.log_request_handling) {
+			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- sanatising URI");
+			log_error_write(srv, __FILE__, __LINE__,  "sb", "URI-path     : ", con->uri.path);
+		}
+
 		/**
 		 *
 		 * call plugins
@@ -375,6 +285,14 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		 *
 		 */
 
+		config_patch_connection(srv, con, COMP_HTTP_URL); /* HTTPurl */
+		config_patch_connection(srv, con, COMP_HTTP_QUERY_STRING); /* HTTPqs */
+
+		/* do we have to downgrade to 1.0 ? */
+		if (!con->conf.allow_http11) {
+			con->request.http_version = HTTP_VERSION_1_0;
+		}
+
 		switch(r = plugins_call_handle_uri_clean(srv, con)) {
 		case HANDLER_GO_ON:
 			break;
@@ -390,12 +308,13 @@ handler_t http_response_prepare(server *srv, connection *con) {
 
 		if (con->request.http_method == HTTP_METHOD_OPTIONS &&
 		    con->uri.path->ptr[0] == '*' && con->uri.path_raw->ptr[1] == '\0') {
-			/* option requests are handled directly without checking of the path */
+			/* option requests are handled directly without checking the path */
 
 			response_header_insert(srv, con, CONST_STR_LEN("Allow"), CONST_STR_LEN("OPTIONS, GET, HEAD, POST"));
 
 			con->http_status = 200;
-			con->file_finished = 1;
+			/* no more content to send */
+			con->send->is_closed = 1;
 
 			return HANDLER_FINISHED;
 		}
@@ -433,14 +352,15 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		buffer_copy_string_buffer(con->physical.doc_root, con->conf.document_root);
 		buffer_copy_string_buffer(con->physical.rel_path, con->uri.path);
 
-#if defined(__WIN32) || defined(__CYGWIN__)
-		/* strip dots from the end and spaces
+		filename_unix2local(con->physical.rel_path);
+#if defined(_WIN32) || defined(__CYGWIN__)
+		/* strip dots and spaces from the end
 		 *
 		 * windows/dos handle those filenames as the same file
 		 *
 		 * foo == foo. == foo..... == "foo...   " == "foo..  ./"
 		 *
-		 * This will affect in some cases PATHINFO
+		 * This will affect PATHINFO in some cases
 		 *
 		 * on native windows we could prepend the filename with \\?\ to circumvent
 		 * this behaviour. I have no idea how to push this through cygwin
@@ -491,15 +411,14 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			break;
 		}
 
-		/* MacOS X and Windows can't distiguish between upper and lower-case
-		 *
-		 * convert to lower-case
+		/* The default Mac OS X and Windows filesystems can't distiguish between
+		 * upper- and lowercase, so convert to lowercase
 		 */
 		if (con->conf.force_lowercase_filenames) {
 			buffer_to_lower(con->physical.rel_path);
 		}
 
-		/* the docroot plugins might set the servername, if they don't we take http-host */
+		/* the docroot plugins might set the servername; if they don't we take http-host */
 		if (buffer_is_empty(con->server_name)) {
 			buffer_copy_string_buffer(con->server_name, con->uri.authority);
 		}
@@ -510,16 +429,22 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		 *
 		 */
 
-		buffer_copy_string_buffer(con->physical.basedir, con->physical.doc_root);
 		buffer_copy_string_buffer(con->physical.path, con->physical.doc_root);
-		BUFFER_APPEND_SLASH(con->physical.path);
+		PATHNAME_APPEND_SLASH(con->physical.path);
+		buffer_copy_string_buffer(con->physical.basedir, con->physical.path);
 		if (con->physical.rel_path->used &&
-		    con->physical.rel_path->ptr[0] == '/') {
+		    con->physical.rel_path->ptr[0] == DIR_SEPERATOR) {
 			buffer_append_string_len(con->physical.path, con->physical.rel_path->ptr + 1, con->physical.rel_path->used - 2);
 		} else {
 			buffer_append_string_buffer(con->physical.path, con->physical.rel_path);
 		}
 
+		/* win32: directories can't have a trailing slash */
+		if (con->physical.path->ptr[con->physical.path->used - 2] == DIR_SEPERATOR) {
+			con->physical.path->ptr[con->physical.path->used - 2] = '\0';
+			con->physical.path->used--;
+		}
+
 		if (con->conf.log_request_handling) {
 			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- after doc_root");
 			log_error_write(srv, __FILE__, __LINE__,  "sb", "Doc-Root     :", con->physical.doc_root);
@@ -543,16 +468,16 @@ handler_t http_response_prepare(server *srv, connection *con) {
 		if (con->conf.log_request_handling) {
 			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- logical -> physical");
 			log_error_write(srv, __FILE__, __LINE__,  "sb", "Doc-Root     :", con->physical.doc_root);
-			log_error_write(srv, __FILE__, __LINE__,  "sb", "Basedir      :", con->physical.basedir);
 			log_error_write(srv, __FILE__, __LINE__,  "sb", "Rel-Path     :", con->physical.rel_path);
 			log_error_write(srv, __FILE__, __LINE__,  "sb", "Path         :", con->physical.path);
 		}
 	}
 
 	/*
-	 * Noone catched away the file from normal path of execution yet (like mod_access)
+	 * No one took the file away from the normal path of execution yet (like mod_access)
+	 *
+	 * we don't have a backend yet, try to resolve the physical path and go on
 	 *
-	 * Go on and check of the file exists at all
 	 */
 
 	if (con->mode == DIRECT) {
@@ -566,13 +491,15 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			log_error_write(srv, __FILE__, __LINE__,  "sb", "Path         :", con->physical.path);
 		}
 
-		if (HANDLER_ERROR != stat_cache_get_entry(srv, con, con->physical.path, &sce)) {
+		switch (stat_cache_get_entry_async(srv, con, con->physical.path, &sce)) {
+		case HANDLER_GO_ON:
 			/* file exists */
 
 			if (con->conf.log_request_handling) {
 				log_error_write(srv, __FILE__, __LINE__,  "s",  "-- file found");
 				log_error_write(srv, __FILE__, __LINE__,  "sb", "Path         :", con->physical.path);
 			}
+
 #ifdef HAVE_LSTAT
 			if ((sce->is_symlink != 0) && !con->conf.follow_symlink) {
 				con->http_status = 403;
@@ -586,6 +513,7 @@ handler_t http_response_prepare(server *srv, connection *con) {
 				return HANDLER_FINISHED;
 			};
 #endif
+
 			if (S_ISDIR(sce->st.st_mode)) {
 				if (con->uri.path->ptr[con->uri.path->used - 2] != '/') {
 					/* redirect to .../ */
@@ -603,7 +531,10 @@ handler_t http_response_prepare(server *srv, connection *con) {
 
 
 			}
-		} else {
+			break;
+		case HANDLER_WAIT_FOR_EVENT:
+			return HANDLER_WAIT_FOR_EVENT;
+		case HANDLER_ERROR:
 			switch (errno) {
 			case EACCES:
 				con->http_status = 403;
@@ -615,8 +546,6 @@ handler_t http_response_prepare(server *srv, connection *con) {
 
 				buffer_reset(con->physical.path);
 				return HANDLER_FINISHED;
-			case ENAMETOOLONG:
-				/* file name to be read was too long. return 404 */
 			case ENOENT:
 				con->http_status = 404;
 
@@ -630,15 +559,18 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			case ENOTDIR:
 				/* PATH_INFO ! :) */
 				break;
+			case EMFILE:
+				return HANDLER_WAIT_FOR_FD;
 			default:
-				/* we have no idea what happend. let's tell the user so. */
+				/* we have no idea what happened, so tell the user. */
 				con->http_status = 500;
-				buffer_reset(con->physical.path);
 
-				log_error_write(srv, __FILE__, __LINE__, "ssbsb",
-						"file not found ... or so: ", strerror(errno),
-						con->uri.path,
-						"->", con->physical.path);
+				ERROR("checking file '%s' (%s) failed: %d (%s) -> sending status 500",
+					BUF_STR(con->uri.path),
+					BUF_STR(con->physical.path),
+					errno, strerror(errno));
+
+				buffer_reset(con->physical.path);
 
 				return HANDLER_FINISHED;
 			}
@@ -648,14 +580,17 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			buffer_copy_string_buffer(srv->tmp_buf, con->physical.path);
 
 			do {
+				struct stat st;
+
 				if (slash) {
 					buffer_copy_string_len(con->physical.path, srv->tmp_buf->ptr, slash - srv->tmp_buf->ptr);
 				} else {
 					buffer_copy_string_buffer(con->physical.path, srv->tmp_buf);
 				}
 
-				if (HANDLER_ERROR != stat_cache_get_entry(srv, con, con->physical.path, &sce)) {
-					found = S_ISREG(sce->st.st_mode);
+				if (0 == stat(con->physical.path->ptr, &(st)) &&
+				    S_ISREG(st.st_mode)) {
+					found = 1;
 					break;
 				}
 
@@ -673,7 +608,7 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			} while ((found == 0) && (slash != NULL) && ((size_t)(slash - srv->tmp_buf->ptr) > (con->physical.basedir->used - 2)));
 
 			if (found == 0) {
-				/* no it really doesn't exists */
+				/* no, it really doesn't exists */
 				con->http_status = 404;
 
 				if (con->conf.log_file_not_found) {
@@ -687,20 +622,6 @@ handler_t http_response_prepare(server *srv, connection *con) {
 				return HANDLER_FINISHED;
 			}
 
-#ifdef HAVE_LSTAT
-			if ((sce->is_symlink != 0) && !con->conf.follow_symlink) {
-				con->http_status = 403;
-
-				if (con->conf.log_request_handling) {
-					log_error_write(srv, __FILE__, __LINE__,  "s",  "-- access denied due symlink restriction");
-					log_error_write(srv, __FILE__, __LINE__,  "sb", "Path         :", con->physical.path);
-				}
-
-				buffer_reset(con->physical.path);
-				return HANDLER_FINISHED;
-			};
-#endif
-
 			/* we have a PATHINFO */
 			if (pathinfo) {
 				buffer_copy_string(con->request.pathinfo, pathinfo);
@@ -721,55 +642,38 @@ handler_t http_response_prepare(server *srv, connection *con) {
 			}
 		}
 
+		config_patch_connection(srv, con, COMP_PHYSICAL_PATH);    /* physical-path  */
+
 		if (con->conf.log_request_handling) {
 			log_error_write(srv, __FILE__, __LINE__,  "s",  "-- handling subrequest");
 			log_error_write(srv, __FILE__, __LINE__,  "sb", "Path         :", con->physical.path);
 		}
 
 		/* call the handlers */
-		switch(r = plugins_call_handle_subrequest_start(srv, con)) {
+		switch(r = plugins_call_handle_start_backend(srv, con)) {
 		case HANDLER_GO_ON:
-			/* request was not handled */
-			break;
 		case HANDLER_FINISHED:
+			/* if we are still here, no one wanted the file; status 403 is ok I think */
+
 		default:
 			if (con->conf.log_request_handling) {
 				log_error_write(srv, __FILE__, __LINE__,  "s",  "-- subrequest finished");
 			}
 
-			/* something strange happend */
+			/* something strange happened */
 			return r;
 		}
-
-		/* if we are still here, no one wanted the file, status 403 is ok I think */
-
-		if (con->mode == DIRECT && con->http_status == 0) {
-			switch (con->request.http_method) {
-			case HTTP_METHOD_OPTIONS:
-				con->http_status = 200;
-				break;
-			default:
-				con->http_status = 403;
-			}
-
-			return HANDLER_FINISHED;
-		}
-
 	}
 
-	switch(r = plugins_call_handle_subrequest(srv, con)) {
-	case HANDLER_GO_ON:
-		/* request was not handled, looks like we are done */
+	if (con->mode == DIRECT) {
+		con->http_status = 403;
+
+		TRACE("%s", "aaaaaaah, sending 403");
+
 		return HANDLER_FINISHED;
-	case HANDLER_FINISHED:
-		/* request is finished */
-	default:
-		/* something strange happend */
-		return r;
+	} else {
+		return HANDLER_GO_ON;
 	}
-
-	/* can't happen */
-	return HANDLER_COMEBACK;
 }
 
 
