@@ -1,18 +1,20 @@
 #include <sys/types.h>
-#include <sys/time.h>
 #include <sys/stat.h>
 
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 #include <stdio.h>
-#include <unistd.h>
 #include <fcntl.h>
-
+#include <time.h>
+#include <fcntl.h>
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
+#include "base.h"
+#include "sys-socket.h"
+#include "sys-files.h"
 
 #ifdef HAVE_PWD_H
 #include <grp.h>
@@ -21,11 +23,18 @@
 
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#else
+#ifdef _WIN32
+#include "xgetopt.h"
+#endif
 #endif
 
 #define FCGI_LISTENSOCK_FILENO 0
 
-#include "sys-socket.h"
+#ifndef UNIX_PATH_MAX
+# define UNIX_PATH_MAX 108
+#endif
+
 
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
@@ -37,9 +46,9 @@ typedef int socklen_t;
 #endif
 
 #ifdef HAVE_SYS_UN_H
-int fcgi_spawn_connection(char *appPath, char **appArgv, char *addr, unsigned short port, const char *unixsocket, int fork_count, int child_count, int pid_fd, int nofork) {
+int fcgi_spawn_connection(char *appPath, char *addr, unsigned short port, const char *unixsocket, int child_count, int pid_fd, int nofork) {
 	int fcgi_fd;
-	int socket_type, status, rc = 0;
+	int socket_type, status;
 	struct timeval tv = { 0, 100 * 1000 };
 
 	struct sockaddr_un fcgi_addr_un;
@@ -58,7 +67,7 @@ int fcgi_spawn_connection(char *appPath, char **appArgv, char *addr, unsigned sh
 
 
 	if (unixsocket) {
-		memset(&fcgi_addr_un, 0, sizeof(fcgi_addr_un));
+		memset(&fcgi_addr, 0, sizeof(fcgi_addr));
 
 		fcgi_addr_un.sun_family = AF_UNIX;
 		strcpy(fcgi_addr_un.sun_path, unixsocket);
@@ -72,13 +81,12 @@ int fcgi_spawn_connection(char *appPath, char **appArgv, char *addr, unsigned sh
 		socket_type = AF_UNIX;
 		fcgi_addr = (struct sockaddr *) &fcgi_addr_un;
 	} else {
-		memset(&fcgi_addr_in, 0, sizeof(fcgi_addr_in));
 		fcgi_addr_in.sin_family = AF_INET;
-		if (addr != NULL) {
-			fcgi_addr_in.sin_addr.s_addr = inet_addr(addr);
-		} else {
-			fcgi_addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
-		}
+                if (addr != NULL) {
+                        fcgi_addr_in.sin_addr.s_addr = inet_addr(addr);
+                } else {
+                        fcgi_addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
+                }
 		fcgi_addr_in.sin_port = htons(port);
 		servlen = sizeof(fcgi_addr_in);
 
@@ -129,124 +137,111 @@ int fcgi_spawn_connection(char *appPath, char **appArgv, char *addr, unsigned sh
 			return -1;
 		}
 
-		while (fork_count-- > 0) {
+		if (!nofork) {
+			child = fork();
+		} else {
+			child = 0;
+		}
 
-			if (!nofork) {
-				child = fork();
-			} else {
-				child = 0;
+		switch (child) {
+		case 0: {
+			char cgi_childs[64];
+			char *b;
+			int max_fd = 0;
+
+			int i = 0;
+
+			/* loose control terminal */
+			setsid();
+
+			/* is save as we limit to 256 childs */
+			sprintf(cgi_childs, "PHP_FCGI_CHILDREN=%d", child_count);
+
+			if(fcgi_fd != FCGI_LISTENSOCK_FILENO) {
+				close(FCGI_LISTENSOCK_FILENO);
+				dup2(fcgi_fd, FCGI_LISTENSOCK_FILENO);
+				close(fcgi_fd);
 			}
 
-			switch (child) {
-			case 0: {
-				char cgi_childs[64];
-				int max_fd = 0;
+			max_fd = open("/dev/null", O_RDWR);
+			close(STDERR_FILENO);
+			dup2(max_fd, STDERR_FILENO);
+			close(max_fd);
 
-				int i = 0;
+			max_fd = open("/dev/null", O_RDWR);
+			close(STDOUT_FILENO);
+			dup2(max_fd, STDOUT_FILENO);
+			close(max_fd);
 
-				/* loose control terminal */
-				setsid();
+			/* we don't need the client socket */
+			for (i = 3; i < max_fd; i++) {
+				if (i != FCGI_LISTENSOCK_FILENO) close(i);
+			}
 
-				/* is safe as we limit to 256 childs */
-				sprintf(cgi_childs, "PHP_FCGI_CHILDREN=%d", child_count);
+			/* create environment */
 
-				if(fcgi_fd != FCGI_LISTENSOCK_FILENO) {
-					close(FCGI_LISTENSOCK_FILENO);
-					dup2(fcgi_fd, FCGI_LISTENSOCK_FILENO);
-					close(fcgi_fd);
+			putenv(cgi_childs);
+
+			/* fork and replace shell */
+			b = malloc(strlen("exec ") + strlen(appPath) + 1);
+			strcpy(b, "exec ");
+			strcat(b, appPath);
+
+			/* exec the cgi */
+			execl("/bin/sh", "sh", "-c", b, NULL);
+
+			exit(errno);
+
+			break;
+		}
+		case -1:
+			/* error */
+			break;
+		default:
+			/* father */
+
+			/* wait */
+			select(0, NULL, NULL, NULL, &tv);
+
+			switch (waitpid(child, &status, WNOHANG)) {
+			case 0:
+				fprintf(stderr, "%s.%d: child spawned successfully: PID: %d\n",
+					__FILE__, __LINE__,
+					child);
+
+				/* write pid file */
+				if (pid_fd != -1) {
+					/* assume a 32bit pid_t */
+					char pidbuf[12];
+
+					snprintf(pidbuf, sizeof(pidbuf) - 1, "%d", child);
+
+					write(pid_fd, pidbuf, strlen(pidbuf));
+					close(pid_fd);
+					pid_fd = -1;
 				}
 
-				max_fd = open("/dev/null", O_RDWR);
-				close(STDERR_FILENO);
-				dup2(max_fd, STDERR_FILENO);
-				close(max_fd);
-
-				max_fd = open("/dev/null", O_RDWR);
-				close(STDOUT_FILENO);
-				dup2(max_fd, STDOUT_FILENO);
-				close(max_fd);
-
-				/* we don't need the client socket */
-				for (i = 3; i < max_fd; i++) {
-					if (i != FCGI_LISTENSOCK_FILENO) close(i);
-				}
-
-				/* create environment */
-
-				putenv(cgi_childs);
-
-				/* fork and replace shell */
-				if (appArgv) {
-					execv(appArgv[0], appArgv);
-
-				} else {
-					char *b = malloc(strlen("exec ") + strlen(appPath) + 1);
-					strcpy(b, "exec ");
-					strcat(b, appPath);
-
-					/* exec the cgi */
-					execl("/bin/sh", "sh", "-c", b, (char *)NULL);
-				}
-
-				exit(errno);
-
 				break;
-			}
 			case -1:
-				/* error */
 				break;
 			default:
-				/* father */
-
-				/* wait */
-				select(0, NULL, NULL, NULL, &tv);
-
-				switch (waitpid(child, &status, WNOHANG)) {
-				case 0:
-					fprintf(stdout, "%s.%d: child spawned successfully: PID: %d\n",
+				if (WIFEXITED(status)) {
+					fprintf(stderr, "%s.%d: child exited with: %d, %s\n",
 						__FILE__, __LINE__,
-						child);
-
-					/* write pid file */
-					if (pid_fd != -1) {
-						/* assume a 32bit pid_t */
-						char pidbuf[12];
-
-						snprintf(pidbuf, sizeof(pidbuf) - 1, "%d", child);
-
-						write(pid_fd, pidbuf, strlen(pidbuf));
-						/* avoid eol for the last one */
-						if (fork_count != 0) {
-							write(pid_fd, "\n", 1);
-						}
-					}
-
-					break;
-				case -1:
-					break;
-				default:
-					if (WIFEXITED(status)) {
-						fprintf(stderr, "%s.%d: child exited with: %d\n",
-							__FILE__, __LINE__, WEXITSTATUS(status));
-						rc = WEXITSTATUS(status);
-					} else if (WIFSIGNALED(status)) {
-						fprintf(stderr, "%s.%d: child signaled: %d\n",
-							__FILE__, __LINE__,
-							WTERMSIG(status));
-						rc = 1;
-					} else {
-						fprintf(stderr, "%s.%d: child died somehow: %d\n",
-							__FILE__, __LINE__,
-							status);
-						rc = status;
-					}
+						WEXITSTATUS(status), strerror(WEXITSTATUS(status)));
+				} else if (WIFSIGNALED(status)) {
+					fprintf(stderr, "%s.%d: child signaled: %d\n",
+						__FILE__, __LINE__,
+						WTERMSIG(status));
+				} else {
+					fprintf(stderr, "%s.%d: child died somehow: %d\n",
+						__FILE__, __LINE__,
+						status);
 				}
-
-				break;
 			}
+
+			break;
 		}
-		close(pid_fd);
-		pid_fd = -1;
 	} else {
 		fprintf(stderr, "%s.%d: socket is already used, can't spawn\n",
 			__FILE__, __LINE__);
@@ -255,7 +250,7 @@ int fcgi_spawn_connection(char *appPath, char **appArgv, char *addr, unsigned sh
 
 	close(fcgi_fd);
 
-	return rc;
+	return 0;
 }
 
 
@@ -267,18 +262,14 @@ void show_version () {
 }
 
 void show_help () {
-	char *b = \
-"Usage: spawn-fcgi [options] -- <fcgiapp> [fcgi app arguments]\n" \
-"\n" \
-"spawn-fcgi v" PACKAGE_VERSION " - spawns fastcgi processes\n" \
-"\n" \
-"Options:\n" \
+	char *b = "spawn-fcgi" "-" PACKAGE_VERSION \
+" - spawns fastcgi processes\n" \
+"usage:\n" \
 " -f <fcgiapp> filename of the fcgi-application\n" \
 " -a <addr>    bind to ip address\n" \
 " -p <port>    bind to tcp-port\n" \
 " -s <path>    bind to unix-domain socket\n" \
 " -C <childs>  (PHP only) numbers of childs to spawn (default 5)\n" \
-" -F <childs>  numbers of childs to fork (default 1)\n" \
 " -P <path>    name of PID-file for spawed process\n" \
 " -n           no fork (for daemontools)\n" \
 " -v           show version\n" \
@@ -296,24 +287,20 @@ int main(int argc, char **argv) {
 	char *fcgi_app = NULL, *changeroot = NULL, *username = NULL,
                *groupname = NULL, *unixsocket = NULL, *pid_file = NULL,
                 *addr = NULL;
-	char **fcgi_app_argv = { NULL };
 	unsigned short port = 0;
 	int child_count = 5;
-	int fork_count = 1;
 	int i_am_root, o;
 	int pid_fd = -1;
 	int nofork = 0;
-	struct sockaddr_un un;
 
 	i_am_root = (getuid() == 0);
 
-	while (-1 != (o = getopt(argc, argv, "c:f:g:hna:p:u:vC:F:s:P:"))) {
+       while(-1 != (o = getopt(argc, argv, "c:f:g:hna:p:u:vC:s:P:"))) {
 		switch(o) {
 		case 'f': fcgi_app = optarg; break;
 		case 'a': addr = optarg;/* ip addr */ break;
 		case 'p': port = strtol(optarg, NULL, 10);/* port */ break;
 		case 'C': child_count = strtol(optarg, NULL, 10);/*  */ break;
-		case 'F': fork_count = strtol(optarg, NULL, 10);/*  */ break;
 		case 's': unixsocket = optarg; /* unix-domain socket */ break;
 		case 'c': if (i_am_root) { changeroot = optarg; }/* chroot() */ break;
 		case 'u': if (i_am_root) { username = optarg; } /* set user */ break;
@@ -328,11 +315,7 @@ int main(int argc, char **argv) {
 		}
 	}
 
-	if (optind < argc) {
-		fcgi_app_argv = &argv[optind];
-	}
-
-	if ((fcgi_app == NULL && fcgi_app_argv == NULL) || (port == 0 && unixsocket == NULL)) {
+	if (fcgi_app == NULL || (port == 0 && unixsocket == NULL)) {
 		show_help();
 		return -1;
 	}
@@ -345,7 +328,7 @@ int main(int argc, char **argv) {
 		return -1;
 	}
 
-	if (unixsocket && strlen(unixsocket) > sizeof(un.sun_path) - 1) {
+	if (unixsocket && strlen(unixsocket) > UNIX_PATH_MAX - 1) {
 		fprintf(stderr, "%s.%d: %s\n",
 			__FILE__, __LINE__,
 			"path of the unix socket is too long\n");
@@ -440,15 +423,6 @@ int main(int argc, char **argv) {
 					"I will not set gid to 0\n");
 				return -1;
 			}
-
-			/* do the change before we do the chroot() */
-			setgid(grp->gr_gid);
-			setgroups(0, NULL); 
-
-			if (username) {
-				initgroups(username, grp->gr_gid);
-			}
-
 		}
 
 		if (changeroot) {
@@ -467,12 +441,18 @@ int main(int argc, char **argv) {
 		}
 
 		/* drop root privs */
+		if (groupname) {
+			setgid(grp->gr_gid);
+		}
 		if (username) {
+			if (groupname) {
+				initgroups(username, grp->gr_gid);
+			}
 			setuid(pwd->pw_uid);
 		}
 	}
 
-       return fcgi_spawn_connection(fcgi_app, fcgi_app_argv, addr, port, unixsocket, fork_count, child_count, pid_fd, nofork);
+       return fcgi_spawn_connection(fcgi_app, addr, port, unixsocket, child_count, pid_fd, nofork);
 }
 #else
 int main() {
